/*
 * rfm69_globals.h
 *
 *  Created on: 7. juni 2018
 *      Author: teig
 */

#ifndef RFM69_GLOBALS_H_
#define RFM69_GLOBALS_H_

#if defined GLOBALS_H_
    // The below is also defined in user space, file "_globals.h"
#else
    // See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#some_xc_timer_macros
    typedef signed int time32_t; // signed int (=signed) or unsigned int (=unsigned) both ok, as long as they are monotoneously increasing
                                 // XC/XMOS 100 MHz increment every 10 ns for max 2exp32 = 4294967296,
                                 // ie. divide by 100 mill = 42.9.. seconds

    #define AFTER_32(a,b) ((a-b)>0)

    #ifdef DO_ASSERTS
        #define ASSERT_DELAY_32(d) do {if (d > INT_MAX) fail("Overflow");} while (0) // Needs <so646.h<, <limits.h> and <xassert.h>
        // INT_MAX is 2147483647 is what fits into 31 bits or last value before a signed 32 bits wraps around
    #else
        #define ASSERT_DELAY_32(d)
    #endif

    #define NOW_32(tmr,time) do {tmr :> time;} while(0) // A tick is 10ns
    // Programming XC on XMOS Devices (Douglas Watt)
    //     If the delay between the two input values fits in 31 bits, timerafter is guaranteed to behave correctly,
    //     otherwise it may behave incorrectly due to overlow or underflow. This means that a timer can be used to
    //     measure up to a total of 2exp31 / (100 mill) = 21s.

    #ifdef __XC__
        // BOOLEAN #include <stdbool.h> if C99
        // See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#bool
        typedef enum {false,true} bool; // 0,1 This typedef matches any integer-type type like long, int, unsigned, char, bool
    #endif
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_DEBUG_PRINT_GLOBAL
    #define DEBUG_PRINT_GLOBAL_LIB _USERMAKEFILE_LIB_RFM69_XC_DEBUG_PRINT_GLOBAL
                                   // 0: all printf off
                                   // 1: controlled locally in each xc file
#else
    #define DEBUG_PRINT_GLOBAL_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_I_RADIO_ANY
    #define I_RADIO_ANY_LIB _USERMAKEFILE_LIB_RFM69_XC_I_RADIO_ANY
    #if (I_RADIO_ANY_LIB==1)
        #warning I_RADIO_ANY_LIB not used
    #endif
#else
    #define I_RADIO_ANY_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_TRANS
    #define ALLOW_SESSION_TYPE_TRANS_LIB _USERMAKEFILE_LIB_RFM69_XC_TRANS

    #if (ALLOW_SESSION_TYPE_TRANS_LIB==1)
        #error ALLOW_SESSION_TYPE_TRANS_LIB==1 not used. Value 2 replaces 1 and TRANS_ASYNCH_WRAPPED_LIB=01
    #endif
#else
    #define ALLOW_SESSION_TYPE_TRANS_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_TRANS_ASYNCH_WRAPPED
    #define TRANS_ASYNCH_WRAPPED_LIB _USERMAKEFILE_LIB_RFM69_XC_TRANS_ASYNCH_WRAPPED
#else
    #define TRANS_ASYNCH_WRAPPED_LIB 0
#endif

#if (ALLOW_SESSION_TYPE_TRANS_LIB==0)
    #if (TRANS_ASYNCH_WRAPPED_LIB==1)
        #error Both must be 1. Set in Makefile
    #endif
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_GET_RADIO_DEBUG_REGS
    #define GET_RADIO_DEBUG_REGS_LIB _USERMAKEFILE_LIB_RFM69_XC_GET_RADIO_DEBUG_REGS
#else
    #define GET_RADIO_DEBUG_REGS_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_DEBUG_SHARED_LOG_VALUE
    #define DEBUG_SHARED_LOG_VALUE_LIB _USERMAKEFILE_LIB_RFM69_XC_DEBUG_SHARED_LOG_VALUE
#else
    #define DEBUG_SHARED_LOG_VALUE_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_DEBUG_FOREVER_LOOP
    #define DEBUG_FOREVER_LOOP_LIB _USERMAKEFILE_LIB_RFM69_XC_DEBUG_FOREVER_LOOP
#else
    #define DEBUG_FOREVER_LOOP_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_VALUE_XSCOPE
    #if (_USERMAKEFILE_LIB_RFM69_XC_VALUE_XSCOPE==1)
        #define VALUE_XSCOPE_LIB(name,value) do {xscope_core_int(name,value);} while(0) // or xscope_int
        #define DEBUG_XSCOPE_LIB 1
    #else
        #define VALUE_XSCOPE_LIB(name,value) // Empty
        #define DEBUG_XSCOPE_LIB 0
    #endif
#else
    #define VALUE_XSCOPE_LIB(name,value) // Empty
    #define DEBUG_XSCOPE_LIB 0
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_STRETCH_IRQ
    #define TEST_STRETCH_IRQ_LIB _USERMAKEFILE_LIB_RFM69_XC_STRETCH_IRQ // was NOT TEST_19MAR2019
#else
    #define TEST_STRETCH_IRQ_LIB 0
#endif


#if (DEBUG_SHARED_LOG_VALUE_LIB==1)

    extern unsigned g_radio_log_value;

    //  DEBUG_FOREVER_LOOP_LIB 14Feb2019 1842181E    0001 1000 0100 0010 0001 1000 0001 1110

    #define DO_RADIO_BITPOS_LEFT                                            9
    //                                       1111 1111 1111 1111 1111 1110 0000 0000
    #define DO_RADIO_LOG_VALUE_MASK_LEFT   0xfffffe00
    #define DO_RADIO_BITPOS_CENTER                                            4
    //                                       0000 0000 0000 0000 0000 0001 1111 0000
    #define DO_RADIO_LOG_VALUE_MASK_CENTER 0x000001f0
    //                                       0000 0000 0000 0000 0000 0000 0000 1111
    #define DO_RADIO_LOG_VALUE_MASK_RIGHT  0x0000000f

    // When CLIENT_WAIT_FOR_RADIO_MAX_MS==16 or lower. Higher: AQU=065. 24Feb2019
    //                                       1A802AE2
    //                                       0001 1010 1000 0000 0010 1010 1110 0010
    //                                          | | |  |           |  | |--14--
    //                                          | | |  |           |  | |  14=sendFrame
    //                                          | | |  |           |  | |
    //                                          28| |  |           |  | |  spiRead                    done
    //                                            27|  |           |  | |  spiWrite                   done
    //                                              25 |           |  | |  spiBurstWriteSeqRegsOrFIFO done
    //                                                 23          |  | |  func_encrypt16             started
    //                                                             13 | |  setHighPower               started
    //                                                                11|  setMode                    started
    //                                                                  9  while_readRegExpectedMask  started

    #define INIT_SHARED_GLOBAL(g_global,value) do {SET_SHARED_GLOBAL(g_global,value);} while(0)

    #define BITOR_ANYPOS_SHARED_GLOBAL(g_global,bitpos) do {bitor_anypos_g_radio_log_value(bitpos);} while(0)
    #define INC_SHARED_GLOBAL(g_global)                 do {inc_g_radio_log_value(); } while(0)
    #define INC_RIGHT_SHARED_GLOBAL(g_global)               inc_right_g_radio_log_value(DO_RADIO_LOG_VALUE_MASK_RIGHT) // returns unsigned
    #define SET_FIELD_SHARED_GLOBAL(g_global,number)    do {set_field_g_radio_log_value(number,DO_RADIO_BITPOS_CENTER,DO_RADIO_LOG_VALUE_MASK_CENTER);} while(0) // MAX number IS 31 SINCE 5 BITS

#else // empty defines
    #define INIT_SHARED_GLOBAL(g_global,value)
    #define BITOR_ANYPOS_SHARED_GLOBAL(g_global,bitpos)
    #define INC_SHARED_GLOBAL(g_global)
    #define INC_RIGHT_SHARED_GLOBAL(g_global)
    #define SET_FIELD_SHARED_GLOBAL(g_global,value)
#endif

#define ALLOW_INLINE_XMOS_ASM 1

typedef enum {low,high} pin_e; // must resolve to 0 and 1 (which they do)

typedef enum {
    logic_normal,  // must equal "false" since bool xor false keeps  bool
                   // In while_.. functions: loop until bit in flag becomes 1
    logic_inverted // must equal "true" since bool xor true  inverts bool
                   // In while_.. functions: loop until bit in flag becomes 0
} logic_e;

#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))

#define t_swap(type,a,b) {type t = a; a = b; b = t;}

#define NUM_ELEMENTS(array) (sizeof(array) / sizeof(array[0])) // Kernighan & Pike p22


// Standard union with an overlapping byte array just works. 4 bytes per unsigned (32 bits):
// Making no-trick memcpy possible and avoiding implicit bytes into unsigned conversions, like
// unsigned value = (((uint32_t) SPI_READ_REG(REG_FRFMSB) << 16) + ((uint16_t) SPI_READ_REG(REG_FRFMID) << 8) + SPI_READ_REG(REG_FRFLSB));
//
#define FOURBYTES_U_LEN 4
typedef struct {
    union {
        uint32_t value; // Overlaps these 8 bytes:
        uint8_t  bytes[FOURBYTES_U_LEN]; // printf ("%s", fourbytes.u.bytes) will print fourbytes.u.bytes[0] first
    } u;
    // ---- EXAMPLE ----
    // fourbytes_u fourbytes;
    // fourbytes.u.bytes[3] = 'A'; // 0x41
    // fourbytes.u.bytes[2] = 'B'; // 0x42
    // fourbytes.u.bytes[1] = 'C'; // 0x43
    // fourbytes.u.bytes[0] = '0'; // 0x30
    // fourbytes.u.bytes[0]++; // To '1'
    // debug_print ("Value=%d,%x, 3=%x, 2=%x, 1=%x, 0=%x s=%s\n",
    //     fourbytes.u.value,    // 1094861617
    //     fourbytes.u.value,    // 0x41424331
    //     fourbytes.u.bytes[3], // 0x41    'A'
    //     fourbytes.u.bytes[2], // 0x42   'B'
    //     fourbytes.u.bytes[1], // 0x43  'C'
    //     fourbytes.u.bytes[0], // 0x31 '1'
    //     fourbytes.u.bytes);   //      "1CBA"
    // PRINTS
    //     "Value=1094861617,41424331, 3=41, 2=42, 1=43, 0=31 s=1CBA"
} fourbytes_u;

#endif
