/*
 * rfm69_xc.h
 *
 *  Created on: 6. juni 2018
 *      Author: teig
 */

#ifndef RFM69_XC_H_
#define RFM69_XC_H_

#ifdef _USERMAKEFILE_LIB_RFM69_XC_SEMANTICS_DO_CRC_ERR_NO_IRQ
    #define SEMANTICS_DO_CRC_ERR_NO_IRQ _USERMAKEFILE_LIB_RFM69_XC_SEMANTICS_DO_CRC_ERR_NO_IRQ
#else
   #define SEMANTICS_DO_CRC_ERR_NO_IRQ 1 // 1 is RFM69. 0 allows messageRadioCRC16Err_IRQ
#endif

#ifdef _USERMAKEFILE_LIB_RFM69_XC_SEMANTICS_DO_LOOP_FOR_RF_IRQFLAGS2_PACKETSENT
    #define SEMANTICS_DO_LOOP_FOR_RF_IRQFLAGS2_PACKETSENT _USERMAKEFILE_LIB_RFM69_XC_SEMANTICS_DO_LOOP_FOR_RF_IRQFLAGS2_PACKETSENT
#else
    #define SEMANTICS_DO_LOOP_FOR_RF_IRQFLAGS2_PACKETSENT 0 // 1 is RFM69
                                                            // 0: saves time in send from max 369 to 280 ms. Mean from 251 to 183 ms
                                                            // 0: max 380 ms if RFM69-board unplugged while powered
#endif

#define ERROR_BITS_NONE 0x00000000
// Often tested for ERROR_BITS_NONE before next function, so the list below could probably have been a value list, not a bit array
typedef enum {
    //                                            //
    ERROR_NO_SPI_UNIT_CONNECTED             =  0, // getDeviceT            0x0001 Reading only zeros (startKIT) or ones (eXplorerKIT) from SPI!
    ERROR_BITNUM_DEVICE_TYPE                =  1, // getDeviceType         0x0002 Not RF_VERSION_VER (0x24 = 00100100)
    //                                            //                              Just as much to verify SPI even if it's bitwise symmetric
    //
    // 1. OBSERVE that all the tests causing any of the below errors MAY or MAY NOT happen if ERROR_NO_SPI_UNIT_CONNECTED or ERROR_BITNUM_DEVICE_TYPE
    //    BUT THESE ARE ONLY VALID WHEN SPI UNIT IS PLUGGED IN AND IT IS OF THE RIGHT TYPE
    // 2. OBSERVE that once the above errors did not happen, then some of the errors below may happen INTERMITTENTLY
    //    so cannot really be used to "give up". Just try sending "next time"
    //
    ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_1   =  2, // initialize (1)        0x0004 X
    ERROR_BITNUM_REG_SYNCVALUE1_AA          =  3, // initialize (2)        0x0008 X
    ERROR_BITNUM_REG_SYNCVALUE1_55          =  4, // initialize (3)        0x0010 X
    ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_3   =  5, // setMode (1)           0x0020
    ERROR_BITNUM_NO_RF69_MODE_SLEEP_EXIT    =  6, // setMode (2)           0x0040
    ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_2   =  7, // sendFrame             0x0080
    ERROR_BITNUM_WAIT_FOR_CSMA_LIMIT_SET_MS =  8, // send                  0x0100 Saw this when another sender was active!
                                                  //                              Also if proper init not done
    ERROR_BITNUM_RF_OSC1_RCCAL_DONE         =  9, // rcCalibration         0x0200
    ERROR_BITNUM_RF_TEMP1_MEAS_RUNNING      = 10, // readTemperature_degC  0x0400
    ERROR_BITNUM_REG_RSSICONFIG             = 11, // readRSSI_dBm          0x0800
    ERROR_BITNUM_SETMODE_FAILED             = 12, // setMode               0x1000

    ERROR_BITNUM_RF_IRQFLAGS2_MODEREADY     = 13, // send                  0x2000
    ERROR_BITNUM_RF_IRQFLAGS2_FIFONOTEMPTY  = 14, // IRQ                   0x4000 Observe that getting this _sometimes_ causes the IRQ error, especially if _USERMAKEFILE_LIB_RFM69_XC_SEMANTICS_DO_CRC_ERR_NO_IRQ=0
    // Observe print_error_bits prints only 4 hex chars (15 above max)
} error_bits_e;

typedef enum {NO_ERR, IS_ERR} is_error_e;

// Read around figure 34 and page 52 of the SEMTECH SX1231H Transceiver manual (Rev 1 - Oct 2011)

// Observe that register REG_PACKETCONFIG1 RegPacketConfig1 bit PacketFormat is set to
// is set to this RF_PACKET1_FORMAT_VARIABLE Variable Length Packet Format in Initialize.
// In that case:
//     In this mode the length of the payload, indicated by the length byte, is given by the
//     first byte of the FIFO and is limited to 255 bytes if AES is not enabled else the message
//     is limited to 64 bytes (i.e. max 66 bytes payload if Address byte is enabled). Note that
//     the length byte itself is not included in its calculation. In this mode, the payload must
//     contain at least 2 bytes, i.e. length + address or message byte.
//
//#define RF69_MAX_DATA_LEN_61 61 // To take advantage of the built in AES/CRC we want to limit the frame size to the internal FIFO size
                                  // So max message length (limited to 61 to support AES hardware encryption
#define   RF69_MAX_FIFO_LEN_65 65 // Not including the first byte PACKETLEN
#define   RF69_MAX_FIFO_LEN_66 66

// ----------------------------------------------------------------------------------
// RFM69 library packet structure: https://lowpowerlab.com/2013/06/20/rfm69-library/
//
// "RFM69 modules support packets 255bytes long, or even unlimited, however since AES
//  hardware encryption must have all 64 encryptable bytes before encryption engine
//  can do its work, this library was limited to 65 bytes total full_payload_01, including a
//  4 header bytes. The full_payload_01 length is automatically controlled by hardware and
//  is not encrypted. Only the DestID, SenderID and CTL bytes are encrypted.
//  Hence 61 other encryptable bytes remain for data."
// -----------------------------------------------------------------------------------

// SEMTECH 5.5.2.2. Variable Length Packet Format
// An illustration of a variable length packet is shown below. It contains the following fields:
//     Preamble (1010...)
//     Sync word (Network ID)
//     Length byte:                      Teig: I think this is from first byte to/from FIFO
//     Optional Address byte (Node ID)   Teig: I think is only set after after setNODEID call
//     Message data
//     Optional 2-bytes CRC checksum     Teig:

typedef struct {
                                        // --- 3 BYTES default not in FIFO, not even in Initialize: ---
    uint8_t preamble0;                  // (*) 0xAA Used by Bit Synchronizer, 24 bits used here:
    uint8_t preamble1;                  // (*) 0xAA
    uint8_t preamble2;                  // (*) 0xAA
                                        // --  2 BYTES done in Initialize, not in FIFO: ---------------
    uint8_t SYN0;                       // (*) RFM69_SYNCVALUE1
    uint8_t SYN1;                       // (*) RFM69_SYNCVALUE2
                                        //     Peer-to-peer network header, 4 bytes:
                                        // --- 1 BYTE FIFO num BYTES after this: ---------------------
    uint8_t header0;                    //     PACKETLEN when RF_PACKET1_FORMAT_VARIABLE it's
                                        //     sizeof(packet_t) + RFM69HEADER_LEN08_IS4
                                        // --- 64 BYTES for FIFO, CALLED PAYLOAD: ---------------------
    uint8_t header1;                    //     TARGETID (DestID)
    uint8_t header2;                    //     SENDERID
    uint8_t header3;                    //     was CTLBYTE
    uint8_t PACKET[APPFULLPAYLOAD_LEN]; //     61 encryptable bytes remain for data
                                        // --- 2 BYTES, not in FIFO: ----------------------------------
    uint8_t CRC16_LSB;                  // (*) CRC-16 LSB
    uint8_t CRC16_MSB;                  // (*) CRC-16 MSB
    //
    // (*) Automtaic handling in hardware blocks: hardware preamble, synch recognition and CRC check
    // From Semtech data sheet
    //     3.5.13. Bit Synchronizer
    //        A preamble (0x55 or 0xAA) of at least 12 bits is required for synchronization,
    //        the longer the synchronization the better the packet success rate
    //     Packet engine with CRC, AES-128 and 66-byte FIFO
    //     5.1.2. Data Operation Modes
    //     1. Continuous mode: ...
    //     2. Packet mode (recommended): user only provides/retrieves full_payload_01 bytes to/from the FIFO.
    //        The packet is automatically built with preamble, Sync word, and optional AES, CRC, and
    //        DC-free encoding schemes The reverse operation is performed in reception. The uC processing
    //        overhead is hence significantly reduced compared to Continuous mode. Depending on the
    //        optional features activated (CRC, AES, etc) the maximum full_payload_01 length is limited to
    //        FIFO size, 255 bytes or unlimited.
} rfm69_packet_t; // This struct is not used, only shown for understanding

typedef enum {
    no_IRQExpected,
    messagePacketSent_IRQExpected,
    messagePacketNotSent_ErrorIRQNotExpected // Like if RFM69-board unplugged while powered
    //
} waitForIRQInterruptCause_e;

typedef enum {                           // Statistics from log file "2018 04 27..Commit 206a5a0.txt"
    void_IRQ                        =  0, // Internal only
    messagePacketSentOk_IRQ         =  1, // When only TX and no RX handling, always
    messagePacketSentOk2_IRQ        =  2, // This is only seen if only TX, or if board not connected
    messagePacketSentOkMissing_IRQ  =  3, // Not seen
    messagePacketSentOkMissing2_IRQ =  4, // Not seen
    messageReceivedOk_IRQ           =  5, // All the time when RX
    messageNotExpectedBy_IRQ        =  6, // Not seen
    messagePacketLenErr_IRQ         =  7, // 196 times with 870K messages @ 1 per second
                                          // Probably gatekeeper for messageAppCRC32Err_IRQ
    messageNotForThisNode_IRQ       =  8, // Not seen
    messageRadioCRC16Err_IRQ        =  9, // Requires SEMANTICS_DO_CRC_ERR_NO_IRQ==0
                                          // If this IRQ never delivered it's a gatekeeper for messageAppCRC32Err_IRQ
    messageAppCRC32Err_IRQ          = 10  // Happens! Search for "NEEDED"
    //
} interruptAndParsingResult_e;

/* typedef enum {
    RF69_315MHZ, // Was 31 as define in RFM69, but it's used as en enum anyhow
    RF69_433MHZ, // Was 43
    RF69_868MHZ, // Was 86
    RF69_915MHZ  // Was 91
    //
} freqBand_e; */

#define KEY_LEN 16
typedef struct {
    uint8_t     nodeID;
    uint32_t    RegFrf;       // REG FREQUENCY (mult with RF69_FSTEP_FLOAT32 to get frequency in Hz)
    uint8_t     key[KEY_LEN]; // KEY only, no nul-terminator
    bool        isRFM69HW;    // IS_RFM69HW_HCW
    //
} rfm69_params_t;

typedef struct {
    uint8_t      APPPROTLEN;
    uint8_t      PACKETLEN;
    uint8_t      SENDERID;
    uint8_t      TARGETID; // should match NODEID
    error_bits_e error_bits;
    uint8_t      RegIrqFlags2;
    //
} some_rfm69_internals_t;

// USE OF SPI_AUX (p_spi_aux) PORT BIT0s
//
// PIN  9 RST   Adafruit RFM68HCW. 100K pull-up to VIN (3V3 with me).
//              This is the Reset pin for the radio. It's pulled high by default.
//              Pull down to ground to put it into reset
//              Via a 74HC4050D (Hex non-inverting HIGH-to-LOW level shifter) to:
// PIN 14 RESET on the HOPERF RFM68HCW board, then it's connected to (PROBABLY VIA AN INVERTER (*)):
// PIN  6 RESET Semtech SX1231H. Semtech writes:
//              7.2.2. Manual Reset
//                  A manual reset of the SX1231H is possible even for applications in which VDD cannot be physically
//                  disconnected. Pin 6 should be pulled high for a hundred microseconds, and then released.
//                  The user should then wait for 5 ms before using the chip.
//
// (*) I have not been able to find the diagram of the HOPERF RFM68HCW board, not even at
//     http://www.hoperf.com/rf_transceiver/modules/RFM69HCW.html
//
// Observe that I have no control of the ports during xTIMEcomposer downloading
// I have observed a 700-800 ms low on signal pins before my code starts
//
#define SPI_AUX0_RST_TIME_A_MS 10 // See comment above
#define SPI_AUX0_RST_TIME_B_MS 10 // 100 us "PULSE" plus time to restart (even if only 5 ms is required for restart finished)

typedef enum {
    FORCETRIGGER_OFF,
    FORCETRIGGER_ON
} forceTrigger_t;

typedef enum {
    RX_TX_IRQ,
    RX_IRQ
} handleIRQ_t;

// For Norway, Norwegian rules:
// See http://www.teigfam.net/oyvind/home/technology/164-my-aquariums-data-radioed-through-the-shelf/#Ref17
// Frekvensbndet 433,050434,790 MHz tillates brukt med en maksimal tillatt utstrlt effekt p 10 mW e.r.p.
// Maksimal sendetid er mindre enn 10 prosent.

// ALLOWED IN NORWAY: 433.050  434.790 (1.74 MHz band)

// Observe RF69_FSTEP_FLOAT32 = 61.03515625 -> 61.035156

// RegFrf (MHz) - carrier frequency
//
#define RF_FRF_314           0x004E8000
#define RF_FRF_315           0x004EC000
#define RF_FRF_316           0x004F0000
#define RF_FRF_433_433000000 0x006C4000 // Not allowed in Norway
#define RF_FRF_433_433919921 0x006C7AE0 // 433.92 AVOID IT!
                                        // Center of legal band (in Norway). Collides with my car key and light remote control
                                        // See http://www.teigfam.net/oyvind/home/technology/164-my-aquariums-data-radioed-through-the-shelf/#frequency_crash_with_my_car_key
#define RF_FRF_433_433705993 0x006C6D2F // Should not hit car key. 433.92-8ppm-(1/10 of band=1.74/10).
                                        // Frequency is according to calculator 0x006C6D2F * (RF69_FSTEP_FLOAT32 as 61.03515625)
                                        // 61.03515625 used by RFM69 since it has a phase locked loop

#define RF_FRF_434           0x006C8000
#define RF_FRF_435           0x006CC000
#define RF_FRF_863           0x00D7C000
#define RF_FRF_864           0x00D80000
#define RF_FRF_865           0x00D84000
#define RF_FRF_866           0x00D88000
#define RF_FRF_867           0x00D8C000
#define RF_FRF_868           0x00D90000
#define RF_FRF_869           0x00D94000
#define RF_FRF_870           0x00D98000
#define RF_FRF_902           0x00E18000
#define RF_FRF_903           0x00E1C000
#define RF_FRF_904           0x00E20000
#define RF_FRF_905           0x00E24000
#define RF_FRF_906           0x00E28000
#define RF_FRF_907           0x00E2C000
#define RF_FRF_908           0x00E30000
#define RF_FRF_909           0x00E34000
#define RF_FRF_910           0x00E38000
#define RF_FRF_911           0x00E3C000
#define RF_FRF_912           0x00E4C000
#define RF_FRF_913           0x00E44000
#define RF_FRF_914           0x00E48000
#define RF_FRF_915           0x00E4C000  // Default 915000000 / 0xE4C000 (14991360) = 61,03515625, as in RFM69 phase locked loop
#define RF_FRF_916           0x00E50000
#define RF_FRF_917           0x00E54000
#define RF_FRF_918           0x00E58000
#define RF_FRF_919           0x00E5C000
#define RF_FRF_920           0x00E60000
#define RF_FRF_921           0x00E64000
#define RF_FRF_922           0x00E68000
#define RF_FRF_923           0x00E6C000
#define RF_FRF_924           0x00E70000
#define RF_FRF_925           0x00E74000
#define RF_FRF_926           0x00E78000
#define RF_FRF_927           0x00E7C000
#define RF_FRF_928           0x00E80000
//
#define RF69_INIT_FREQUENCY_433919921_HZ 433919921L // Value fits HEX, shown with DEBUG_PRINT_RFM69_FREQ. 433.92 is in the center, same as our car key!
                                                    // See http://www.teigfam.net/oyvind/home/technology/164-my-aquariums-data-radioed-through-the-shelf/#frequency_crash_with_my_car_key
#define RF69_INIT_FREQUENCY_433705993_HZ 433705993L // Should not hit car key. 433.92-8ppm-(1/10 of band=1.74/10)

// Control readAllRegs
#ifdef _USERMAKEFILE_LIB_RFM69_XC_RADIO_IF_READALLREGS
    // ROBERTXMOS See http://www.xcore.com/viewtopic.php?f=26&t=6718
    // 0: Code size smaller
    // 1: Code size larger
    #define RADIO_IF_READALLREGS _USERMAKEFILE_LIB_RFM69_XC_RADIO_IF_READALLREGS
#else
    #define RADIO_IF_READALLREGS 0
#endif

//       If called would cause the linker to kick in the floating point library:
unsigned freq_register_value_to_Hz (const uint32_t register_value); // Approximate
uint32_t freq_Hz_to_register_value (const unsigned frequency_Hz);   // Approximate

#if (GET_RADIO_DEBUG_REGS_LIB==1)
    #define NUM_DEBUG_BYTES 5

    typedef enum  {
        iof_RegOpMode = 0,
        iof_RegIrqFlags1 = 1,
        iof_RegIrqFlags2 = 2,
        iof_radio_mode = 3,
        iof_waitForIRQInterruptCause = 4
    } iof_debug_bytes_e;

    typedef struct {
        uint8_t data[NUM_DEBUG_BYTES];
    } debug_data_t;
#endif

typedef enum {
    debug_none,
    debug_just_read_some_registers,
    debug_flush_RegIrqFlags2,
    debug_mode_0_1, // THIS SOLVED THE IRQ PROBLEM!
    debug_void      // At the end of enum. Only used for incrementing and modulo arithemtics
} debug_state_e;

typedef int16_t dBm_t;

// For those interface functions that start with _uspi
typedef enum {
    session_none_void_trans,                            // VALUE_XSCOPE_LIB 65  90
    session_do_aux_adafruit_rfm69hcw_RST_pulse_trans1,  // VALUE_XSCOPE_LIB 66  91
    session_send_trans1,                                // VALUE_XSCOPE_LIB 67  92
    session_readRSSI_dBm_trans1,                        // VALUE_XSCOPE_LIB 68  93
    session_handleSPIInterrupt_trans1,                  // VALUE_XSCOPE_LIB 69  94
    session_ultimateIRQclear_trans1,                    // VALUE_XSCOPE_LIB 70  95
    session_receiveDone_trans1,                         // VALUE_XSCOPE_LIB 71  96
    session_initialize_trans1,                          // VALUE_XSCOPE_LIB 72  97
    session_encrypt16_trans1,                           // VALUE_XSCOPE_LIB 73  98
    session_getDeviceType_trans1,                       // VALUE_XSCOPE_LIB 74  99
    session_rcCalibration_trans1,                       // VALUE_XSCOPE_LIB 75 100
    session_readTemperature_degC_trans1,                // VALUE_XSCOPE_LIB 76 101
    session_setHighPower_trans1,                        // VALUE_XSCOPE_LIB 77 102
    session_setMode_trans1,                             // VALUE_XSCOPE_LIB 78 103
    session_setPowerLevel_dBm_trans1,                   // VALUE_XSCOPE_LIB 79 104
    session_setFrequencyRegister,                       // VALUE_XSCOPE_LIB 80 105
    session_getFrequencyRegister                        // VALUE_XSCOPE_LIB 81 106
    #if (GET_RADIO_DEBUG_REGS_LIB==1)
        ,
        session_getDebug                                // VALUE_XSCOPE_LIB 82 107
    #endif
} session_transx_id_e;

#define WAIT_FOR_RADIO_MAX_MS 200

#ifndef CLIENT_WAIT_FOR_RADIO_MAX_MS
    #define CLIENT_WAIT_FOR_RADIO_MAX_MS (WAIT_FOR_RADIO_MAX_MS * 3) // Good margin, we surveil it with maxtime_used_us_trans1to2
                                                                     // Also, remember that context switches with printf might happen in between!
#else                                                                // Tested with DEBUG_FOREVER_LOOP_LIB
    #warning using externally defined
#endif

typedef struct {
    error_bits_e error_bits;
    is_error_e   is_error;
} error_t;

#if (ALLOW_SESSION_TYPE_TRANS_LIB>=1)
    typedef struct {
        some_rfm69_internals_t      return_some_rfm69_internals;
        packet_t                    return_PACKET;
        interruptAndParsingResult_e return_interruptAndParsingResult;
    } from_handleSPIInterrupt_t;

    typedef struct {
        time32_t   start_time_trans1;
        unsigned   maxtime_allowed_ms_trans1to2; // CLIENT_WAIT_FOR_RADIO_MAX_MS. Long, like 600 ms. Will include context switches and other work!
        unsigned   maxtime_used_us_trans1to2;    // Short, even like 820 us
        bool       timed_out_trans1to2;          // maxtime_allowed_ms_trans1to2 too short or something is hanging
        unsigned   radio_log_value;              // Only meaningful if DEBUG_SHARED_LOG_VALUE_LIB==1
    } timing_transx_t;

    // Only for those interface functions that start with _uspi
    typedef struct {
        session_transx_id_e id_trans1;
        unsigned            test_inc_1000;
        union {
            waitForIRQInterruptCause_e waitForIRQInterruptCause; // From send_trans3
            dBm_t                      rssi_dBm;                 // From readRSSI_dBm
            from_handleSPIInterrupt_t  handleSPIInterrupt;       // From handleSPIInterrupt
            bool                       receiveDone;              // From receiveDone
            uint8_t                    deviceType;               // From getDeviceType
            int8_t                     temperature_degC;         // From readTemperature_degC
            uint32_t                   register_value;           // From getFrequencyRegister
            #if (GET_RADIO_DEBUG_REGS_LIB==1)
                debug_data_t           debug_data;               // From getDebug
            #endif
        } u_out;
    } return_trans3_t; // TODO: move inside
#endif

typedef interface radio_if_t { // All functions may set error_bits, so getAndClearErrorBits must be called after each logical session
    // Ref RFM69 library: Become internal functions
    //                   receiveBegin
    //                   sendFrame
    //                   setHighPowerRegs
    // Ref RFM69 library: Dropped/deleted/obsoleted functions
    //                   interruptHandler       and also isr0 >= handled by event on port p_spi_irq instead
    //                   maybeInterrupts        an inline
    //                   sendWithRetry          easier to just do an extra call to send
    //                   sendACK                all about..
    //                   ACKRequested           .. ACKing..
    //                   ACKReceived            ..removed
    //                   canSend                ..removed as interface call
    //                   setNetwork             It was REG_SYNCVALUE2 that got ist value as "NETWORKID". Now just use RFM69_SYNCVALUE2
    //                   sleep                  ..removed. Use setMode with RF69_MODE_SLEEP instead (saves about 200 bytes, 18Feb2019)
    //
    // These interface functions do not use SPI and do not need any timeout monitoring.
    // The safest way to use them is via the functions that end with _iff, where any real interface call is prohibited once a timoeut has been detected on any of the SPI-monitored functions
    void                       do_spi_aux_pin        (const unsigned maskof_pin, const pin_e value);
    {error_bits_e, is_error_e} getAndClearErrorBits  (void);
    error_t                    getAndClearErrorBits_ (void);
    void                       setListenToAll        (const bool doListenToAll); // Was promiscuous
    uint8_t                    setNODEID             (const uint8_t newNODEID);  // return_previous_NODEID

    // These interface functions use SPI and have an optional layer for timeout monitoring. They end with _trans1 (not here, below)
    // The safest way to use them is via the functions that end with _iff_trans, where any real interface call is prohibited once a timoeut has been detected

    #if (ALLOW_SESSION_TYPE_TRANS_LIB==0)
        void                                                            uspi_do_aux_adafruit_rfm69hcw_RST_pulse (const unsigned maskof_pin); // MUST BE CALLED FIRST AND ONLY ONCE, EVEN BEFORE initialize. Reason: it blocks on await_spi_port_init_by_all_clients
        waitForIRQInterruptCause_e                                      uspi_send                               (const uint8_t  TARGETID_toAddress,  const packet_t PACKET);
        uint8_t                                                         uspi_getDeviceType                      (void);
        dBm_t                                                           uspi_readRSSI_dBm                       (const forceTrigger_t forceTrigger); // Received Signal Strength Indication
        {some_rfm69_internals_t, packet_t, interruptAndParsingResult_e} uspi_handleSPIInterrupt                 (void); // packet_t also returns CommHeaderRFM69
        void                                                            uspi_setPowerLevel_dBm                  (const uint8_t powerLevel_dBm);
        void                                                            uspi_encrypt16                          (const char key[], unsigned const static len); // 15 chars plus nul-terminator is 16 bytes
        void                                                            uspi_initialize                         (const rfm69_params_t init);
        void                                                            uspi_setHighPower                       (const bool isHighPowerOn); // For RFM69HW only: you must call setHighPower(true) after initialize() or else transmission won't work
        bool                                                            uspi_receiveDone                        (void);
    #endif
    void                                                                uspi_ultimateIRQclear                   (void);
    void                                                                uspi_rcCalibration                      (void);
    int8_t                                                              uspi_readTemperature_degC               (const int8_t calOffset_degC);
    void                                                                uspi_setMode                            (const uint8_t newMode);
    void                                                                uspi_setFrequencyRegister               (const uint32_t register_value);
    uint32_t                                                            uspi_getFrequencyRegister               (void); // returns register_value from RFM69
    #if (GET_RADIO_DEBUG_REGS_LIB==1)
        debug_data_t                                                    uspi_getDebug                           (const debug_state_e debug_state); // data indexed with iof_debug_bytes_e
    #endif

    #if (RADIO_IF_READALLREGS == 1)
        void                                                            uspi_readAllRegs                        (void); // Service function
    #endif

    #if (ALLOW_SESSION_TYPE_TRANS_LIB>=1)

        // _trans1 do_start_job            Phase 1, then follows
        // _trans2 job_done_notify         phase 2, then follows
        // _trans3 notify_clear_get_result phase 3

        // COST: one chanend
        // FIRST DO ANY _trans1. Then by contract do NO OTHER uspi_.. calls before session_trans2 is received, then you MUST do session_trans3. Then all ready for anything.
        //
        //                               ---- Only for those interface functions that start with _uspi ----
                                time32_t do_aux_adafruit_rfm69hcw_RST_pulse_trans1 (const unsigned maskof_pin);
                                time32_t                         initialize_trans1 (const rfm69_params_t init);
                                time32_t                      getDeviceType_trans1 (void);
                                time32_t                          encrypt16_trans1 (const char key[KEY_LEN]); // len not possible here
                                time32_t                               send_trans1 (const uint8_t TARGETID_toAddress, const packet_t PACKET);
                                time32_t                       readRSSI_dBm_trans1 (const forceTrigger_t forceTrigger);
                                time32_t                 handleSPIInterrupt_trans1 (void);
                                time32_t                   ultimateIRQclear_trans1 (void);
                                time32_t                        receiveDone_trans1 (void);
                                time32_t                      rcCalibration_trans1 (void);
                                time32_t               readTemperature_degC_trans1 (const int8_t calOffset_degC);
                                time32_t                       setHighPower_trans1 (const bool isHighPowerOn);
                                time32_t                            setMode_trans1 (const uint8_t newMode);
                                time32_t                  setPowerLevel_dBm_trans1 (const uint8_t powerLevel_dBm);
                                time32_t               setFrequencyRegister_trans1 (const uint32_t register_value);
                                time32_t               getFrequencyRegister_trans1 (void);
        #if (GET_RADIO_DEBUG_REGS_LIB==1)
                                time32_t                           getDebug_trans1 (const debug_state_e debug_state); // data indexed with iof_debug_bytes_e
        #endif
        [[notification]]        slave void                          session_trans2 (void);  // Also handled in do_sessions_trans2to3
        [[clears_notification]] return_trans3_t                     session_trans3 (void);  // --"--
    #endif
} radio_if_t;

typedef struct {
    pin_e pin_value;
    bool  pin_delayed_high_event_next; // If this is true, the below is false
    bool  pin_was_high_too_long;       // Is false if above is true
} irq_val_t;

typedef enum {
    pin_gone_low,          // "gone" high to indicate edge triggering
    pin_gone_high,         // "gone" low  ditto
    pin_still_high_timeout // Will be repeated until pin_gone_low
} irq_update_e;

typedef struct probe_pins_t {
    out port probe_when_irq;
} probe_pins_t;

// [[distributable]] // If it only responds to interface calls (but in memory usage it actually costs a _little_ more)
   [[combinable]]    // Allows a timerafter
// ....                 Nothing specified would be liek any occam task with code allowed after the while-select main loop.
//                      With [[combinable]] we simulate this code in timerafter
void RFM69_driver (
        server radio_if_t    i_radio,
               out port      p_spi_aux,
        client spi_master_if i_spi,
               unsigned      spi_client
);

#if (DEBUG_SHARED_LOG_VALUE_LIB==1)
    void clr_radio_log_value (void);
#endif

#if (ALLOW_SESSION_TYPE_TRANS_LIB>=1)

    // ---- For those interface functions that DO NOT start with _uspi ----
    void                             do_spi_aux_pin_iff        (const bool timed_out_trans1to2, client radio_if_t i_radio, const unsigned maskof_pin, const pin_e value);
    {error_bits_e, is_error_e} getAndClearErrorBits_iff        (const bool timed_out_trans1to2, client radio_if_t i_radio);
    void                             setListenToAll_iff        (const bool timed_out_trans1to2, client radio_if_t i_radio, const bool doListenToAll); // Was promiscuous
    uint8_t                               setNODEID_iff        (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t newNODEID);

    // ---- For those interface functions that start with _uspi ----
    #if (TRANS_ASYNCH_WRAPPED_LIB==0)
        time32_t do_aux_adafruit_rfm69hcw_RST_pulse_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const unsigned maskof_pin); // MUST BE CALLED FIRST AND ONLY ONCE, EVEN BEFORE initialize. Reason: it blocks on await_spi_port_init_by_all_clients
        time32_t                         initialize_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const rfm69_params_t init);
        time32_t                      getDeviceType_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio);
        time32_t                               send_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t TARGETID_toAddress, const packet_t PACKET);
        time32_t                       readRSSI_dBm_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const forceTrigger_t forceTrigger);
        time32_t                 handleSPIInterrupt_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio);
        time32_t                        receiveDone_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio);
        time32_t                   ultimateIRQclear_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio);
        time32_t                  setPowerLevel_dBm_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t powerLevel_dBm);
        time32_t                          encrypt16_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const char key[KEY_LEN]); // len not possible here
        time32_t                       setHighPower_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const bool isHighPowerOn);
    #endif

    time32_t                   readTemperature_degC_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const int8_t calOffset_degC);
    time32_t                                setMode_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t newMode);
    time32_t                   setFrequencyRegister_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint32_t register_value);
    time32_t                   getFrequencyRegister_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio);
    #if (GET_RADIO_DEBUG_REGS_LIB==1)
        time32_t                           getDebug_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const debug_state_e debug_state); // data indexed with iof_debug_bytes_e
    #endif

    #if (ALLOW_SESSION_TYPE_TRANS_LIB==2)
        void        do_aux_adafruit_rfm69hcw_RST_pulse_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const unsigned maskof_pin); // MUST BE CALLED FIRST AND ONLY ONCE, EVEN BEFORE initialize. Reason: it blocks on await_spi_port_init_by_all_clients
        void                                initialize_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const rfm69_params_t init);
        uint8_t                          getDeviceType_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans);
        waitForIRQInterruptCause_e                send_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const uint8_t TARGETID_toAddress, const packet_t PACKET);
        dBm_t                             readRSSI_dBm_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const forceTrigger_t forceTrigger); // Received Signal Strength Indication
        interruptAndParsingResult_e handleSPIInterrupt_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, some_rfm69_internals_t &return_some_rfm69_internals, packet_t &return_PACKET);
        bool                               receiveDone_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans);
        void                          ultimateIRQclear_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans);
        void                         setPowerLevel_dBm_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const uint8_t powerLevel_dBm);
        void                                 encrypt16_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const char key[KEY_LEN]); // len not possible here
        void                              setHighPower_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const bool isHighPowerOn); // For RFM69HW only: you must call setHighPower(true) after initialize() or else transmission won't work
        bool                               receiveDone_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans);
        #if (GET_RADIO_DEBUG_REGS_LIB==1)
            debug_data_t                      getDebug_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const debug_state_e debug_state); // data indexed with iof_debug_bytes_e
        #endif

    #endif

    // Contains a nested select: a task calling this cannot be [[combinable]] (and then also not [[distributable]])
    // This the caller has to have a core on its own
    //
    void do_sessions_trans2to3 ( // Returns g_radio_log_value. 0 if DEBUG_SHARED_LOG_VALUE_LIB==0
        client radio_if_t       i_radio,
               timing_transx_t  &session_trans,
               return_trans3_t  &return_trans3);
#endif

[[combinable]] // Cannot be [[distributable]] since pinsneq
void IRQ_interrupt_task (
                chanend       ?c_irq_update,        // Sends irq_update_e
                in port        p_irq,
                probe_pins_t &?p_probe,             // "null" in call to omit
         const unsigned        irq_high_max_time_ms
        );

#endif
