/*
 * rfm69_crc.xc
 *
 *  Created on: 6. juni 2018
 *      Author: teig
 */

#include <stdint.h>   // uint8_t

#include "rfm69_globals.h"
#include "rfm69_crc.h"

// Found at http://www.xcore.com/viewtopic.php?f=26&t=5048&p=25852&hilit=CRC#p25852

crc32_t
calc_CRC32 (
        uint32_t data[], static const int num_words,
        crc32_t  expected_crc)
{
    crc32_t crc      = 0xFFFFFFFF;
    uint32_t polynom = 0xEDB88320; // Reversed polynomial for CRC-32, see https://en.wikipedia.org/wiki/Cyclic_redundancy_check

    #if (ALLOW_INLINE_XMOS_ASM == 1)

        // From the Assembly Programming Manual (X9432B):
        //    crc32     d, r, p     Word CRC
        //              d           register used as a destination operand
        //                 r        register used as a resource identifier
        //                    p     An optional predicate symbol may be specified as the 3rd argument
        //    crc32_inc d, e, x, y, bitp Word CRC with address increment (xCORE-200 only)
        //    crc8      r, o, d, p  8-step CRC
        //    crcn      d, x, p, n  Variable step CRC (xCORE-200 only)

        for (unsigned int i = 0; i < num_words; i++)
        {
            asm volatile("crc32 %0, %2, %3" : "=r" (crc) : "0" (crc), "r"  (data[i]), "r" (polynom));
        }

        asm volatile("crc32 %0, %2, %3" : "=r" (crc) : "0" (crc), "r" (expected_crc), "r" (polynom));
    #else
        #error Missing C code for CRC
        // Steal it from http://www.hackersdelight.org/hdcodetxt/crc.c.txt
    #endif

    return ~crc;
}

