/*
 * RFM69_xc.xc
 *
 *  Created on: 21. feb. 2018Mask
 *      Author: teig
 */

// **********************************************************************************************
// CODE INSPIRED MOSTLY BY THE RFM69 library
// https://github.com/LowPowerLab/RFM69
// **********************************************************************************************
//     ******************************************************************************************
//     Driver definition for HopeRF RFM69W/RFM69HW/RFM69CW/RFM69HCW, Semtech SX1231/1231H
//     ******************************************************************************************
//     Copyright Felix Rusu 2016, http://www.LowPowerLab.com/contact
//     ******************************************************************************************
//     License
//     ******************************************************************************************
//     This program is free software; you can redistribute it
//     and/or modify it under the terms of the GNU General
//     Public License as published by the Free Software
//     Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will
//     be useful, but WITHOUT ANY WARRANTY; without even the
//     implied warranty of MERCHANTABILITY or FITNESS FOR A
//     PARTICULAR PURPOSE. See the GNU General Public
//     License for more details.
//
//     Licence can be viewed at
//     http://www.gnu.org/licenses/gpl-3.0.txt
//
//     Please maintain this license information along with authorship
//     and copyright notices in any redistribution of this code

// **********************************************************************************************
// CODE INSPIRED ALSO BY THE RadioHead library
// http://www.airspayce.com/mikem/arduino/RadioHead/
// **********************************************************************************************
//     RH_RF69.cpp etc.
//     Copyright (C) 2011 Mike McCauley

// **********************************************************************************************
// PORTED TO XC FOR XMOS PROCESSORS BY YVIND TEIG
// http://www.teigfam.net/oyvind/home/technology/164-my-aquariums-data-radioed-through-the-shelf
// Porting has been done as needed, not to keep the SW as more like a general driver.
// So functions there have not been ported here
// **********************************************************************************************

#include <xs1.h>
#include <platform.h>  // slice
#include <timer.h>     // delay_milliseconds(200), XS1_TIMER_HZ etc
#include <stdint.h>    // uint8_t
#include <stdio.h>     // debug_print
#include <xccompat.h>  // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
#include <iso646.h>    // not etc.
#include <spi.h>
#include <xscope.h>

#include "rfm69_globals.h"
#include "rfm69_spi_driver.h"

#include "rfm69_crc.h"
#include "rfm69_commprot.h"
#include "rfm69_registers.h"
#include "module_xc_ptr.h" // When DEBUG_SHARED_LOG_VALUE_LIB==1
#include "rfm69_xc.h"

#define DEBUG_PRINT_RFM69_HIGHPRI 1
#define debug_print_highpri(fmt, ...) do { if((DEBUG_PRINT_RFM69_HIGHPRI==1) and (DEBUG_PRINT_GLOBAL_LIB==1)) printf(fmt, __VA_ARGS__); } while (0)

#define DEBUG_PRINT_RFM69 0
#define debug_print(fmt, ...) do { if((DEBUG_PRINT_RFM69==1) and (DEBUG_PRINT_GLOBAL_LIB==1)) printf(fmt, __VA_ARGS__); } while (0)

#ifdef _USERMAKEFILE_LIB_RFM69_XC_DEBUG_PRINT_EXCEPTION
    #define debug_print_x(fmt, ...) do { if(_USERMAKEFILE_LIB_RFM69_XC_DEBUG_PRINT_EXCEPTION) printf(fmt, __VA_ARGS__); } while (0)
#else
    #define debug_print_x(fmt, ...)
#endif

void print_add_error_bits_iff (
        const uint8_t      text[],
        const error_bits_e error_bits) {

    if (error_bits != ERROR_BITS_NONE) {
        debug_print("\n__ %s error_bits(%04X)\n\n", text, error_bits);
        // Always print. "__ " means at the end of a function but no corresponding start,
        // as compared to, like "setHighPowerRegs" and "-- setHighPowerRegs"
    } else {
        debug_print("__ %s\n", text);
    }
}

void print_error_bits_highpri (
        const error_bits_e error_bits) {

    debug_print_highpri("\nerror_bits(%04X)\n", error_bits); // Always print
}

#if (DEBUG_PRINT_RFM69==1)
    #define PRINT_ADD_ERROR_BITS_IFF(t,e) print_add_error_bits_iff(t,e)
#else
    #define PRINT_ADD_ERROR_BITS_IFF(t,e)
#endif

#if (DEBUG_PRINT_RFM69_HIGHPRI==1)
    #define PRINT_ERROR_BITS_HIGHPRI(e) print_error_bits_highpri(e)
#else
    #define PRINT_ERROR_BITS_HIGHPRI(e)
#endif

#define DEBUG_PRINT_RFM69_REGISTER_DETAIL 0

#if (WARNINGS==1)
    // Only used for checking, to make picking up value from makefile visible
    #if (_USERMAKEFILE_LIB_RFM69_XC_USER == 1)
        #warning _USERMAKEFILE_LIB_RFM69_XC_USER is _app_rfm69_on_xmos_native
    #elif (_USERMAKEFILE_LIB_RFM69_XC_USER == 2)
        #warning _USERMAKEFILE_LIB_RFM69_XC_USER is _Aquarium_1_x
    #elif (_USERMAKEFILE_LIB_RFM69_XC_USER == 3)
        #warning _USERMAKEFILE_LIB_RFM69_XC_USER is _Aquarium_rfm69_client
    #else
        #error _USERMAKEFILE_LIB_RFM69_XC_USER NOT DEFINED // Define in User makefile as XCC_FLAGS
    #endif
#endif

#define SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY _USERMAKEFILE_LIB_RFM69_XC_SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE


// **********************************************************************************
// Driver definition for HopeRF RFM69W/RFM69HW/RFM69CW/RFM69HCW, Semtech SX1231/1231H
// **********************************************************************************
// Copyright Felix Rusu 2016, http://www.LowPowerLab.com/contact
// **********************************************************************************
// License
// **********************************************************************************
// This program is free software; you can redistribute it
// and/or modify it under the terms of the GNU General
// Public License as published by the Free Software
// Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will
// be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A
// PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// Licence can be viewed at
// http://www.gnu.org/licenses/gpl-3.0.txt
//
// Please maintain this license information along with authorship
// and copyright notices in any redistribution of this code
// **********************************************************************************

#define RSSI_CSMA_LIMIT_DBM   (-90) // Upper RX signal sensitivity threshold in dBm for carrier sense access
#define RSSI_RX_LIMIT_DBM    (-110) // RSSI trigger level for Rssi interrupt
                                    // ( -40) Nothing received even if RSSI received normally is (-36) dBM
                                    // ( -50) Picks up most of the values when TX->RX on my bench
                                    // (-110) RFM69 library value
                                    // (-114) Default is REG_RSSI_TO_DBM(RF_RSSITHRESH_VALUE)

#define REG_RSSI_TO_DBM(val) (-(val/2)) // Like REG_RSSI_TO_DBM(RegRssiThresh)
#define DBM_TO_RSSI_REG(val) (-(val*2)) // Like DBM_TO_RSSI_REG(RSSI_RX_LIMIT_DBM)
//                                                                                                 Enabled blocks
#define RF69_MODE_SLEEP   (RF_OPMODE_MODE_SLEEP       >> RF_OPMODE_MODE_BITPOS) // =0  XTAL OFF    None
#define RF69_MODE_STANDBY (RF_OPMODE_MODE_STANDBY     >> RF_OPMODE_MODE_BITPOS) // =1  XTAL ON     Top regulator and crystal oscillator. DEFAULT value in SX1231
//                                                                                 =1  LISTEN MODE when RF_OPMODE_LISTEN_ON in REG_OPMODE
#define RF69_MODE_SYNTH   (RF_OPMODE_MODE_SYNTHESIZER >> RF_OPMODE_MODE_BITPOS) // =2  PLL ON      Frequency synthesizer
#define RF69_MODE_TX      (RF_OPMODE_MODE_TRANSMITTER >> RF_OPMODE_MODE_BITPOS) // =3  TX MODE     Frequency synthesizer and transmitter
#define RF69_MODE_RX      (RF_OPMODE_MODE_RECEIVER    >> RF_OPMODE_MODE_BITPOS) // =4  RX MODE     Frequency synthesizer and receiver

#define RF69_MODE_VOID    255

// #define null 0  // XC already has a "null" in the language

#define RF69_BROADCAST_ADDR 255
//
#define WAIT_FOR_CSMA_LIMIT_SET_MS      (WAIT_FOR_RADIO_MAX_MS/2) // Was RF69_CSMA_LIMIT_MS      1000. Could this be AQU=065?
#define WAIT_FOR_PACKET_SENT_MS         (WAIT_FOR_RADIO_MAX_MS/2) // Was WAIT_FOR_PACKET_SENT_MS 1000. AQU=065?
                                                                  // With baud rate RF_BITRATEMSB_55555 70 bytes (560 bit) takes 1/100 sec = 10 ms. Plus "not ready" 100 ms should be fine
#define WAIT_FOR_REGISTER_VALUE_MS      WAIT_FOR_RADIO_MAX_MS     // 200 in the original
#define WAIT_FOR_REGISTER_TEMP1_MEAS_MS WAIT_FOR_RADIO_MAX_MS     // In RFM69 there was no timeout like this
#define WAIT_FOR_SLEEP_MODE_EXIT_MS     WAIT_FOR_RADIO_MAX_MS     // In RFM69 there was no timeout like this
#define WAIT_FOR_RCCALIBRATION_MS       WAIT_FOR_RADIO_MAX_MS     // In RFM69 there was no timeout like this

#define NO_REGISTER_VALUE                   0
#define NO_REGISTER         NO_REGISTER_VALUE
#define SOME_REGISTER      (NO_REGISTER_VALUE+1)
#define COURSE_TEMP_OFFSET               (-90)  // puts the temperature reading in the ballpark, user can fine tune the returned value (Teig: Was COURSE_TEMP_COEF)

// The crystal oscillator frequency of the RF69 module (from RadioHead library)
#define RH_RF69_FXOSC 32000000.0

// The Frequency Synthesizer step = RH_RF69_FXOSC / 2exp19(=524288)  (p13 in datasheet)
#define RF69_FSTEP_FLOAT32  (RH_RF69_FXOSC / 524288) // = 61.03515625 used by RFM69 since it has a phase locked loop
                                                     // = 61.035156   is what I get with float (4 bytes) or even double (8 bytes),
                                                     //               even if I give the compiler 61.03515625 directly

typedef struct {
    error_bits_e               error_bits;
    uint8_t                    radio_mode;       // Was _mode. RF69_MODE_STANDBY. Should have been a struct!
    bool                       doListenToAll;    // Was _promiscuousMode
    uint8_t                    powerLevel_dBm;
    bool                       isRadio_RFM69HW;  // Was _RFM69HW
    //
    uint8_t                    nodeID;           // Was _address
    uint8_t                    device_type;      // Is RF_VERSION_VER
    //
    uint8_t                    APPPROTLEN;
    uint8_t                    PACKETLEN;
    waitForIRQInterruptCause_e waitForIRQInterruptCause;

    // ----------------------- Some overlapping in storage, but each for their own use -----------------------
    uint8_t                    TARGETID;         // should match NODEID. When received it was sent as GATEWAYID qwe
    uint8_t                    SENDERID;
    uint8_t                    CTLBYTE;
    packet_t                   PACKET;           // Contains PACKETLEN, TARGETID, SENDERID and CTLBYTE
} rfm69_internals_t;

#define ISRFM69HW false // From param "bool isRFM69HW=false"

// "INTERNAL FUNCTION" means that its service is needed by another i_radio.xxx interface call's call
// Some are not so, but it's ok to make them functions anyhow (like func_initialize, but it saved 756 bytes to make it a function!)

// PARAMETER SEQUENCE STANDARD: i_spi, spi_client, &spi_params, &rfm69_internals

// INTERNAL FUNCTION
bool
canSend (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);
// INTERNAL FUNCTION
void
do_getSomeRadioInternals (
        rfm69_internals_t      &rfm69_internals,
        some_rfm69_internals_t &return_some_rfm69_internals,
        packet_t               &PACKET);
void
func_encrypt16 (
        const char            key[],
        unsigned const static len,
        client spi_master_if  i_spi,
        spi_params_t          &spi_params,
        rfm69_internals_t     &rfm69_internals);

{interruptAndParsingResult_e, uint8_t} // RegIrqFlags2
handleSPIInterrupt (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

{interruptAndParsingResult_e, uint8_t} // RegIrqFlags2
_handleSPIInterrupt (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

void
func_initialize (
        const rfm69_params_t init,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
int16_t
readRSSI_dBm (
        const forceTrigger_t forceTrigger,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
int8_t
readTemperature_degC (
        const int8_t         calOffset,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
void
receiveBegin (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

void
_receiveBegin (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
bool
func_receiveDone (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

void
func_send (
        const uint8_t         TARGETID_toAddress,
        const packet_t        PACKET,
        client spi_master_if  i_spi,
        spi_params_t          &spi_params,
        rfm69_internals_t     &rfm69_internals);

// INTERNAL FUNCTION
void
sendFrame ( // timeout_delta_ms
        const uint8_t         TARGETID_toAddress,
        const packet_t        PACKET,
        client spi_master_if  i_spi,
        spi_params_t          &spi_params,
        rfm69_internals_t     &rfm69_internals);

void
setHighPower (
        const bool           isHighPowerOn,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
void
setHighPowerRegs (
        const bool           isHighPowerOn,
        client spi_master_if i_spi,
        spi_params_t         &spi_params);

// INTERNAL FUNCTION
void
setMode (
        const unsigned       debug_caller_id,
        const uint8_t        newMode,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
void
_setMode (
        const unsigned       debug_caller_id,
        const uint8_t        newMode,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
void
while_not_canSend_do_receiveDone (
        const unsigned       debug_caller_id,
        const unsigned       timeout_ms,   // Like WAIT_FOR_CSMA_LIMIT_SET_MS
        const unsigned       error_bitnum, // like ERROR_BITNUM_CAN_SEND
        client spi_master_if i_spi,        // For SPI_WRITE_REG and SPI_READ_REG macros
        spi_params_t         &spi_params,  // --"--
        rfm69_internals_t    &rfm69_internals);

// INTERNAL FUNCTION
uint8_t // readFromRegValue
while_readRegExpectedMask (
        const unsigned       debug_caller_id,
        const uint8_t        readFromReg,     // Like REG_IRQFLAGS1
        const uint8_t        readFromRegMask, // Like RF_IRQFLAGS1_MODEREADY
        const logic_e        invertedLogic,   // like logic_normal:   loop until bit in flag becomes 1
                                              //      logic_inverted: loop until bit in flag becomes 0
        const unsigned       timeout_ms,      // Like 200
        const unsigned       error_bitnum,       // like ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY
        client spi_master_if i_spi,           // For SPI_WRITE_REG and SPI_READ_REG macros
        spi_params_t         &spi_params,     // --"--
        rfm69_internals_t    &rfm69_internals);

#if (DEBUG_SHARED_LOG_VALUE_LIB==1)
    unsigned g_radio_log_value;
#endif

// ---------- THE CODE ----------

// -----------------------------------------------------------------------------------------------------
// TEST: OF FLOAT CALCULATIONS IN freq_register_value_to_Hz AND freq_Hz_to_register_value IN APPLICATION:
//
// unsigned frequencyHz = freq_register_value_to_Hz (MY_RFM69_FREQ_REGS);
// freq_Hz_to_register_value (frequencyHz); // Just for comparing and printing
//
// causes the following printing:
//     6C6D2F reg MUL fstep4 61.035156 = 433705984 Hz
//     433705984 Hz DIV fstep4 61.035156 = 6C6D2F reg

//       If this or the below is called it would cause the linker to kick in the floating point library:
unsigned freq_register_value_to_Hz (const uint32_t register_value) { // Test: 6C6D2F

    // Approximate since RF69_FSTEP_FLOAT32 becomes 61.035156 instead of 61.03515625, using float or double:

    unsigned    frequency_Hz;
    const float fstep4 = RF69_FSTEP_FLOAT32; // Using double (name fstep8) with %lf in print below does not help, still only 6 decimals

    // I think the floating point library (mul, div, plus and minus used), that's included once the compiler sees a need for it,
    // on a (now obsoleted)startKIT (with silicon on board: XS1-A8-DEV) takes about 2584 bytes. On an eXplorerKIT
    // (with silicon on board:XE216-512-TQ128) it takes 2636 bytes.
    // See http://www.teigfam.net/oyvind/home/technology/141-xc-is-c-plus-x/#floating_point_library
    //
    // With RF69_FSTEP_FLOAT32 defined as a float the compiler will do all these TYPE CONVERSIONS (which is more than TYPE CAST!)
    // automatically, without all the explicit typing. However, have beeen brought up with occam I like to see this explicit:

    frequency_Hz = (unsigned) (((float)register_value) * fstep4);

    debug_print ("%06X reg MUL fstep%u %f = %u Hz\n", register_value, sizeof fstep4, fstep4, frequency_Hz);
    //                                 %lf for double
    //          6C6D2F reg MUL fstep4 61.035156 = 433705984 Hz
    // Theoretical value with calculator and FSTEP as 61.03515625 is 433705993 (9 more, or about 2 ppm)
    // The RFM69 uses a phased locked loop

    return frequency_Hz; // Test: 433705984 (9 too little, or about 2 ppm)
}

//       If this or the above is called it would cause the linker to kick in the floating point library:
uint32_t freq_Hz_to_register_value (const unsigned frequency_Hz) { //Test: 433705984

    // Approximate since RF69_FSTEP_FLOAT32 becomes 61.035156 instead of 61.03515625, using float or double:

    uint32_t    register_value;
    const float fstep4 = RF69_FSTEP_FLOAT32; // Using double (name fstep8) with %lf in print below does not help, still only 6 decimals

    register_value = (uint32_t) (((float)frequency_Hz) / fstep4);

    debug_print ("%u Hz DIV fstep%u %f = %06X reg\n", frequency_Hz, sizeof fstep4, fstep4, register_value);
    //                              %lf for double
    //     433705984 Hz DIV fstep4 61.035156 = 6C6D2F reg

    return register_value; // Test: 6C6D2F WHICH IS BACK TO freq_register_value_to_Hz's input value
}


// INTERNAL FUNCTION
bool
canSend (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,24); // NEXT FROM 25 in rfm69_spi_driver.xc
    debug_print ("%s", "canSend\n");

    bool canSendNow;

    if ((rfm69_internals.radio_mode == RF69_MODE_RX) and (rfm69_internals.PACKETLEN == 0)) {
        canSendNow = (readRSSI_dBm (FORCETRIGGER_OFF, i_spi, spi_params, rfm69_internals) < RSSI_CSMA_LIMIT_DBM);
        // if signal stronger than -100dBm is detected assume channel activity
    } else {
        canSendNow = false;
    }

    if (canSendNow) {
        setMode (7, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
    } else {}

    debug_print ("%s", "-- canSend\n");

    return canSendNow;
} // canSend


void
do_getSomeRadioInternals (
        rfm69_internals_t      &rfm69_internals,
        some_rfm69_internals_t &return_some_rfm69_internals,
        packet_t               &PACKET)
{
    return_some_rfm69_internals.error_bits = rfm69_internals.error_bits;

    return_some_rfm69_internals.APPPROTLEN = rfm69_internals.APPPROTLEN;
    return_some_rfm69_internals.PACKETLEN  = rfm69_internals.PACKETLEN;
    return_some_rfm69_internals.TARGETID   = rfm69_internals.TARGETID;
    return_some_rfm69_internals.SENDERID   = rfm69_internals.SENDERID;

    PACKET = rfm69_internals.PACKET;
}

// To enable encryption: rfm69_internals.encrypt("ABCDEFGHIJKLMNOP");
// To disable encryption: rfm69_internals.encrypt(null) or rfm69_internals.encrypt(0)
// KEY HAS TO BE 16 bytes

// Here is a message of 20 bytes (5 words) seen on the receiver side with the same KEY on both sides.
//     RX[0]: 00636213 CommHeaderRFM69
//     RX[1]: 00010114 appHeading
//     RX[2]: 00071F63 07=   1F=appPowerLevel, 63=appNODEID,
//     RX[3]: 00000009 appSeqCnt
//     RX[4]: 54931951 CRC32
//            RSSI -17, P 31, RXappSeqCnt 9
//
// Here is about the same message seen on the receiver side, but the sender KEY has changed with one of 16 chars:
//     RX[0]: CA2CD513 3391935763
//     RX[1]: E7030427 3875734567
//     RX[2]: 7A32CC6B 2050149483
//     RX[3]: 3A948B48 982813512
//     RX[4]: 6121FF07 1629617927
//            RSSI -48, fail numAppCRCerrs 1

void
func_encrypt16 (
        const char           key[], unsigned const static len,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,23);
    debug_print ("func_encrypt16 %s len %d\n", key, len);

    setMode (8, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
    if (len==KEY_LEN) {
        spiBurstWriteSeqRegsOrFIFO (REG_AESKEY1, key, len, i_spi, spi_params);
    } else {}

    // REG_PACKETCONFIG2 (RegPacketConfig2 in manual, page 75)
    // RF_PACKET2_AES_ON AesOn Enable the AES encryption/decryption:
    // BIT0=Off
    // BIT1=On (full_payload_01 limited to 66 bytes maximum)
    //
    uint8_t RegPacketConfig2 = SPI_READ_REG (REG_PACKETCONFIG2);
    if (len==KEY_LEN) {
        RegPacketConfig2 or_eq RF_PACKET2_AES_ON; // Encryption is on
    } else {
        RegPacketConfig2 and_eq (compl RF_PACKET2_AES_ON); // Encryption is off
    }
    //
    SPI_WRITE_REG (REG_PACKETCONFIG2, RegPacketConfig2);
    debug_print ("%s", "-- func_encrypt16\n");
} // func_encrypt16

{interruptAndParsingResult_e, uint8_t} // RegIrqFlags2
handleSPIInterrupt (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    // WE GET BOTH TX AND RX INTERRUPTS IN HERE, BUT NOT SO FOR THE RFM69 LIBRARY (SEE FILE "RFM69.cpp")
    //     RFM69 RX interrupt:
    //         Handled in void RFM69::interruptHandler() - interrupt gets called when a packet is received
    //         Condition to do anything at all (except reading RSSI):
    //         if (_mode == RF69_MODE_RX && (readReg(REG_IRQFLAGS2) & RF_IRQFLAGS2_PAYLOADREADY)) {...} else {nothing}
    //     RFM69 TX interrupt:
    //         Handled inline in void RFM69::sendFrame by just busy-polling for the interrupt pin for WAIT_FOR_PACKET_SENT_MS
    //         and then optionally (*) wait for RF_IRQFLAGS2_PACKETSENT to become set. Then there is this call:
    //         setMode (RF69_MODE_STANDBY)
    //         (*) That line is commented out, but I guess it's meant to handle a situation where WAIT_FOR_PACKET_SENT_MS passes before
    //             it's actually sent. It doesn't look like REG_IRQFLAGS2 has to be read at all

    // Equivalent to Arduino: attachInterrupt(_interruptNum, rfm69_internals::isr0, RISING);
    // About SPI_IRQ, search for RF_DIOMAPPING1_DIO0_01 here
    //     https://learn.adafruit.com/adafruit-rfm69hcw-and-rfm96-rfm95-rfm98-lora-packet-padio-breakouts/pinouts
    //         G0 - the rfm69_internals's "GPIO 0" pin, also known as the IRQ pin, used for interrupt request notification
    //         from the rfm69_internals to the microcontroller, 3.3V logic level
    //     https://www.semtech.com/uploads/documents/sx1231.pdf
    //         It's called DIO0 on the diagram: Digital I/O, software configured
    //         5.3.2. DIO Pins Mapping in Packet Mode
    //         Mode RX: Diox mapping 01 -> full_payload_01 ready

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,22);

    interruptAndParsingResult_e interruptAndParsingResult_priority = void_IRQ;
    interruptAndParsingResult_e interruptAndParsingResult          = void_IRQ;


    uint8_t RegIrqFlags2 = SPI_READ_REG (REG_IRQFLAGS2); //            RegIrqFlags2, we have only seen these combinations:
    //                                         0x66                   ...
    //                                             0x64               IS_MYTARGET_SLAVE CRC wrong from some RX low signal
    //                                                 0x44           IS_MYTARGET_SLAVE CRC wrong from some RX low signal
    //                                                     0x46       message_received_too_short_for_IRQ
    //                                                         0x08   IS_MYTARGET_MASTER
    //  #define RF_IRQFLAGS2_FIFOFULL     0x80                        FifoFull
    //  #define RF_IRQFLAGS2_FIFONOTEMPTY 0x40  X   X   X   X         FifoNotEmpty
    //  #define RF_IRQFLAGS2_FIFOLEVEL    0x20  X   X                 FifoLevel
    //  #define RF_IRQFLAGS2_FIFOOVERRUN  0x10                        FifoOverrun
    //  #define RF_IRQFLAGS2_PACKETSENT   0x08                  X     PacketSent
    //  #define RF_IRQFLAGS2_PAYLOADREADY 0x04  X   X   X   X         PayloadReady
    //  #define RF_IRQFLAGS2_CRCOK        0x02  X           X         CrcOk
    //  #define RF_IRQFLAGS2_LOWBAT       0x01                        not present on RFM69/SX1231

    debug_print("%s", "IRQ handleSPIInterrupt\n");
    if ((rfm69_internals.waitForIRQInterruptCause == messagePacketSent_IRQExpected) or
        (rfm69_internals.waitForIRQInterruptCause == messagePacketNotSent_ErrorIRQNotExpected)){

        if ((RegIrqFlags2 bitand RF_IRQFLAGS2_PACKETSENT) != 0) {
            if ((RegIrqFlags2 bitand (compl RF_IRQFLAGS2_PACKETSENT)) == 0) {
                interruptAndParsingResult = messagePacketSentOk_IRQ; // No other bits set
            } else {
                // === This is only seen if only TX, or if board not connected ===
                interruptAndParsingResult = messagePacketSentOk2_IRQ; // Other bits set
            }
        } else {
            if ((RegIrqFlags2 bitand (compl RF_IRQFLAGS2_PACKETSENT)) != 0) {
                interruptAndParsingResult = messagePacketSentOkMissing_IRQ; // No other bits set
            } else {
                interruptAndParsingResult = messagePacketSentOkMissing2_IRQ; // Other bits set
            }
        }

        debug_print("%s", "IRQ TX sent\n");

        rfm69_internals.waitForIRQInterruptCause = no_IRQExpected;

    } else if ((rfm69_internals.radio_mode == RF69_MODE_RX) and ((RegIrqFlags2 bitand RF_IRQFLAGS2_PAYLOADREADY) != 0)) {

        bool radioCRC16err = false;

        debug_print("%s", "IRQ RX\n");

        // RFM69 library had a setMode (RF69_MODE_STANDBY), not needed in any sense as far as I can see
        // Reinserted 29Oct2018 RFM69=001 TODO remove
        #if (SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY == 1)
            setMode (10, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
            // Is this because "Another important feature is ability to fill and empty the FIFO in Sleep/Stdby mode,
            // ensuring optimum power consumption and adding more flexibility for the software."
        #endif

        if ((RegIrqFlags2 bitand RF_IRQFLAGS2_CRCOK) == 0) { // "OK" is a '1'
            // Only possible if SEMANTICS_DO_CRC_ERR_NO_IRQ == 0
            // Requires RF_PACKET1_CRCAUTOCLEAR_OFF
            interruptAndParsingResult_priority = messageRadioCRC16Err_IRQ; // May be overwritten below
            radioCRC16err = true;
            // Will happen quite often since incoming from other type of radios probably would be seen here

            // By default, if the CRC check fails then the FIFO is automatically cleared and no interrupt is generated.
            // If the CRC fails the PayloadReady interrupt is not generated and the FIFO is cleared.
            // This function can be overridden by setting CrcAutoClearOff = 1, forcing the availability of PayloadReady
            // interrupt and the payload in the FIFO even if the CRC fails.

            // https://groups.google.com/forum/#!topic/radiohead-arduino/nhtkTIVM7fM
            //     Application's data field, error modes?
            // https://stackoverflow.com/questions/49885413/what-is-the-possibility-of-crc16-collisions-on-20-bytes-of-data-2
            //     What is the possibility of CRC16 collisions on 20 bytes of data? (2)
            // http://www.xcore.com/viewtopic.php?p=33015#p33015
            //     How much 433 MHz emitted by a typical board?

        } else {}

        // We want to see what's coming in also if radioCRC16err:

        BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,21);

        SPI_BEGIN_TRANSACTION();
        {
            SPI_TRANSFER8 (REG_FIFO bitand (compl SPI_WRITE_MASK)); // START REG_FIFO, ignoring return (=read) value

            /// Fill CommHeaderRFM69
            for (unsigned index = IOF_PACKETLEN_0; index < RFM69HEADER_LEN08_IS4; index++) {
                rfm69_internals.PACKET.u.packet_u1_uint8_arr[index] = SPI_TRANSFER8 (SPI_RD_FIFO_CMD);
            }
            rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN += LENM1_LEN;

            rfm69_internals.PACKETLEN = rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN; // First. Special FIFO admin since RF_PACKET1_FORMAT_VARIABLE

            if (rfm69_internals.PACKETLEN != PACKET_LEN08) {
                interruptAndParsingResult = messagePacketLenErr_IRQ;
            } else {

                rfm69_internals.TARGETID   = rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.TARGETID; // Sent as GATEWAYID
                rfm69_internals.SENDERID   = rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.SENDERID;
                rfm69_internals.APPPROTLEN = rfm69_internals.PACKETLEN - RFM69HEADER_LEN08_IS4; // Already read these

                for (unsigned index = 0; index < rfm69_internals.APPPROTLEN; index++) {
                    uint8_t reg = SPI_TRANSFER8 (SPI_RD_FIFO_CMD);
                    rfm69_internals.PACKET.u.packet_u1_uint8_arr[RFM69HEADER_LEN08_IS4 + index] = reg;
                }

                if (radioCRC16err == true) {
                    // No code since CRC32 of application level error is not meaningful if CRC16 error of lower level
                } else {
                    // CRC32 error is only meaningful if CRC16 is passed as fine
                    crc32_t appCRC32 = calc_CRC32 (rfm69_internals.PACKET.u.packet_u2_uint32_arr, APPCRC_OVER_LEN32, 0);
                    bool appCRC32err = not (appCRC32 == rfm69_internals.PACKET.u.packet_u0.appCRC32);
                    if (appCRC32err) {
                        interruptAndParsingResult_priority = messageAppCRC32Err_IRQ;
                    } else {}
                    debug_print ("RX CRC32 %08X OK:%u\n", rfm69_internals.PACKET.u.packet_u0.appCRC32, not appCRC32err);
                }
            }
        }
        SPI_END_TRANSACTION();

        BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,20);

        {   // Flush the FIFO, most often for ZERO bytes!
            // RFM69=001 instead of SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY and SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE (coupled now)
            // which is not as general (for all cases) (but the RF69_MODE_RX->RF69_MODE_STANDBY then RF69_MODE_STANDBY->RF69_MODE_RX clears FIFO,
            // but I don't like the idea NOT to listen for a while). I try this. It's according to
            // SEMTECH SX1231H datasheet 5.2.2.3 (p45). Interrupt Sources and Flags: In other words, FifoNotEmpty state must be checked
            //     after each read operation for a decision on the next one (FifoNotEmpty = 1: more byte(s) to read; FifoNotEmpty = 0: no more byte to read)
            // We have read all data that we think we need. Let's just discard what more there might be!
            // We see some times that even this is not enough, this will not clear the RF_IRQFLAGS2_FIFONOTEMPTY flag. This may happen regardless of SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE
            // when debugging and printing, or only if SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE=0 when FLASHed. This is why we made ultimateIRQclear to make any situation recoverable
            uint8_t reg;
            reg = while_readRegExpectedMask (8,               // "008-B loop". discarding reg (Never seen it loop:)
                    REG_IRQFLAGS2, RF_IRQFLAGS2_FIFONOTEMPTY, // RegIrqFlags2, FifoNotEmpty
                    logic_inverted,                           // loop until bit in flag becomes 0
                    WAIT_FOR_REGISTER_VALUE_MS,               // timeout_ms (Client ver 0822 did not help to increase by 5 times)
                    ERROR_BITNUM_RF_IRQFLAGS2_FIFONOTEMPTY,   // error_bitnum
                    i_spi, spi_params, rfm69_internals);
        }

        if (interruptAndParsingResult_priority != void_IRQ) {
            interruptAndParsingResult = interruptAndParsingResult_priority;
        } else {}

        if (interruptAndParsingResult == void_IRQ) {
            if (
                (rfm69_internals.doListenToAll) or
                (rfm69_internals.TARGETID == rfm69_internals.nodeID) or
                (rfm69_internals.TARGETID == RF69_BROADCAST_ADDR)) {

                setMode (11, RF69_MODE_RX, i_spi, spi_params, rfm69_internals);
                interruptAndParsingResult = messageReceivedOk_IRQ;
            } else {
                interruptAndParsingResult = messageNotForThisNode_IRQ;
            }
        } else {} // Already handled

    } else {
        interruptAndParsingResult = messageNotExpectedBy_IRQ;
    }

    if (interruptAndParsingResult != messageReceivedOk_IRQ) { // TODO why? Check original
        setMode (10, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
    } else {}

    debug_print ("PRO %u, LEN %u, ID %u,%u\n",
            rfm69_internals.doListenToAll,
            rfm69_internals.PACKETLEN,
            rfm69_internals.TARGETID, rfm69_internals.nodeID);

    debug_print("-- IRQ handleSPIInterrupt with %u\n", interruptAndParsingResult);
    debug_print("handleSPIInterrupt %u %08X %u\n", interruptAndParsingResult, RegIrqFlags2, rfm69_internals.PACKETLEN);
    return {interruptAndParsingResult, RegIrqFlags2};

    // Observe that there is a possible receiveBegin call on return from handleSPIInterrupt!

} // handleSPIInterrupt

{interruptAndParsingResult_e, uint8_t} // RegIrqFlags2
_handleSPIInterrupt (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    // WE GET BOTH TX AND RX INTERRUPTS IN HERE, BUT NOT SO FOR THE RFM69 LIBRARY (SEE FILE "RFM69.cpp")
    //     RFM69 RX interrupt:
    //         Handled in void RFM69::interruptHandler() - interrupt gets called when a packet is received
    //         Condition to do anything at all (except reading RSSI):
    //         if (_mode == RF69_MODE_RX && (readReg(REG_IRQFLAGS2) & RF_IRQFLAGS2_PAYLOADREADY)) {...} else {nothing}
    //     RFM69 TX interrupt:
    //         Handled inline in void RFM69::sendFrame by just busy-polling for the interrupt pin for WAIT_FOR_PACKET_SENT_MS
    //         and then optionally (*) wait for RF_IRQFLAGS2_PACKETSENT to become set. Then there is this call:
    //         setMode (RF69_MODE_STANDBY)
    //         (*) That line is commented out, but I guess it's meant to handle a situation where WAIT_FOR_PACKET_SENT_MS passes before
    //             it's actually sent. It doesn't look like REG_IRQFLAGS2 has to be read at all

    // Equivalent to Arduino: attachInterrupt(_interruptNum, rfm69_internals::isr0, RISING);
    // About SPI_IRQ, search for RF_DIOMAPPING1_DIO0_01 here
    //     https://learn.adafruit.com/adafruit-rfm69hcw-and-rfm96-rfm95-rfm98-lora-packet-padio-breakouts/pinouts
    //         G0 - the rfm69_internals's "GPIO 0" pin, also known as the IRQ pin, used for interrupt request notification
    //         from the rfm69_internals to the microcontroller, 3.3V logic level
    //     https://www.semtech.com/uploads/documents/sx1231.pdf
    //         It's called DIO0 on the diagram: Digital I/O, software configured
    //         5.3.2. DIO Pins Mapping in Packet Mode
    //         Mode RX: Diox mapping 01 -> full_payload_01 ready

    VALUE_XSCOPE_LIB (RFM69_VALUE,1002);
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,22);

    interruptAndParsingResult_e interruptAndParsingResult_priority = void_IRQ;
    interruptAndParsingResult_e interruptAndParsingResult          = void_IRQ;


    uint8_t RegIrqFlags2 = SPI_READ_REG (REG_IRQFLAGS2); //            RegIrqFlags2, we have only seen these combinations:
    //                                         0x66                   ...
    //                                             0x64               IS_MYTARGET_SLAVE CRC wrong from some RX low signal
    //                                                 0x44           IS_MYTARGET_SLAVE CRC wrong from some RX low signal
    //                                                     0x46       message_received_too_short_for_IRQ
    //                                                         0x08   IS_MYTARGET_MASTER
    //  #define RF_IRQFLAGS2_FIFOFULL     0x80                        FifoFull
    //  #define RF_IRQFLAGS2_FIFONOTEMPTY 0x40  X   X   X   X         FifoNotEmpty
    //  #define RF_IRQFLAGS2_FIFOLEVEL    0x20  X   X                 FifoLevel
    //  #define RF_IRQFLAGS2_FIFOOVERRUN  0x10                        FifoOverrun
    //  #define RF_IRQFLAGS2_PACKETSENT   0x08                  X     PacketSent
    //  #define RF_IRQFLAGS2_PAYLOADREADY 0x04  X   X   X   X         PayloadReady
    //  #define RF_IRQFLAGS2_CRCOK        0x02  X           X         CrcOk
    //  #define RF_IRQFLAGS2_LOWBAT       0x01                        not present on RFM69/SX1231

    debug_print("%s", "IRQ handleSPIInterrupt\n");
    if ((rfm69_internals.waitForIRQInterruptCause == messagePacketSent_IRQExpected) or
        (rfm69_internals.waitForIRQInterruptCause == messagePacketNotSent_ErrorIRQNotExpected)){

        if ((RegIrqFlags2 bitand RF_IRQFLAGS2_PACKETSENT) != 0) {
            if ((RegIrqFlags2 bitand (compl RF_IRQFLAGS2_PACKETSENT)) == 0) {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1003);
                interruptAndParsingResult = messagePacketSentOk_IRQ; // No other bits set
            } else {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1004);
                // === This is only seen if only TX, or if board not connected ===
                interruptAndParsingResult = messagePacketSentOk2_IRQ; // Other bits set
            }
        } else {
            if ((RegIrqFlags2 bitand (compl RF_IRQFLAGS2_PACKETSENT)) != 0) {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1005);
                interruptAndParsingResult = messagePacketSentOkMissing_IRQ; // No other bits set
            } else {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1006);
                interruptAndParsingResult = messagePacketSentOkMissing2_IRQ; // Other bits set
            }
        }

        debug_print("%s", "IRQ TX sent\n");

        rfm69_internals.waitForIRQInterruptCause = no_IRQExpected;

    } else if ((rfm69_internals.radio_mode == RF69_MODE_RX) and ((RegIrqFlags2 bitand RF_IRQFLAGS2_PAYLOADREADY) != 0)) {

        VALUE_XSCOPE_LIB (RFM69_VALUE,1007);
        bool radioCRC16err = false;

        debug_print("%s", "IRQ RX\n");

        // RFM69 library had a setMode (RF69_MODE_STANDBY), not needed in any sense as far as I can see
        // Reinserted 29Oct2018 RFM69=001 TODO remove
        #if (SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY == 1)
            VALUE_XSCOPE_LIB (RFM69_VALUE,1008);
            setMode (10, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
            // Is this because "Another important feature is ability to fill and empty the FIFO in Sleep/Stdby mode,
            // ensuring optimum power consumption and adding more flexibility for the software."
        #endif

        if ((RegIrqFlags2 bitand RF_IRQFLAGS2_CRCOK) == 0) { // "OK" is a '1'
            VALUE_XSCOPE_LIB (RFM69_VALUE,1009);
            // Only possible if SEMANTICS_DO_CRC_ERR_NO_IRQ == 0
            // Requires RF_PACKET1_CRCAUTOCLEAR_OFF
            interruptAndParsingResult_priority = messageRadioCRC16Err_IRQ; // May be overwritten below
            radioCRC16err = true;
            // Will happen quite often since incoming from other type of radios probably would be seen here

            // By default, if the CRC check fails then the FIFO is automatically cleared and no interrupt is generated.
            // If the CRC fails the PayloadReady interrupt is not generated and the FIFO is cleared.
            // This function can be overridden by setting CrcAutoClearOff = 1, forcing the availability of PayloadReady
            // interrupt and the payload in the FIFO even if the CRC fails.

            // https://groups.google.com/forum/#!topic/radiohead-arduino/nhtkTIVM7fM
            //     Application's data field, error modes?
            // https://stackoverflow.com/questions/49885413/what-is-the-possibility-of-crc16-collisions-on-20-bytes-of-data-2
            //     What is the possibility of CRC16 collisions on 20 bytes of data? (2)
            // http://www.xcore.com/viewtopic.php?p=33015#p33015
            //     How much 433 MHz emitted by a typical board?

        } else {
            VALUE_XSCOPE_LIB (RFM69_VALUE,1010);
        }

        // We want to see what's coming in also if radioCRC16err:

        BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,21);

        VALUE_XSCOPE_LIB (RFM69_VALUE,1801);
        SPI_BEGIN_TRANSACTION();
        VALUE_XSCOPE_LIB (RFM69_VALUE,1802);

        {
            VALUE_XSCOPE_LIB (RFM69_VALUE,1011);
            SPI_TRANSFER8 (REG_FIFO bitand (compl SPI_WRITE_MASK)); // START REG_FIFO, ignoring return (=read) value

            /// Fill CommHeaderRFM69
            for (unsigned index = IOF_PACKETLEN_0; index < RFM69HEADER_LEN08_IS4; index++) {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1012+index);
                rfm69_internals.PACKET.u.packet_u1_uint8_arr[index] = SPI_TRANSFER8 (SPI_RD_FIFO_CMD);
            }
            rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN += LENM1_LEN;

            rfm69_internals.PACKETLEN = rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN; // First. Special FIFO admin since RF_PACKET1_FORMAT_VARIABLE

            if (rfm69_internals.PACKETLEN != PACKET_LEN08) {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1101);
                interruptAndParsingResult = messagePacketLenErr_IRQ;
            } else {

                rfm69_internals.TARGETID   = rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.TARGETID; // Sent as GATEWAYID
                rfm69_internals.SENDERID   = rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.SENDERID;
                rfm69_internals.APPPROTLEN = rfm69_internals.PACKETLEN - RFM69HEADER_LEN08_IS4; // Already read these

                for (unsigned index = 0; index < rfm69_internals.APPPROTLEN; index++) {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1200+index);
                    uint8_t reg = SPI_TRANSFER8 (SPI_RD_FIFO_CMD);
                    rfm69_internals.PACKET.u.packet_u1_uint8_arr[RFM69HEADER_LEN08_IS4 + index] = reg;
                }

                if (radioCRC16err == true) {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1200);
                    // No code since CRC32 of application level error is not meaningful if CRC16 error of lower level
                } else {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1201);
                    // CRC32 error is only meaningful if CRC16 is passed as fine
                    crc32_t appCRC32 = calc_CRC32 (rfm69_internals.PACKET.u.packet_u2_uint32_arr, APPCRC_OVER_LEN32, 0);
                    bool appCRC32err = not (appCRC32 == rfm69_internals.PACKET.u.packet_u0.appCRC32);
                    if (appCRC32err) {
                        interruptAndParsingResult_priority = messageAppCRC32Err_IRQ;
                    } else {}
                    debug_print ("RX CRC32 %08X OK:%u\n", rfm69_internals.PACKET.u.packet_u0.appCRC32, not appCRC32err);
                }
            }
        }
        VALUE_XSCOPE_LIB (RFM69_VALUE,1803);
        SPI_END_TRANSACTION();
        VALUE_XSCOPE_LIB (RFM69_VALUE,1804);

        BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,20);

        {   // Flush the FIFO, most often for ZERO bytes!
            // RFM69=001 instead of SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY and SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE (coupled now)
            // which is not as general (for all cases) (but the RF69_MODE_RX->RF69_MODE_STANDBY then RF69_MODE_STANDBY->RF69_MODE_RX clears FIFO,
            // but I don't like the idea NOT to listen for a while). I try this. It's according to
            // SEMTECH SX1231H datasheet 5.2.2.3 (p45). Interrupt Sources and Flags: In other words, FifoNotEmpty state must be checked
            //     after each read operation for a decision on the next one (FifoNotEmpty = 1: more byte(s) to read; FifoNotEmpty = 0: no more byte to read)
            // We have read all data that we think we need. Let's just discard what more there might be!
            // We see some times that even this is not enough, this will not clear the RF_IRQFLAGS2_FIFONOTEMPTY flag. This may happen regardless of SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE
            // when debugging and printing, or only if SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE=0 when FLASHed. This is why we made ultimateIRQclear to make any situation recoverable
            VALUE_XSCOPE_LIB (RFM69_VALUE,1301);
            uint8_t reg;
            reg = while_readRegExpectedMask (8,               // "008-B loop". discarding reg (Never seen it loop:)
                    REG_IRQFLAGS2, RF_IRQFLAGS2_FIFONOTEMPTY, // RegIrqFlags2, FifoNotEmpty
                    logic_inverted,                           // loop until bit in flag becomes 0
                    WAIT_FOR_REGISTER_VALUE_MS,               // timeout_ms (Client ver 0822 did not help to increase by 5 times)
                    ERROR_BITNUM_RF_IRQFLAGS2_FIFONOTEMPTY,   // error_bitnum
                    i_spi, spi_params, rfm69_internals);
            VALUE_XSCOPE_LIB (RFM69_VALUE,1302);
        }

        if (interruptAndParsingResult_priority != void_IRQ) {
            VALUE_XSCOPE_LIB (RFM69_VALUE,1303);
            interruptAndParsingResult = interruptAndParsingResult_priority;
        } else {}

        if (interruptAndParsingResult == void_IRQ) {
            if (
                (rfm69_internals.doListenToAll) or
                (rfm69_internals.TARGETID == rfm69_internals.nodeID) or
                (rfm69_internals.TARGETID == RF69_BROADCAST_ADDR)) {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1304);

                setMode (11, RF69_MODE_RX, i_spi, spi_params, rfm69_internals);
                interruptAndParsingResult = messageReceivedOk_IRQ;
            } else {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1305);
                interruptAndParsingResult = messageNotForThisNode_IRQ;
            }
        } else {
            VALUE_XSCOPE_LIB (RFM69_VALUE,1306);
        } // Already handled

    } else {
        VALUE_XSCOPE_LIB (RFM69_VALUE,1307);
        interruptAndParsingResult = messageNotExpectedBy_IRQ;
    }

    if (interruptAndParsingResult != messageReceivedOk_IRQ) { // TODO why? Check original
        VALUE_XSCOPE_LIB (RFM69_VALUE,1308);
        // AQU=065 comment out this line removes the error:
        _setMode (10, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
    } else {}

    debug_print ("PRO %u, LEN %u, ID %u,%u\n",
            rfm69_internals.doListenToAll,
            rfm69_internals.PACKETLEN,
            rfm69_internals.TARGETID, rfm69_internals.nodeID);

    debug_print("-- IRQ handleSPIInterrupt with %u\n", interruptAndParsingResult);
    debug_print("handleSPIInterrupt %u %08X %u\n", interruptAndParsingResult, RegIrqFlags2, rfm69_internals.PACKETLEN);
    VALUE_XSCOPE_LIB (RFM69_VALUE,1309);
    return {interruptAndParsingResult, RegIrqFlags2};

    // Observe that there is a possible receiveBegin call on return from handleSPIInterrupt!

} // _handleSPIInterrupt


void
func_initialize (
        const rfm69_params_t init,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    debug_print("%s", "IRQ func_initialize\n");

    uint8_t reg;

    #define NUM_REGISTERS_INIT 23
    #define IOF_REG_BITRATEMSB  2
    #define IOF_REG_BITRATELSB  3
    //
    uint8_t       reg_value_readback [NUM_REGISTERS_INIT];
    const uint8_t CONFIG             [NUM_REGISTERS_INIT] [2] =
    {
        // ## iof_config
        /* 00 0x01 */ { REG_OPMODE, RF_OPMODE_SEQUENCER_OFF bitor RF_OPMODE_LISTEN_OFF bitor RF_OPMODE_MODE_STANDBY },
        /* 01 0x02 */ { REG_DATAMODUL, RF_DATAMODUL_DATAMODE_PACKET bitor
                          RF_DATAMODUL_MODULATIONTYPE_FSK bitor
                          RF_DATAMODUL_MODULATIONSHAPING_00 }, // no shaping
        /* 02 0x03 */ { REG_BITRATEMSB, RF_BITRATEMSB_55555},
        /* 03 0x04 */ { REG_BITRATELSB, RF_BITRATELSB_55555},
        /* 04 0x05 */ { REG_FDEVMSB, RF_FDEVMSB_50000}, // default: 5KHz, (FDEV + BitRate / 2 <= 500KHz)
        /* 05 0x06 */ { REG_FDEVLSB, RF_FDEVLSB_50000},
        /* 06 0x07 */ { REG_FRFMSB, (uint8_t) ((init.RegFrf >> 16) bitand 0xff)}, // mask low BYTE (zero code by compiler, type transfer does it)
        /* 07 0x08 */ { REG_FRFMID, (uint8_t) ((init.RegFrf >>  8) bitand 0xff)}, // --"--
        /* 08 0x09 */ { REG_FRFLSB, (uint8_t) ((init.RegFrf      ) bitand 0xff)}, // --"--
        // .. 0x11    { REG_PALEVEL, RF_PALEVEL_PA0_ON bitor RF_PALEVEL_PA1_OFF bitor RF_PALEVEL_PA2_OFF bitor RF_PALEVEL_OUTPUTPOWER_11111},
        // .. 0x13    { REG_OCP, RF_OCP_ON bitor RF_OCP_TRIM_95 }, // over current protection (default is 95mA)
        //              REG_RXBW defaults are { REG_RXBW, RF_RXBW_DCCFREQ_010 bitor RF_RXBW_MANT_24 bitor RF_RXBW_EXP_5} (RxBw: 10.4KHz)
        /* 09 0x19 */ { REG_RXBW, RF_RXBW_DCCFREQ_010 bitor RF_RXBW_MANT_16 bitor RF_RXBW_EXP_2 }, // (BitRate < 2 * RxBw)
        // .. 0x19    { REG_RXBW, RF_RXBW_DCCFREQ_010 bitor RF_RXBW_MANT_24 bitor RF_RXBW_EXP_3 }, // for BR-19200
        /* 10 0x25 */ { REG_DIOMAPPING1, RF_DIOMAPPING1_DIO0_01 }, // DIO0 is the only IRQ we're using
        /* 11 0x26 */ { REG_DIOMAPPING2, RF_DIOMAPPING2_CLKOUT_OFF }, // DIO5 ClkOut disable for power saving
        /* 12 0x28 */ { REG_IRQFLAGS2, RF_IRQFLAGS2_FIFOOVERRUN }, // writing to this bit ensures that the FIFO & status flags are reset
        /* 13 0x29 */ { REG_RSSITHRESH, DBM_TO_RSSI_REG(RSSI_RX_LIMIT_DBM) },
        // 14 0x2D    { REG_PREAMBLELSB, RF_PREAMBLESIZE_LSB_VALUE } // default 3 preamble bytes 0xAAAAAA /
        /* 15 0x2E */ { REG_SYNCCONFIG, RF_SYNC_ON bitor RF_SYNC_FIFOFILL_AUTO bitor RF_SYNC_SIZE_2 bitor RF_SYNC_TOL_0 },
        /* 16 0x2F */ { REG_SYNCVALUE1, RFM69_SYNCVALUE1 }, // Was 0x2D
        /* 17 0x30 */ { REG_SYNCVALUE2, RFM69_SYNCVALUE2 }, // Was init.NETWORKID
        /* 18 0x37 */ { REG_PACKETCONFIG1,
                        RF_PACKET1_FORMAT_VARIABLE bitor // So first byte with FIFO num bytes following
                        RF_PACKET1_DCFREE_OFF bitor // So neither Manchester Encoding nor Data whitening (so radio "bits" may have a DC component)
                        RF_PACKET1_CRC_ON bitor
                        #if (SEMANTICS_DO_CRC_ERR_NO_IRQ == 1)
                            RF_PACKET1_CRCAUTOCLEAR_ON bitor // Default and RFM69
                        #else
                            RF_PACKET1_CRCAUTOCLEAR_OFF bitor // Now I can do statistiscs with messageRadioCRC16Err_IRQ
                        #endif
                        RF_PACKET1_ADRSFILTERING_OFF }, // Allows "doListenToAll" to be built on top of this
        /* 19 0x38 */ { REG_PAYLOADLENGTH, RF69_MAX_FIFO_LEN_65 }, // in variable length mode: the max frame size, not used in TX
        // .. 0x39    { REG_NODEADRS, init.nodeID }, // turned off because we're not using address filtering. But settable in setAddres
        //              TX on FIFO not empty (So RF_FIFOTHRESH_VALUE FifoThreshold not used):
        /* 20 0x3C */ { REG_FIFOTHRESH, RF_FIFOTHRESH_TXSTART_FIFONOTEMPTY bitor RF_FIFOTHRESH_VALUE },
        /* 21 0x3D */ { REG_PACKETCONFIG2,
                        RF_PACKET2_RXRESTARTDELAY_2BITS bitor
                        RF_PACKET2_AUTORXRESTART_ON bitor
                        RF_PACKET2_AES_OFF }, // RXRESTARTDELAY must match transmitter PA ramp-down time (bitrate dependent)
        // .. 0x3D    { REG_PACKETCONFIG2,
        //              RF_PACKET2_RXRESTARTDELAY_NONE bitor
        //              RF_PACKET2_AUTORXRESTART_ON bitor
        //              RF_PACKET2_AES_OFF },
        //              for BR-19200. RXRESTARTDELAY must match transmitter PA ramp-down time (bitrate dependent)
        /* 22 0x6F */ { REG_TESTDAGC, RF_DAGC_IMPROVED_LOWBETA0 }
        //              run DAGC continuously in RX mode for Fading Margin Improvement, recommended default for AfcLowBetaOn=0
        // 22 == NUM_REGISTERS_INIT
    };

    // Used above:
    // init.freqBand
    rfm69_internals.SENDERID        = init.nodeID;
    rfm69_internals.nodeID          = init.nodeID; // No REG_NODEADRS so not set above, use setAddres to tell SX1231H
    rfm69_internals.isRadio_RFM69HW = init.isRFM69HW;

    // RFM69 had a write to REG_SYNCVALUE1 of 0xAA and readback until 0xAA then the same for 0x55 here.
    // Don't know why, if it were not to synch before CONFIG[][] write? TODO find out?
    // Removed 2018 04 28 then while_writeReg_readRegExpected also went

    for (uint8_t iof_config = 0; iof_config < NUM_REGISTERS_INIT; iof_config++) {
        reg_value_readback[iof_config] = SPI_WRITE_REG (CONFIG[iof_config][0], CONFIG[iof_config][1]);
    }

    // Encryption is persistent between resets and can trip you up during debugging.
    // Disable it during initialization so we always start from a known state.
    func_encrypt16 ("0", 0, i_spi, spi_params, rfm69_internals); // setMode to RF69_MODE_STANDBY

    setHighPower (rfm69_internals.isRadio_RFM69HW, i_spi, spi_params, rfm69_internals); // called regardless if it's a RFM69W or RFM69HW

    setMode (12, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals); // setMode to RF69_MODE_STANDBY since we don't know about what's going on in setHighPower

    reg = while_readRegExpectedMask (1,             // "001-B loop". discarding reg (Never seen it loop:)
            REG_IRQFLAGS1, RF_IRQFLAGS1_MODEREADY,  // readFromReg, readFromRegMask
            logic_normal,                           // loop until bit in flag becomes 1
            WAIT_FOR_REGISTER_VALUE_MS,             // timeout_ms
            ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_1,  // error_bitnum
            i_spi, spi_params, rfm69_internals);

    debug_print("%s", "-- IRQ func_initialize\n");
} // func_initialize


// Get the Received Signal Strength Indication (RSSI)
// INTERNAL FUNCTION
int16_t
readRSSI_dBm (
        const forceTrigger_t forceTrigger,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,19);

    debug_print("%s", "readRSSI_dBm\n");
    int16_t return_rssi;
    if (forceTrigger == FORCETRIGGER_ON) { // This seems to take much longer time
        // RSSI trigger not needed if DAGC is in continuous mode
        SPI_WRITE_REG (REG_RSSICONFIG, RF_RSSI_START);

        // wait for RSSI_Ready
        // In the orignal code there was no timeout here
        uint8_t reg;
        reg = while_readRegExpectedMask (2,   // "003-B loop". discarding reg (Never seen it loop:)
                REG_RSSICONFIG, RF_RSSI_DONE, // readFromReg, readFromRegMask
                logic_normal,                 // loop until bit in flag becomes 1
                WAIT_FOR_REGISTER_VALUE_MS,   // timeout_ms. In the original there was no timeout here
                ERROR_BITNUM_REG_RSSICONFIG,  // error_bitnum
                i_spi, spi_params, rfm69_internals);
    } else {}

    return_rssi = (int16_t) (SPI_READ_REG (REG_RSSIVALUE));
    //
    // From SX1231H Manual:
    //     Absolute value of the RSSI in dBm, 0.5dB steps. RSSI = -RssiValue/2 [dBm]
    //     From High Sensitivity: down to -120 dBm at 1.2 kbps
    //     115 dB+ Dynamic Range RSSI
    //
    //     3.4.9. RSSI
    //         The RSSI block evaluates the amount of energy available within the receiver channel bandwidth.
    //         Its resolution is 0.5 dB, and it has a wide dynamic range to accommodate both small and large
    //         signal levels that may be present. Its acquisition time is very short, taking only 2 bit periods.
    //         The RSSI sampling must occur during the reception of preamble in FSK (RF_DATAMODUL_MODULATIONTYPE_FSK
    //         default), and constant 1 reception in OOK (RF_DATAMODUL_MODULATIONTYPE_OOK not here).
    return_rssi = REG_RSSI_TO_DBM (return_rssi);

    debug_print ("%s", "-- readRSSI_dBm\n");
    return return_rssi;
} // readRSSI_dBm


// INTERNAL FUNCTION
int8_t
readTemperature_degC (
        const int8_t         calOffset,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,18);
    debug_print("%s", "readTemperature_degC\n");
    signed  return_temperature_degC;
    uint8_t oldmode = rfm69_internals.radio_mode; // Idea from radioHead

    if (rfm69_internals.radio_mode != RF69_MODE_STANDBY) {
        setMode (13, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals); // May update rfm69_internals.radio_mode
    } else {}

    // REG_TEMP1 = RegTemp1 Temperature Sensor control
    // BIT3: Triggers the temperature measurement when set. Always reads 0.
    //
    SPI_WRITE_REG (REG_TEMP1, RF_TEMP1_MEAS_START);

    // BIT2: RF_TEMP1_MEAS_RUNNING TempMeasRunning
    // Set to 1 while the temperature measurement is running.
    // Toggles back to 0 when the measurement has completed. The receiver can not be used while measuring temperature
    // It takes less than 100 microseconds for the RFM69HCW to evaluate the temperature
    //
    uint8_t reg;
    reg = while_readRegExpectedMask (3,          // "003-B loop". discarding reg (Never seen it loop:)
            REG_TEMP1, RF_TEMP1_MEAS_RUNNING,    // readFromReg, readFromRegMask
            logic_inverted,                      // loop until bit in flag becomes 0
            WAIT_FOR_REGISTER_TEMP1_MEAS_MS,     // timeout_ms. In the original there was no timeout here
            ERROR_BITNUM_RF_TEMP1_MEAS_RUNNING,  // error_bitnum
            i_spi, spi_params, rfm69_internals);

    // REG_TEMP2 = RegTemp2 Temperature readout (reset-value 0x00)
    //             RegTemp2 Value is -40 to +85 with -1 degC per LSB

                                             // CORRESPONDS ABOUT TO 21 DEGC ON MY DESK:
    uint8_t temp2  = SPI_READ_REG (REG_TEMP2);    // 0x90 (0xFFFFFF90) is bin 10010000 is dec -112
    int8_t  temp2c = (int8_t) (compl temp2); // 0x6F (0x0000006F) is bin 01101111 is dec  111
                                             // compl == one's complement, negation == one's complement + 1 (has sign)
                                             // RadioHead uses compl (so result is one less than negation)

    //                                   111    +       (-90)           + 0  = 21 DEGC
    return_temperature_degC = (int8_t) (temp2c + COURSE_TEMP_OFFSET + calOffset);

    debug_print("temp2(0x%02X) temp2c(0x%02X) is %d degC\n", temp2, temp2c, return_temperature_degC);
    // 'complement' corrects the slope, rising temp = rising val
    //  COURSE_TEMP_OFFSET puts reading in the ballpark, user can add additional correction

    if (rfm69_internals.radio_mode != oldmode) {
        setMode (14, oldmode, i_spi, spi_params, rfm69_internals);
    } else {}

    debug_print("%s", "-- readTemperature_degC\n");

    return return_temperature_degC;
}

// INTERNAL FUNCTION
void
receiveBegin (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,17);
    debug_print("%s", "receiveBegin ---- CLR! ---\n");
    unsigned value = 0; // For printf only

    rfm69_internals.APPPROTLEN = 0;
    rfm69_internals.SENDERID = 0;
    rfm69_internals.TARGETID = 0;
    rfm69_internals.PACKETLEN = 0;

    // TODO This block is probably not necessary (studied and rewritten 30Oct2018)
    uint8_t RegPacketConfig2 = SPI_READ_REG (REG_PACKETCONFIG2);
    if ((RegPacketConfig2 bitand RF_PACKET2_AUTORXRESTART_ON) == 0) { // This was not originally tested.
        // See Datasheet for RegPacketConfig2 AutoRxRestartOn: 0 Off -> RestartRx can be used.
        value = 1;
        if ((SPI_READ_REG (REG_IRQFLAGS2) bitand RF_IRQFLAGS2_PAYLOADREADY) != 0) { // IRQ is high!
            SPI_WRITE_REG (REG_PACKETCONFIG2, RegPacketConfig2 bitor RF_PACKET2_RXRESTART); // (RestartRx). Avoid RX deadlocks(?)
            value = 2;
        } else {}
    } else {}

    SPI_WRITE_REG (REG_DIOMAPPING1, RF_DIOMAPPING1_DIO0_01); // set DIO0 to "PayloadReady" in receive mode.
    // Opens for possible interrupt. If RX then RX ready, it TX then TX ready
    // See Table 22 DIO Mapping, Packet Mode

    setMode (15, RF69_MODE_RX, i_spi, spi_params, rfm69_internals); // TODO why is this not before the REG_DIOMAPPING1 write above?
    debug_print_x ("-- receiveBegin did %u\n", value);
} // receiveBegin

void
_receiveBegin (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,17);
    debug_print("%s", "receiveBegin ---- CLR! ---\n");
    unsigned value = 0; // For printf only

    rfm69_internals.APPPROTLEN = 0;
    rfm69_internals.SENDERID = 0;
    rfm69_internals.TARGETID = 0;
    rfm69_internals.PACKETLEN = 0;

    // TODO This block is probably not necessary (studied and rewritten 30Oct2018)
    VALUE_XSCOPE_LIB (RFM69_VALUE,3000);
    uint8_t RegPacketConfig2 = SPI_READ_REG (REG_PACKETCONFIG2);
    if ((RegPacketConfig2 bitand RF_PACKET2_AUTORXRESTART_ON) == 0) { // This was not originally tested.
        // See Datasheet for RegPacketConfig2 AutoRxRestartOn: 0 Off -> RestartRx can be used.
        value = 1;
        if ((SPI_READ_REG (REG_IRQFLAGS2) bitand RF_IRQFLAGS2_PAYLOADREADY) != 0) { // IRQ is high!
            VALUE_XSCOPE_LIB (RFM69_VALUE,3001);
            SPI_WRITE_REG (REG_PACKETCONFIG2, RegPacketConfig2 bitor RF_PACKET2_RXRESTART); // (RestartRx). Avoid RX deadlocks(?)
            value = 2;
        } else {
            VALUE_XSCOPE_LIB (RFM69_VALUE,3002);
        }
    } else {
        VALUE_XSCOPE_LIB (RFM69_VALUE,3003);
    }

    SPI_WRITE_REG (REG_DIOMAPPING1, RF_DIOMAPPING1_DIO0_01); // set DIO0 to "PayloadReady" in receive mode.
    // Opens for possible interrupt. If RX then RX ready, it TX then TX ready
    // See Table 22 DIO Mapping, Packet Mode
    VALUE_XSCOPE_LIB (RFM69_VALUE,3004);
    // AQU=065 comment out this line removes the error: (when setMode)
    _setMode (15, RF69_MODE_RX, i_spi, spi_params, rfm69_internals); // TODO why is this not before the REG_DIOMAPPING1 write above?
    VALUE_XSCOPE_LIB (RFM69_VALUE,3005);
    debug_print_x ("-- receiveBegin did %u\n", value);
} // receiveBegin


// checks if a packet was received and/or puts transceiver in receive (ie RX or listen) mode
// INTERNAL FUNCTION
bool
func_receiveDone (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    unsigned receiveDone_case;
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,16);
    debug_print("%s", "func_receiveDone\n");

    if ((rfm69_internals.radio_mode == RF69_MODE_RX) and (rfm69_internals.PACKETLEN > 0)) {
        // After every received packet

        // RFM69 had a setMode (RF69_MODE_STANDBY) here, but it required another call to this i_radio.uspi_receiveDone()
        // to do the setMode (RF69_MODE_RX) in receiveBegin below. This seems to take mode into RF69_MODE_STANDBY
        // as a temporary state only. This does not seem neceessary, it looks like the FIFO is emptied anyhow
        // Commit 359618a (did not find it 29Oct2018, but found the code so there is a commit 29=ct2018 with this code being present)
        // was the last version with this semantics,
        // controlled with SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY (2018 05 21 code)
        // and SEMANTICS_DO_INTERMEDIATE_RECEIVEDONE there (2018 05 21 code)

        // UPDATE 29Oct2018: RFM69=001
        // The problem with "IRQ 7 (messagePacketLenErr_IRQ) fault is that the IRQ line is kept high with REG_IRQFLAGS2 0x66,
        // and that the transition RF69_MODE_RX->RF69_MODE_STANDBY will not clear the FIFO, but RF69_MODE_STANDBY->RF69_MODE_RX
        // will clear the FIFO! So the original code makes sense!
        // Reinserting code from 2018 05 21 29Oct2018:
        #if (SEMANTICS_DO_INTERMEDIATE_RF69_MODE_STANDBY == 1) // TODO remove
            setMode (16, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals); // enables interrupts
            // Is this because "Another important feature is ability to fill and empty the FIFO in Sleep/Stdby mode,
            // ensuring optimum power consumption and adding more flexibility for the software."
        #endif

        receiveDone_case = 1; // returns true
    } else if (rfm69_internals.radio_mode == RF69_MODE_RX) { // already in RX no full_payload_01 yet
        // This did not hit the restore after silent for 10 seconds in case of _Aquarium_rfm69_client, because the restore hit the above case
        // Also, with file "2018 10 04 A...txt" we see two "receiveBegin" here, so this is not necessary:
        // receiveBegin (i_spi, spi_params, rfm69_internals); // Does an setMode (RF69_MODE_RX)

        receiveDone_case = 2; // returns false
    } else {
        // Initially once
        receiveBegin (i_spi, spi_params, rfm69_internals); // Does an setMode (RF69_MODE_RX)
        receiveDone_case = 3; // returns false
    }

    debug_print_x ("-- func_receiveDone did %u\n", receiveDone_case);

    return (receiveDone_case == 1);
} // func_receiveDone


void
func_send (
        const uint8_t        TARGETID_toAddress, // GATEWAYID
        const packet_t       PACKET,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,15);
    debug_print("%s", "func_send\n");

    uint8_t RegPacketConfig2 = SPI_READ_REG (REG_PACKETCONFIG2); // RF_PACKET2_RXRESTART (RestartRx) bit is per def 0 when read
    SPI_WRITE_REG (REG_PACKETCONFIG2, RegPacketConfig2 bitor RF_PACKET2_RXRESTART); // (RestartRx). Avoids RX deadlocks

    while_not_canSend_do_receiveDone (4,             // Only usage of this function
                                                     // Largest loop: 004-A loop 005. Most often it's 001 or zero loops (sending only)
                                                     // In 8.5 days log @ 1 per second: 2:004-A loop 004, 236:004-A loop 003, 4036:004-A loop 002, 64433:004-A loop 001 of 870K messages
                                                     // Exception 004-A loop 021 when RFM69-board unplugged while powered
            WAIT_FOR_CSMA_LIMIT_SET_MS,              // timeout_ms
            ERROR_BITNUM_WAIT_FOR_CSMA_LIMIT_SET_MS, // error_bitnum
            i_spi,
            spi_params,
            rfm69_internals);



    if (rfm69_internals.error_bits == ERROR_BITS_NONE) {
        sendFrame (TARGETID_toAddress, PACKET, i_spi, spi_params, rfm69_internals);
    } else {}

    if (rfm69_internals.error_bits == ERROR_BITS_NONE) {
        rfm69_internals.waitForIRQInterruptCause = messagePacketSent_IRQExpected;
    } else {
        rfm69_internals.waitForIRQInterruptCause = messagePacketNotSent_ErrorIRQNotExpected;
    }

    debug_print("%s", "-- func_send\n");
} // func_send


// INTERNAL FUNCTION
void
sendFrame (
        const uint8_t         TARGETID_toAddress, // GATEWAYID
        const packet_t        PACKET,
        client spi_master_if  i_spi,
        spi_params_t          &spi_params,
        rfm69_internals_t     &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,14); // AQU=065 log 24Feb2019
    debug_print("%s", "sendFrame\n");
    uint8_t  reg;
    setMode (17, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals); // turn off receiver to prevent reception when filling fifo

    reg = while_readRegExpectedMask (5,             // "005-B loop". discarding reg (Never seen it loop:) // AQU=065 log 24Feb2019
            REG_IRQFLAGS1, RF_IRQFLAGS1_MODEREADY,  // readFromReg, readFromRegMask
            logic_normal,                           // loop until bit in flag becomes 1
            WAIT_FOR_REGISTER_VALUE_MS,             // timeout_ms. In the original there was no timeout here
            ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_2,  // error_bitnum
            i_spi, spi_params, rfm69_internals);

    if (rfm69_internals.error_bits == ERROR_BITS_NONE) {
        SPI_WRITE_REG (REG_DIOMAPPING1, RF_DIOMAPPING1_DIO0_00); // DIO0 is "Packet Sent"

        rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN  = PACKET_LEN08;
        rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.TARGETID   = TARGETID_toAddress;
        rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.SENDERID   = rfm69_internals.nodeID;
        rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.padding_03 = PACKET_INIT_VAL08;

        for (unsigned iof_buff = 0; iof_buff < APPNORMPAYLOAD_LEN; iof_buff++) {
            rfm69_internals.PACKET.u.packet_u0.appPayload_uint8_arr[iof_buff] = PACKET.u.packet_u1_uint8_arr[RFM69HEADER_LEN08_IS4 + iof_buff];
        }

        rfm69_internals.PACKET.u.packet_u0.appCRC32 =
                calc_CRC32 (rfm69_internals.PACKET.u.packet_u2_uint32_arr, APPCRC_OVER_LEN32, 0); // First this..
        debug_print ("TX CRC %08X\n", rfm69_internals.PACKET.u.packet_u0.appCRC32);

        rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN -= LENM1_LEN; // ..then this, since first byte to FIFO is num bytes not including first byte..
        // SEND TO RADIO:
        spiBurstWriteSeqRegsOrFIFO (REG_FIFO, rfm69_internals.PACKET.u.packet_u1_uint8_arr, PACKET_LEN08, i_spi, spi_params); // ..for this.. // AQU=065 log 24Feb2019
        //
        rfm69_internals.PACKET.u.packet_u0.CommHeaderRFM69.PACKETLEN += LENM1_LEN; // ..then this (good for printing in client since CRC32 uses it)

        setMode (18, RF69_MODE_TX, i_spi, spi_params, rfm69_internals);
        // IRQ goes high after this, with RF_IRQFLAGS2_PACKETSENT IN

        #if (SEMANTICS_DO_LOOP_FOR_RF_IRQFLAGS2_PACKETSENT == 1)
            reg = while_readRegExpectedMask (6,             // "006-B loop". discarding reg. Largest loop: 006-B loop 028 times. Most often it's 004 times.
                                                            // Each loop is fast, so 028 is not much visible (sending only)
                    REG_IRQFLAGS2, RF_IRQFLAGS2_PACKETSENT, // readFromReg, readFromRegMask
                    logic_normal,                           // loop until bit in flag becomes 1
                    WAIT_FOR_PACKET_SENT_MS,                // timeout_ms
                    ERROR_BITNUM_RF_IRQFLAGS2_MODEREADY,    // error_bitnum
                    i_spi, spi_params, rfm69_internals);
        #else
            // waitForIRQInterruptCause handled by client
            // Now how fast client sends again is important; interlock that with the return return_waitForIRQInterruptCause from func_send
        #endif

    } else {} // IS_ERR

    debug_print("%s", "-- sendFrame\n");

} // sendFrame


// For RFM69HW only: you must call setHighPower(true) after func_initialize() or else transmission won't work
// Also study comments in setPowerLevel
void
setHighPower (
        const bool           isHighPowerOn, // Teig, original "bool onOff"...
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,13);
    debug_print("%s", "setHighPower\n");
    rfm69_internals.isRadio_RFM69HW = isHighPowerOn;
    SPI_WRITE_REG (REG_OCP, rfm69_internals.isRadio_RFM69HW ? RF_OCP_ON : RF_OCP_OFF); // ... had true as off!
    if (rfm69_internals.isRadio_RFM69HW) { // turning ON
        SPI_WRITE_REG (REG_PALEVEL, (SPI_READ_REG (REG_PALEVEL) bitand RF_PALEVEL_OUTPUTPOWER_MASK) bitor RF_PALEVEL_PA1_ON bitor RF_PALEVEL_PA2_ON);
       // enable P1 & P2 amplifier stages. The powerLevel bits are kept
    } else {
        SPI_WRITE_REG (REG_PALEVEL, RF_PALEVEL_PA0_ON bitor RF_PALEVEL_PA1_OFF bitor RF_PALEVEL_PA2_OFF bitor rfm69_internals.powerLevel_dBm); // enable P0 only
    }

    debug_print("%s", "-- setHighPower\n");
} // setHighPower


// INTERNAL FUNCTION
// Also study comments in setPowerLevel
void
setHighPowerRegs (
        const bool           isHighPowerOn, // Teig, original "bool onOff"...
        client spi_master_if i_spi,
        spi_params_t         &spi_params)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,12);
    debug_print("%s", "setHighPowerRegs\n");
    SPI_WRITE_REG (REG_TESTPA1_H, isHighPowerOn ? RF_TESTPA1_H_HIGH_POWER : RF_TESTPA1_H_RX_OR_PA0);
    SPI_WRITE_REG (REG_TESTPA2_H, isHighPowerOn ? RF_TESTPA2_H_HIGH_POWER : RF_TESTPA2_H_RX_OR_PA0);
    debug_print("%s", "-- setHighPowerRegs\n");
} // setHighPowerRegs


// INTERNAL FUNCTION
// Observe that REG_OPMODE must have RF_OPMODE_SEQUENCER_OFF set for this to work!
// Observe that RF_OPMODE_LISTEN_ON in REG_OPMODE is never set! TODO

// As related to messageAppCRC32Err_IRQ
//   1. Observe that ListenMode with RF_OPMODE_LISTEN_ON set in REG_OPMODE (RegOpMode) has not been investigated.
//      None of the libraries I have ported from have related to ListenMode either. It's basically
//      meant to be used to save power, wich is not a big issue for these cases I guess
//   2. Also observe that REG_AUTOMODES (RegAutoModes) has not been set to anything else than with the default values
//      RF_AUTOMODES_ENTER_OFF,  RF_AUTOMODES_EXIT_OFF and RF_AUTOMODES_INTERMEDIATE_SLEEP
//      None of the libraries I have ported from have related to AutoModes either.
//   3. In other words, I don't know if trying out the above would cause messageAppCRC32Err_IRQ never to happen

void
setMode (
        const unsigned       debug_caller_id,
        const uint8_t        newMode,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,11);
    const uint8_t oldMode = rfm69_internals.radio_mode;

    // Cost 708 bytes with -O2 31496
    debug_print
        ("setMode (%u) from (%s) to (%s)\n",
            debug_caller_id,
            (oldMode == RF69_MODE_VOID)    ? "RF69_MODE_VOID"    :
            (oldMode == RF69_MODE_SLEEP)   ? "RF69_MODE_SLEEP"   :
            (oldMode == RF69_MODE_STANDBY) ? "RF69_MODE_STANDBY" :
            (oldMode == RF69_MODE_SYNTH)   ? "RF69_MODE_SYNTH"   :
            (oldMode == RF69_MODE_TX)      ? "RF69_MODE_TX"      :
            (oldMode == RF69_MODE_RX)      ? "RF69_MODE_RX"      : "?",
            (newMode == RF69_MODE_VOID)    ? "RF69_MODE_VOID?"   :
            (newMode == RF69_MODE_SLEEP)   ? "RF69_MODE_SLEEP"   :
            (newMode == RF69_MODE_STANDBY) ? "RF69_MODE_STANDBY" :
            (newMode == RF69_MODE_SYNTH)   ? "RF69_MODE_SYNTH"   :
            (newMode == RF69_MODE_TX)      ? "RF69_MODE_TX"      :
            (newMode == RF69_MODE_RX)      ? "RF69_MODE_RX"      : "?"
        );

    if (newMode == rfm69_internals.radio_mode) {
        debug_print ("%s\n", "-- setMode same, not needed");
    } else {

        bool error = false;
        spi_byte_t reg;
        //
        reg = SPI_READ_REG(REG_OPMODE);                     // Read first (to keep the other bits, Teig assumes)
        reg = reg bitand (compl RF_OPMODE_MODE_MASK);       // clear the 3 bits inside, keep the others
        reg = reg bitor (newMode << RF_OPMODE_MODE_BITPOS); // insert new 3 bits instead
        reg = reg bitor RF_OPMODE_SEQUENCER_OFF;            // Not necessary since it's set already set in func_initialize, but in general it's necessary (Teig added this)
                                                            // Controls the automatic Sequencer (see section 4.2 ): 0
                                                            // BIT7: 1 -> Mode is forced by the user (RF_OPMODE_SEQUENCER_OFF)
        SPI_WRITE_REG (REG_OPMODE, reg);

        if (reg != SPI_READ_REG(REG_OPMODE)) {
            rfm69_internals.error_bits or_eq (1 << ERROR_BITNUM_SETMODE_FAILED);
            error = true;
            debug_print ("setMode to (%u) failed\n", newMode);
        } else {
            switch (newMode) {
                case RF69_MODE_TX: {
                    if (rfm69_internals.isRadio_RFM69HW) setHighPowerRegs (true, i_spi, spi_params);
                } break;
                case RF69_MODE_RX: {
                    if (rfm69_internals.isRadio_RFM69HW) setHighPowerRegs (false, i_spi, spi_params);
                } break;
                default: {
                    // No code
                    // Teig no "return" any more since we cover more cases
                } break;
            }

            if (rfm69_internals.radio_mode == RF69_MODE_SLEEP) {
                debug_print ("%s\n", "was RF69_MODE_SLEEP");
                // The original code has "rfm69_internals.radio_mode==RF69_MODE_SLEEP" in the loop, how it will ever get out of because of the
                // "not RF69_MODE_SLEEP anymore" it I don't see. It probably is a coding error and SPI_READ_REG will take it instead (so SPI has to work!)
                // So I just add an error message here (ERROR_BITNUM_NO_RF69_MODE_SLEEP_EXIT) to see if I can catch it

                // We are using packet mode, so this check is not really needed
                // But waiting for mode ready is necessary when going from sleep because the FIFO may not be immediately available from previous mode
                reg = while_readRegExpectedMask (7,           // "007-B loop" discarding reg (Never seen it loop:)
                       REG_IRQFLAGS1, RF_IRQFLAGS1_MODEREADY, // readFromReg, readFromRegMask
                       logic_normal,                          // loop until bit in flag becomes 1
                       WAIT_FOR_SLEEP_MODE_EXIT_MS,           // timeout_ms. In the original there was no timeout here
                       ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_3, // error_bitnum (only sets it, clearing is by getAndClearErrorBits)
                       i_spi, spi_params, rfm69_internals);

                if ((rfm69_internals.error_bits bitand (1 << ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_3)) != 0) {
                   error = true;
                   rfm69_internals.error_bits or_eq (1 << ERROR_BITNUM_NO_RF69_MODE_SLEEP_EXIT);
                } else {}
            } else {}
        }

        if (error) {
            PRINT_ADD_ERROR_BITS_IFF ("setMode error", rfm69_internals.error_bits);
        } else {
            rfm69_internals.radio_mode = newMode; // This is the only place we set it
        }
        debug_print("%s", "-- setMode\n");
    }
} // setMode

// INTERNAL FUNCTION
void
_setMode (
        const unsigned       debug_caller_id,
        const uint8_t        newMode,
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    VALUE_XSCOPE_LIB (RFM69_VALUE,10001); // AQU=065
    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,11);
    const uint8_t oldMode = rfm69_internals.radio_mode;

    // Cost 708 bytes with -O2 31496
    debug_print
        ("setMode (%u) from (%s) to (%s)\n",
            debug_caller_id,
            (oldMode == RF69_MODE_VOID)    ? "RF69_MODE_VOID"    :
            (oldMode == RF69_MODE_SLEEP)   ? "RF69_MODE_SLEEP"   :
            (oldMode == RF69_MODE_STANDBY) ? "RF69_MODE_STANDBY" :
            (oldMode == RF69_MODE_SYNTH)   ? "RF69_MODE_SYNTH"   :
            (oldMode == RF69_MODE_TX)      ? "RF69_MODE_TX"      :
            (oldMode == RF69_MODE_RX)      ? "RF69_MODE_RX"      : "?",
            (newMode == RF69_MODE_VOID)    ? "RF69_MODE_VOID?"   :
            (newMode == RF69_MODE_SLEEP)   ? "RF69_MODE_SLEEP"   :
            (newMode == RF69_MODE_STANDBY) ? "RF69_MODE_STANDBY" :
            (newMode == RF69_MODE_SYNTH)   ? "RF69_MODE_SYNTH"   :
            (newMode == RF69_MODE_TX)      ? "RF69_MODE_TX"      :
            (newMode == RF69_MODE_RX)      ? "RF69_MODE_RX"      : "?"
        );

    if (newMode == rfm69_internals.radio_mode) {
        VALUE_XSCOPE_LIB (RFM69_VALUE,10002);
        debug_print ("%s\n", "-- setMode same, not needed");
    } else {
        VALUE_XSCOPE_LIB (RFM69_VALUE,10003); // AQU=065
        bool error = false;
        spi_byte_t reg;
        //
        reg = SPI_READ_REG(REG_OPMODE);                     // Read first (to keep the other bits, Teig assumes)

        // AQU=065f comment out these lines removes the error qwe
        // if     SPI_MASTER_POS 1
        // But if SPI_MASTER_POS 2 it also works with this code!

        #if (TEST_STRETCH_IRQ_LIB==1)
            // No code
            // Was TEST_19MAR2019==0
            // see file ..SDS00060.png IRQ now has a period of 4 seconds, but is low 2 and high 2 secs. This could explain the different behaviour,
            // that the IRQ does not deadlock since during sending the IRQ stays high!?
        #else
            reg = reg bitand (compl RF_OPMODE_MODE_MASK);       // clear the 3 bits inside, keep the others
            reg = reg bitor (newMode << RF_OPMODE_MODE_BITPOS); // insert new 3 bits instead
            reg = reg bitor RF_OPMODE_SEQUENCER_OFF;            // Not necessary since it's set already set in func_initialize, but in general it's necessary (Teig added this)
                                                                // Controls the automatic Sequencer (see section 4.2 ): 0
                                                                // BIT7: 1 -> Mode is forced by the user (RF_OPMODE_SEQUENCER_OFF)
            // VALUE_XSCOPE_LIB (RFM69_VALUE,10005); // AQU=065
        #endif

        SPI_WRITE_REG (REG_OPMODE, reg); // AQU=065f
        VALUE_XSCOPE_LIB (RFM69_VALUE,10004); // AQU=065

        if (reg != SPI_READ_REG(REG_OPMODE)) {
            VALUE_XSCOPE_LIB (RFM69_VALUE,10006);
            rfm69_internals.error_bits or_eq (1 << ERROR_BITNUM_SETMODE_FAILED);
            error = true;
            debug_print ("setMode to (%u) failed\n", newMode);
        } else {
            VALUE_XSCOPE_LIB (RFM69_VALUE,10007); // AQU=065
            switch (newMode) {
                case RF69_MODE_TX: {
                    if (rfm69_internals.isRadio_RFM69HW) {
                        VALUE_XSCOPE_LIB (RFM69_VALUE,10008);
                        setHighPowerRegs (true, i_spi, spi_params);
                    }
                } break;
                case RF69_MODE_RX: {
                    if (rfm69_internals.isRadio_RFM69HW) {
                        VALUE_XSCOPE_LIB (RFM69_VALUE,10009);
                        setHighPowerRegs (false, i_spi, spi_params);
                    }
                } break;
                default: {
                    // No code
                    // Teig no "return" any more since we cover more cases
                    VALUE_XSCOPE_LIB (RFM69_VALUE,10010); // AQU=065 RF69_MODE_STANDBY
                } break;
            }

            if (rfm69_internals.radio_mode == RF69_MODE_SLEEP) {
                VALUE_XSCOPE_LIB (RFM69_VALUE,10011);
                debug_print ("%s\n", "was RF69_MODE_SLEEP");
                // The original code has "rfm69_internals.radio_mode==RF69_MODE_SLEEP" in the loop, how it will ever get out of because of the
                // "not RF69_MODE_SLEEP anymore" it I don't see. It probably is a coding error and SPI_READ_REG will take it instead (so SPI has to work!)
                // So I just add an error message here (ERROR_BITNUM_NO_RF69_MODE_SLEEP_EXIT) to see if I can catch it

                // We are using packet mode, so this check is not really needed
                // But waiting for mode ready is necessary when going from sleep because the FIFO may not be immediately available from previous mode
                reg = while_readRegExpectedMask (7,           // "007-B loop" discarding reg (Never seen it loop:)
                       REG_IRQFLAGS1, RF_IRQFLAGS1_MODEREADY, // readFromReg, readFromRegMask
                       logic_normal,                          // loop until bit in flag becomes 1
                       WAIT_FOR_SLEEP_MODE_EXIT_MS,           // timeout_ms. In the original there was no timeout here
                       ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_3, // error_bitnum (only sets it, clearing is by getAndClearErrorBits)
                       i_spi, spi_params, rfm69_internals);

                if ((rfm69_internals.error_bits bitand (1 << ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY_3)) != 0) {
                   error = true;
                   rfm69_internals.error_bits or_eq (1 << ERROR_BITNUM_NO_RF69_MODE_SLEEP_EXIT);
                } else {}
            } else {
                VALUE_XSCOPE_LIB (RFM69_VALUE,10012); // AQU=065
            }
        }

        if (error) {
            PRINT_ADD_ERROR_BITS_IFF ("setMode error", rfm69_internals.error_bits);
        } else {
            rfm69_internals.radio_mode = newMode; // This is the only place we set it
        }
        debug_print("%s", "-- setMode\n");
        VALUE_XSCOPE_LIB (RFM69_VALUE,10013); // AQU=065
    }
} // _setMode


// INTERNAL FUNCTION
//
// EVEN IF THIS FUNCTION HAS AN OUTER LOOP TO CATCH A RADIO THAT WOULD NOT RESPOND IN TIME,
// THEN ANY FOREVER-BLOCKING SPI CALL WOULD STILL BE FOREVER BLOCKING. See AQU=065 in the aquarium code.
// I have used the synchronous version of lib_spi (spi_master or spi_master_2 which is [[distributable]]).
// Asynch version (spi_master_async which is [[combinable]]) of the SPI might not necessarily help solving
// the problem, but it might isolate the problem easier. To isolate the problem then I guess make the
// protocol with this library asynchronous would be better. I will work with this and call it RFM69=009
// This comment also goes for while_readRegExpectedMask (below)
//
void
while_not_canSend_do_receiveDone (
        const unsigned       debug_caller_id,
        const unsigned       timeout_ms,   // Like WAIT_FOR_CSMA_LIMIT_SET_MS
        const unsigned       error_bitnum, // like ERROR_BITNUM_CAN_SEND
        client spi_master_if i_spi,        // For SPI_WRITE_REG and SPI_READ_REG macros
        spi_params_t         &spi_params,  // --"--
        rfm69_internals_t    &rfm69_internals)
{
    timer      tmr;
    time32_t   now_time_ticks;
    time32_t   start_time_ticks;
    time32_t   timeout_at_ticks;
    bool       timed_out;
    bool       canSend_ok;
    unsigned   loop_cnt = 0;

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,10);
    debug_print ("%s", "while_not_canSend_do_receiveDone\n");
    tmr :> start_time_ticks;
    timeout_at_ticks = start_time_ticks + (timeout_ms * XS1_TIMER_KHZ);

    do {
        canSend_ok = canSend (i_spi, spi_params, rfm69_internals);

        if (not canSend_ok) {
            // Do another round
            tmr :> now_time_ticks;
            timed_out = AFTER_32 (now_time_ticks, timeout_at_ticks);
            if (not timed_out) {
                func_receiveDone (i_spi, spi_params, rfm69_internals);
                loop_cnt++;
                // Do another round
            } else {
                rfm69_internals.error_bits or_eq (1<<error_bitnum);
                debug_print ("%s", "timeout!\n");
            }
        } else {} // Finished

    } while ((not canSend_ok) and (not timed_out));

    if (loop_cnt == 0) {
        debug_print ("%s\n", "-- while_not_canSend_do_receiveDone");
    } else {
        debug_print ("%03u-A loop %03u\n", debug_caller_id, loop_cnt);
    }
}

// INTERNAL FUNCTION
//
// See comment at while_not_canSend_do_receiveDone above

uint8_t // readFromRegValue
while_readRegExpectedMask (
        const unsigned       debug_caller_id,
        const uint8_t        readFromReg,     // Like REG_IRQFLAGS1
        const uint8_t        readFromRegMask, // Like RF_IRQFLAGS1_MODEREADY
        const logic_e        invertedLogic,   // like logic_normal:   loop until bit in flag becomes 1
                                              //      logic_inverted: loop until bit in flag becomes 0
        const unsigned       timeout_ms,      // Like 200
        const unsigned       error_bitnum,    // like ERROR_BITNUM_RF_IRQFLAGS1_MODEREADY
        client spi_master_if i_spi,           // For SPI_WRITE_REG and SPI_READ_REG macros
        spi_params_t         &spi_params,     // --"--
        rfm69_internals_t    &rfm69_internals)
{
    timer      tmr;
    time32_t   now_time_ticks;
    time32_t   start_time_ticks;
    time32_t   timeout_at_ticks;
    bool       timed_out;
    uint8_t    readFromRegValue;
    bool       return_read_ok;
    unsigned   loop_cnt = 0;

    BITOR_ANYPOS_SHARED_GLOBAL(g_radio_log_value,DO_RADIO_BITPOS_LEFT); // "9);"
    debug_print ("%s", "while_readRegExpectedMask\n");
    tmr :> start_time_ticks;
    timeout_at_ticks = start_time_ticks + (timeout_ms * XS1_TIMER_KHZ);

    do {
        readFromRegValue = SPI_READ_REG (readFromReg); // SPI_READ_REG
        return_read_ok   = ((readFromRegValue bitand readFromRegMask) != 0) xor invertedLogic;

        if (not return_read_ok) {
            tmr :> now_time_ticks;
            timed_out = AFTER_32 (now_time_ticks, timeout_at_ticks);
            if (not timed_out) {
                 // Do another round
                loop_cnt++;
            } else {
                rfm69_internals.error_bits or_eq (1<<error_bitnum);
                debug_print ("%s", "timeout!\n");
            }
        } else {} // Finished

    } while ((not return_read_ok) and (not timed_out));

    if (debug_caller_id == 8) {
        if (loop_cnt == 0) {
            debug_print_x ("%s\n", "-- while_readRegExpectedMask");
        } else {
            debug_print_x ("%03u-B loop %03u\n", debug_caller_id, loop_cnt);
        }
    } else {
        if (loop_cnt == 0) {
            debug_print ("%s\n", "-- while_readRegExpectedMask");
        } else {
            debug_print ("%03u-B loop %03u\n", debug_caller_id, loop_cnt);
        }
    }

   return readFromRegValue; // Plus rfm69_internals.error_bits
}
#ifndef DEBUG_SHARED_LOG_VALUE_LIB
    #error DEBUG_SHARED_LOG_VALUE_LIB not defined
#elif (DEBUG_SHARED_LOG_VALUE_LIB==1)

    void clr_radio_log_value (void) {
        INIT_SHARED_GLOBAL(g_radio_log_value,0);
    }

#endif

#if (ALLOW_SESSION_TYPE_TRANS_LIB>=1)

    // THESE FUNCTIONS MUST RUN ON THE CALL STACK OF THE CALLING TASK TO AVOID TASK SWITCHING IF timed_out_trans1to2
    // THEREFORE SEPARATE FUNCTIONS

    // ------------------------------------------------------------
    //  For those interface functions that DO NOT start with _uspi
    //  They don't use SPI and need no timeout monitoring
    // ------------------------------------------------------------
    {error_bits_e, is_error_e} getAndClearErrorBits_iff (const bool timed_out_trans1to2, client radio_if_t i_radio) {
        error_bits_e error_bits = ERROR_BITS_NONE;
        is_error_e   is_error   = NO_ERR;
        if (not timed_out_trans1to2) {
            {error_bits, is_error} = i_radio.getAndClearErrorBits();
        } else {}
        return {error_bits, is_error};
    }

    void do_spi_aux_pin_iff  (const bool timed_out_trans1to2, client radio_if_t i_radio, const unsigned maskof_pin, const pin_e value) {
        if (not timed_out_trans1to2) {
            i_radio.do_spi_aux_pin (maskof_pin, value);
        } else {}
    }

    void setListenToAll_iff (const bool timed_out_trans1to2, client radio_if_t i_radio, const bool doListenToAll) {
        if (not timed_out_trans1to2) {
            i_radio.setListenToAll (doListenToAll);
        } else {}
    }

    uint8_t setNODEID_iff (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t newNODEID) {
        uint8_t return_previous_NODEID = 0;
        if (not timed_out_trans1to2) {
            return_previous_NODEID = i_radio.setNODEID (newNODEID);
        } else {}
        return return_previous_NODEID;
    }

    // ---------------------------------------------------
    // For those interface functions that start with _uspi
    // They use SPI and do need _transx timeout monitoring
    // ---------------------------------------------------

    time32_t do_aux_adafruit_rfm69hcw_RST_pulse_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const unsigned maskof_pin) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.do_aux_adafruit_rfm69hcw_RST_pulse_trans1 (maskof_pin);
        } else {}
        return start_time;
    }

    time32_t initialize_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const rfm69_params_t init) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.initialize_trans1 (init);
        } else {}
        return start_time;
    }

    time32_t getDeviceType_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.getDeviceType_trans1 ();
        } else {}
        return start_time;
    }

    #if (TRANS_ASYNCH_WRAPPED_LIB==0)
        time32_t send_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t TARGETID_toAddress, const packet_t PACKET) {
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.send_trans1 (TARGETID_toAddress, PACKET);
            } else {}
            return start_time;
        }

        time32_t readRSSI_dBm_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const forceTrigger_t forceTrigger) {
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.readRSSI_dBm_trans1 (forceTrigger);
            } else {}
            return start_time;
        }

        time32_t handleSPIInterrupt_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio) {
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.handleSPIInterrupt_trans1();
            } else {}
            return start_time;
        }

        time32_t receiveDone_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio) {
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.receiveDone_trans1();
            } else {}
            return start_time;
        }

        time32_t ultimateIRQclear_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio) {
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.ultimateIRQclear_trans1();
            } else {}
            return start_time;
        }

        time32_t encrypt16_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const char key[KEY_LEN]) {
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.encrypt16_trans1 (key);
            } else {}
            return start_time;
        }
    #endif

    time32_t rcCalibration_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.rcCalibration_trans1();
        } else {}
        return start_time;
    }

    time32_t readTemperature_degC_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const int8_t calOffset_degC) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.readTemperature_degC_trans1(calOffset_degC);
        } else {}
        return start_time;
    }

    time32_t setHighPower_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const bool isHighPowerOn) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.setHighPower_trans1(isHighPowerOn);
        } else {}
        return start_time;
    }

    time32_t setMode_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t newMode) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.setMode_trans1(newMode);
        } else {}
        return start_time;
    }

    time32_t setPowerLevel_dBm_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint8_t powerLevel_dBm) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.setPowerLevel_dBm_trans1(powerLevel_dBm);
        } else {}
        return start_time;
    }

    time32_t setFrequencyRegister_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const uint32_t setFrequencyRegister_trans1) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.setFrequencyRegister_trans1(setFrequencyRegister_trans1);
        } else {}
        return start_time;
    }

    time32_t getFrequencyRegister_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio) {
        time32_t start_time;
        if (not timed_out_trans1to2) {
            start_time = i_radio.getFrequencyRegister_trans1();
        } else {}
        return start_time;
    }

    #if (GET_RADIO_DEBUG_REGS_LIB==1)
        time32_t getDebug_iff_trans1 (const bool timed_out_trans1to2, client radio_if_t i_radio, const debug_state_e debug_state) { // data indexed with iof_debug_bytes_e
            time32_t start_time;
            if (not timed_out_trans1to2) {
                start_time = i_radio.getDebug_trans1(debug_state);
            } else {}
            return start_time;
        }
    #endif

    // Contains a nested select: a task calling this cannot be [[combinable]] (and then also not [[distributable]])
    // This the caller has to have a core on its own
    //
    // The alternative to this is to lay this code out "one level up", and introduce a state machine for it there.
    // This would allow for a [[COMBINABLE]] task and the sequential waiting done here would not block for other
    // actions. However, in a typical top level this "blocking sub-thread" probably is no problem.
    // Another matter is that repeated calls on i_radio would, if the code was moved one level up, have to be
    // stopped until the previous one was finished, introducing even another state. So this solution here
    // is all over the simplest
    //
    void do_sessions_trans2to3 (
            client radio_if_t       i_radio,
                   timing_transx_t  &timing_transx,
                   return_trans3_t  &return_trans3)
    {
        if (not timing_transx.timed_out_trans1to2) {
            timer    tmr;
            time32_t timeout;
            time32_t start_time_trans1;
            time32_t end_time_trans1;

            tmr :> timeout;
            timeout += (timing_transx.maxtime_allowed_ms_trans1to2 * XS1_TIMER_KHZ);

            select {
                case i_radio.session_trans2() : { // notification or signal or ready. If RFM69 register is awaited for forever or SPI hangs then no event here
                    tmr :> end_time_trans1;
                    return_trans3 = i_radio.session_trans3(); // Get data. No SPI. Clears notification
                    #if (DEBUG_SHARED_LOG_VALUE_LIB==1)
                        timing_transx.radio_log_value = INC_RIGHT_SHARED_GLOBAL(g_radio_log_value); // CLR LEFT and CENTER bits and INC RIGHT bits
                    #else
                        timing_transx.radio_log_value++;
                    #endif
                } break;
                case tmr when timerafter(timeout) :> end_time_trans1: {
                    timing_transx.timed_out_trans1to2 = true; // This is pathological. Radio should probably be reset by HW pin, if it's not SPI that hangs
                    // Keep g_radio_log_value!
                    // Also, this would from now on allow all the _trans1 and this do_sessions_trans2to3, but no SPI code would be run
                } break;
            }

            timing_transx.maxtime_used_us_trans1to2 =
                    max (timing_transx.maxtime_used_us_trans1to2, (end_time_trans1 - timing_transx.start_time_trans1)/XS1_TIMER_MHZ);

        } else {
            // No code, don't do it!
        }
    }

    // For those interface functions that start with _uspi
    typedef struct {
        union {
            unsigned       maskof_pin;
            struct {
                uint8_t    TARGETID_toAddress;
                packet_t   PACKET;
            } send;
            forceTrigger_t forceTrigger;
            rfm69_params_t init;
            uint8_t        key[KEY_LEN]; // KEY only, no nul-terminator
            int8_t         calOffset_degC;
            bool           isHighPowerOn;
            uint8_t        newMode;
            uint8_t        powerLevel_dBm;
            uint32_t       register_value;
            #if (GET_RADIO_DEBUG_REGS_LIB==1)
                debug_state_e debug_state;
            #endif
        } u_in;
    } param_for_trans1_t;

    // For those interface functions that start with _uspi
    typedef struct {
        param_for_trans1_t param_for_trans1;
        timer              tmr; // Does not seem to need a separate timer
        time32_t           time;
        bool               do_async_trans2;
         return_trans3_t    return_trans3;
        // NO:             timed_out_trans1to2; This property is set by do_sessions_trans2to3 and must therefore by supplied to _to_ RFM69_driver the calls like send_trans1
    } internal_transx_t;

    #if (ALLOW_SESSION_TYPE_TRANS_LIB==2)
        waitForIRQInterruptCause_e send_iff_asynch (
                client radio_if_t i_radio,
                timing_transx_t   &session_trans,
                const uint8_t     TARGETID_toAddress,
                const packet_t    PACKET)
        {
            return_trans3_t return_trans3;
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.send_trans1 (TARGETID_toAddress, PACKET);
                {
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
            return return_trans3.u_out.waitForIRQInterruptCause; // Only valid if not timed out
        }


        dBm_t readRSSI_dBm_iff_asynch (
                client radio_if_t    i_radio,
                timing_transx_t      &session_trans,
                const forceTrigger_t forceTrigger) // Received Signal Strength Indication
        {
            return_trans3_t return_trans3;
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.readRSSI_dBm_trans1 (forceTrigger);
                {
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
            return return_trans3.u_out.rssi_dBm; // Only valid if not timed out
        }


        interruptAndParsingResult_e handleSPIInterrupt_iff_asynch (
                 client radio_if_t      i_radio,
                 timing_transx_t        &session_trans,
                 some_rfm69_internals_t &return_some_rfm69_internals, // Only updated if not timed out
                 packet_t               &return_PACKET)               // Only updated if not timed out
        {
            return_trans3_t return_trans3;
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.handleSPIInterrupt_trans1 ();
                {
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2

                    return_some_rfm69_internals = return_trans3.u_out.handleSPIInterrupt.return_some_rfm69_internals;
                    return_PACKET               = return_trans3.u_out.handleSPIInterrupt.return_PACKET;
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
            return return_trans3.u_out.handleSPIInterrupt.return_interruptAndParsingResult; // Only valid if not timed out
        }


        bool receiveDone_iff_asynch (
                client radio_if_t i_radio,
                timing_transx_t &session_trans)
        {
            return_trans3_t return_trans3;
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.receiveDone_trans1 ();
                {
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
            return return_trans3.u_out.receiveDone; // Only valid if not timed out
        }

        void ultimateIRQclear_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans)
        {
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.ultimateIRQclear_trans1 ();
                {
                    return_trans3_t return_trans3; // No return value to use
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
        }

        void setPowerLevel_dBm_iff_asynch  (client radio_if_t i_radio, timing_transx_t &session_trans, const uint8_t powerLevel_dBm)
        {
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.setPowerLevel_dBm_trans1 (powerLevel_dBm);
                {
                    return_trans3_t return_trans3; // No return value to use
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
        }

        void encrypt16_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const char key[KEY_LEN]) // len not possible here
        {
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.encrypt16_trans1 (key);
                {
                    return_trans3_t return_trans3; // No return value to use
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
        }

        void do_aux_adafruit_rfm69hcw_RST_pulse_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const unsigned maskof_pin) // MUST BE CALLED FIRST AND ONLY ONCE, EVEN BEFORE initialize. Reason: it blocks on await_spi_port_init_by_all_clients
        {
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.do_aux_adafruit_rfm69hcw_RST_pulse_trans1 (maskof_pin);
                {
                    return_trans3_t return_trans3; // No return value to use
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
        }

        void initialize_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const rfm69_params_t init)
        {
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.initialize_trans1 (init);
                {
                    return_trans3_t return_trans3; // No return value to use
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
        }

        uint8_t getDeviceType_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans)
        {
            return_trans3_t return_trans3;
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.getDeviceType_trans1 ();
                {
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
            return return_trans3.u_out.deviceType; // Only valid if not timed out
        }

        void setHighPower_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const bool isHighPowerOn) // For RFM69HW only: you must call setHighPower(true) after initialize() or else transmission won't work
        {
            if (not session_trans.timed_out_trans1to2) {
                session_trans.start_time_trans1 = i_radio.setHighPower_trans1 (isHighPowerOn);
                {
                    return_trans3_t return_trans3; // No return value to use
                    do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                }
            } else {
                // Server may be pathologically blocked, don't use i_radio
            }
        }

        #if (GET_RADIO_DEBUG_REGS_LIB==1)
            debug_data_t getDebug_iff_asynch (client radio_if_t i_radio, timing_transx_t &session_trans, const debug_state_e debug_state) // data indexed with iof_debug_bytes_e
            {
                debug_data_t debug_data;
                if (not session_trans.timed_out_trans1to2) {
                    session_trans.start_time_trans1 = i_radio.getDebug_trans1 (debug_state);
                    {
                        return_trans3_t return_trans3; // No return value to use
                        do_sessions_trans2to3 (i_radio, session_trans, return_trans3); // Will detect timed_out_trans1to2
                        debug_data = return_trans3.u_out.debug_data;
                    }
                } else {
                    // Server may be pathologically blocked, don't use i_radio
                }
                return debug_data;
            }
        #endif

    #endif
#endif

void func_do_aux_adafruit_rfm69hcw_RST_pulse (
        const unsigned       maskof_pin,
        unsigned             &spi_aux_bits,
        out port             p_spi_aux,  // SPI unit restart in SPI_AUX
        client spi_master_if i_spi,
        unsigned             spi_client)  // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN

{
    // See http://www.xcore.com/viewtopic.php?f=26&t=6446#p32106 on why I don't have the port as param here
    debug_print ("spi_client %u await_spi_port_init_by_all_clients\n", spi_client);
    i_spi.await_spi_port_init_by_all_clients();
    debug_print ("spi_client %u await_spi_port_init_by_all_clients done\n", spi_client);
    {
        // DOWNLOAD CODE                    HIGH             LOW
        //    |---|                          |----------------|
        //    |   |   |----------------------|      10 ms     | (*)
        // --/    \---|                                       |--- 10 ms ----!--
        //                                   !                               !USE SPI
        //                await_spi_port_init_by_all_clients!

        spi_aux_bits or_eq maskof_pin;               // BIT HIGH
        p_spi_aux <: spi_aux_bits;                   // PIN HIGH
        delay_milliseconds (SPI_AUX0_RST_TIME_A_MS); // 10 ms
        spi_aux_bits and_eq (compl maskof_pin);      // BIT LOW
        p_spi_aux <: spi_aux_bits;                   // PIN LOW
        delay_milliseconds (SPI_AUX0_RST_TIME_B_MS); // 10 ms

        // (*)
        //    - IF RST IS GIVEN A LOW RST PULSE AND THEN LEFT HIGH (this cost me hours of work to find out!),
        //      THEN SX1231H _WILL_ RESPOND, BUT _NOT_ CORRECT!
        //    - Also, if the RadioHead library running the Adafruit Feather M0 and RFM69HCW is modified to take
        //      the RST line finally HIGH it also fails. We can see that some while(until some condition)
        //      then never will quit. Teig: in my code I have wrapped all these in timoeuts because that's
        //      good programming practice

        // RadioHead library code is like this (milliseconds)
        //
        // manual reset
        // digitalWrite(RFM69_RST, HIGH);
        // delay(10);
        // digitalWrite(RFM69_RST, LOW);
        // delay(10);

        // RFM69 library does not handle the RESET pin at all!
    }
}

void func_do_spi_aux_pin (
        const unsigned maskof_pin,
        const pin_e    value,
        unsigned       &spi_aux_bits,
        out port       p_spi_aux) // SPI unit restart in SPI_AUX
{
    if (value == high) {
        spi_aux_bits or_eq maskof_pin;
    } else { // low
        spi_aux_bits and_eq (compl maskof_pin);
    }
    p_spi_aux <: spi_aux_bits;
}

uint8_t func_getDeviceType (
        client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        unsigned             spi_client,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    rfm69_internals.device_type  = spiRead (REG_VERSION, i_spi, spi_params);

    if ((rfm69_internals.device_type == 0x00) or (rfm69_internals.device_type == 0xFF)) {
        rfm69_internals.error_bits or_eq (1 << ERROR_NO_SPI_UNIT_CONNECTED);
        // If all zeros: we could have kept track of all bits read from the start and cleared this errosr state on first non-zero
        // If all ones: the opposite of the above, however,
        // we dont' do this since if the line is floating we wouldn't know
        // However, we have tested on no connected unit, and on the startKIT zero is read always
    } else if (rfm69_internals.device_type != RF_VERSION_VER) {
        rfm69_internals.error_bits or_eq (1 << ERROR_BITNUM_DEVICE_TYPE);
    } else {}

    return rfm69_internals.device_type;
}

void func_handleSPIInterrupt (
        some_rfm69_internals_t      &return_some_rfm69_internals,
        packet_t                    &return_PACKET,
        interruptAndParsingResult_e &return_interruptAndParsingResult,
        client spi_master_if        i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        spi_params_t                &spi_params,
        rfm69_internals_t           &rfm69_internals)
{
        // AQU=065e Commenting out all of this removed the AQU=065e, but not individual calls!
        // AQU=065e However, all four calls commented out removes it!
        // AQU=065e commenting out either handleSPIInterrupt or _receiveBegin causes error

        uint8_t RegIrqFlags2;
        VALUE_XSCOPE_LIB (RFM69_VALUE,1001);

        {return_interruptAndParsingResult, RegIrqFlags2} = _handleSPIInterrupt (i_spi, spi_params, rfm69_internals);
        VALUE_XSCOPE_LIB (RFM69_VALUE,2001);

        do_getSomeRadioInternals (rfm69_internals, return_some_rfm69_internals, return_PACKET);
        VALUE_XSCOPE_LIB (RFM69_VALUE,2002);

        return_some_rfm69_internals.RegIrqFlags2 = RegIrqFlags2; // AQU=065e no change to keep

        if (return_interruptAndParsingResult != messageReceivedOk_IRQ) {
            VALUE_XSCOPE_LIB (RFM69_VALUE,2003);
            // Any junk remaining in the FIFO will be cleared next time we go to receive mode.
            _receiveBegin (i_spi, spi_params, rfm69_internals); // Since this will clear the rfm69_internals copied over above
        } else {
            VALUE_XSCOPE_LIB (RFM69_VALUE,2004);
        }
        VALUE_XSCOPE_LIB (RFM69_VALUE,2005);

        debug_print_highpri ("func_handleSPIInterrupt %u %08X %u\n", return_interruptAndParsingResult, return_some_rfm69_internals.RegIrqFlags2, return_some_rfm69_internals.PACKETLEN);
        VALUE_XSCOPE_LIB (RFM69_VALUE,2006);
}

void func_rcCalibration (
        client spi_master_if i_spi,
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    // Trigger the calibration of the RC oscillator when set. Always reads 0.
    // RC calibration must be triggered in RF_OPMODE_MODE_STANDBY (default mode)
    //
    SPI_WRITE_REG (REG_OSC1, RF_OSC1_RCCAL_START);

    uint8_t reg;
    reg = while_readRegExpectedMask (0,      // "000-B loop". discarding reg (Never seen it loop:)
            REG_OSC1, RF_OSC1_RCCAL_DONE,    // readFromReg, readFromRegMask
            logic_normal,                    // invertedLogic
            WAIT_FOR_RCCALIBRATION_MS,       // timeout_ms. In the original there was no timeout here
            ERROR_BITNUM_RF_OSC1_RCCAL_DONE, // error_bitnum
            i_spi, spi_params, rfm69_internals);
}

// set *transmit/TX* output power: 0=min APPPPOWERLEVEL_MIN_0, 31=max APPPPOWERLEVEL_MAX_DBM
// this results in a "weaker" transmitted signal, and directly results in a lower RSSI at the receiver
// the power configurations are explained in the SX1231H datasheet (Table 10 on p21; RegPaLevel p66):
// http://www.semtech.com/images/datasheet/sx1231h.pdf
// valid powerLevel parameter values are 0-31 and result in a directly proportional effect on the output/transmission power
// this function implements 2 modes as follows:
//       - for RFM69W the range is from 0-31 [-18dBm to 13dBm] (PA0 only on RFIO pin)
//       - for RFM69HW the range is from 0-31 [5dBm to 20dBm]  (PA1 & PA2 on PA_BOOST pin & high Power PA settings -
//         see section 3.3.7 in datasheet, p22)

// Teig: spelling it out here:
// From the Semtech SX1231H manual describing RegPaLevel and Power Amplifiers (PA) truth table:
//                                        Connecting antenna and Power Amplifiers PA0, PA1 and PA2
//                                        PA0                   output/input on pin RFIO     (pin 21)
//                                                PA1 and PA2   outputs      on pin PA_BOOST (pin 23)
// ---------------------------------------###-----###-----###--------------------------------------------------------------
// Output power setting, with 1 dB steps  PA0On   PA1On   PA2On  -----Range----  #Steps  Values   ------Formula--------
// Pout = -18 + OutputPower [dBm], with     1       0       0    -18 to +13 dBm    32     0..31   -18 dBm + OutputPower
// Pout = -18 + OutputPower [dBm], with     0       1       0     -2 to +13 dBm    16    16..31   -18 dBm + OutputPower
// Pout = -14 + OutputPower [dBm], with     0       1       1     +2 to +17 dBm    16    16..31   -14 dBm + OutputPower
// Pout = -11 + OutputPower [dBm], with     0       1       1     +5 to +20 dBm    16    16..31   -11 dBm + OutputPower (*)
// (*) When setHighPower set, which I think we MUST do on the RFM69HCW with SX1231H, so this is ALWAYS available
//
void func_setPowerLevel_dBm (
        const uint8_t        powerLevel_dBm,
        client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    rfm69_internals.powerLevel_dBm = (powerLevel_dBm > APPPPOWERLEVEL_MAX_DBM ? APPPPOWERLEVEL_MAX_DBM : powerLevel_dBm);
    if (rfm69_internals.isRadio_RFM69HW) {
        rfm69_internals.powerLevel_dBm /= 2; // See #Steps above
    } else {}

    uint8_t RegPaLevel = SPI_READ_REG (REG_PALEVEL);
    RegPaLevel = (RegPaLevel bitand compl RF_PALEVEL_OUTPUTPOWER_MASK) bitor rfm69_internals.powerLevel_dBm;
    SPI_WRITE_REG (REG_PALEVEL, RegPaLevel);
}

void func_setFrequencyRegister (
        const uint32_t       register_value,
        client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    uint8_t     oldMode = rfm69_internals.radio_mode;
    fourbytes_u register_value_;

    register_value_.u.value = register_value;

    if (oldMode == RF69_MODE_TX) {
        setMode (1, RF69_MODE_RX, i_spi, spi_params, rfm69_internals);
    } else {}

    SPI_WRITE_REG (REG_FRFMSB, register_value_.u.bytes[2]);
    SPI_WRITE_REG (REG_FRFMID, register_value_.u.bytes[1]);
    SPI_WRITE_REG (REG_FRFLSB, register_value_.u.bytes[0]);

    if (oldMode == RF69_MODE_RX) {
        setMode (2, RF69_MODE_SYNTH, i_spi, spi_params, rfm69_internals);
    } else {}

    setMode (3, oldMode, i_spi, spi_params, rfm69_internals);
}

uint32_t func_getFrequencyRegister (
        client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    uint32_t return_register_value;

    fourbytes_u register_value;

    register_value.u.value    = 0; // All four bytes
    register_value.u.bytes[2] = SPI_READ_REG (REG_FRFMSB);
    register_value.u.bytes[1] = SPI_READ_REG (REG_FRFMID);
    register_value.u.bytes[0] = SPI_READ_REG (REG_FRFLSB);

    return_register_value = register_value.u.value;
    return return_register_value;
}

// When reading while_readRegExpectedMask REG_IRQFLAGS2 bit RF_IRQFLAGS2_FIFONOTEMPTY is not enough
// we need a mechanism to recover from all situations:
// See SEMTECH SX1231H datasheet 5.2.2.3 (p46):
//   Table 20 Status of FIFO when Switching Between Different Modes of the Chip
//     From Stdby/Sleep to Rx  FIFO status Cleared
//     From Rx          to Tx  FIFO status Cleared WHAT WE DO BELOW SINCE WE ARE IN RF69_MODE_RX
//     From Tx          to Any FIFO status Cleared
// The good thing is that with this solution it's not down to RF69_MODE_STANDBY between every reception,
// and _maybe_ we will see more packages and lose fewer
//
void func_ultimateIRQclear (
        client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        spi_params_t         &spi_params,
        rfm69_internals_t    &rfm69_internals)
{
    setMode (23, RF69_MODE_STANDBY, i_spi, spi_params, rfm69_internals);
    setMode (24, RF69_MODE_RX,      i_spi, spi_params, rfm69_internals);
}

#if (GET_RADIO_DEBUG_REGS_LIB==1)

    debug_data_t func_getDebug (
            const debug_state_e  debug_state,
            client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
            spi_params_t         &spi_params,
            rfm69_internals_t    &rfm69_internals)
    {
        debug_data_t debug_data;
        if (debug_state == debug_just_read_some_registers) {                                           // Values read not _directly_ associated with IRQ reception
                                                                                                       // 90 D8 00 04 00
                                                                                                       // 90 D9 40 04 00
                                                                                                       // 90 D9 60 04 00
                                                                                                       // 90 D9 66 04 00
                                                                                                       //
            debug_data.data [iof_RegOpMode]                = SPI_READ_REG (REG_OPMODE);                // 0x90 0x90  0x90   RF_OPMODE_SEQUENCER_OFF + RF_OPMODE_MODE_RECEIVER
            debug_data.data [iof_RegIrqFlags1]             = SPI_READ_REG (REG_IRQFLAGS1);             // 0xD8              RF_IRQFLAGS1_MODEREADY + RF_IRQFLAGS1_RXREADY + RF_IRQFLAGS1_PLLLOCK + RF_IRQFLAGS1_RSSI
                                                                                                       //      0xD9  0xD9 + RF_IRQFLAGS1_SYNCADDRESSMATCH
            debug_data.data [iof_RegIrqFlags2]             = SPI_READ_REG (REG_IRQFLAGS2);             // 0x00
                                                                                                       //      0x40       + RF_IRQFLAGS2_FIFONOTEMPTY
                                                                                                       //            0x66 + RF_IRQFLAGS2_FIFOLEVEL + RF_IRQFLAGS2_PAYLOADREADY + RF_IRQFLAGS2_CRCOK
            debug_data.data [iof_radio_mode]               = rfm69_internals.radio_mode;               // 0x04 0x04  0x04   RF69_MODE_RX
            debug_data.data [iof_waitForIRQInterruptCause] = rfm69_internals.waitForIRQInterruptCause; // 0x00 0x00  0x00

        } else if (debug_state == debug_flush_RegIrqFlags2) {
            // Flush the FIFO, most often for ZERO bytes! See read of RF_IRQFLAGS2_FIFONOTEMPTY in code above
            debug_data.data [iof_RegIrqFlags2] = while_readRegExpectedMask (9, // "009-B loop". discarding reg (Never seen it loop:)
                    REG_IRQFLAGS2, RF_IRQFLAGS2_FIFONOTEMPTY,                  // RegIrqFlags2, FifoNotEmpty
                    logic_inverted,                                            // loop until bit in flag becomes 0
                    WAIT_FOR_REGISTER_VALUE_MS,                                // timeout_ms
                    ERROR_BITNUM_RF_IRQFLAGS2_FIFONOTEMPTY,                    // error_bitnum.
                    i_spi, spi_params, rfm69_internals);
            // rfm69_internals.error_bits now updated. Get it with i_radio.getAndClearErrorBits
        } else if (debug_state == debug_mode_0_1) { // As ultimateIRQclear (above)
            func_ultimateIRQclear (i_spi, spi_params, rfm69_internals);
        }
        else {} // debug_none

        return debug_data;
    }
#endif

// ------------------------------------------------------------------------------------------------------------------------------------------------
// TASK "RFM69_driver" with methods like in RFM69.cpp
// ------------------------------------------------------------------------------------------------------------------------------------------------
//
[[combinable]]
void RFM69_driver (
        server radio_if_t    i_radio,
        out port             p_spi_aux,  // SPI unit restart in SPI_AUX
        client spi_master_if i_spi,      // Used to talk with lib_spi, who owns SPI_MOSI, SPI_CLK, SPI_MISO and SPI_CS_EN
        unsigned             spi_client)
{
    #if (ALLOW_SESSION_TYPE_TRANS_LIB>=1)
        internal_transx_t internal_transx;

        internal_transx.do_async_trans2 = false;
        internal_transx.return_trans3.test_inc_1000 = 0;
    #endif

    INIT_SHARED_GLOBAL(g_radio_log_value,0);

    error_bits_e error_bits_previous = ERROR_BITS_NONE;

    spi_params_t spi_params = {
             spi_client,
             SPI_CLOCK_SPEED_KHZ,
             SPI_MODE_0_CORRECTED,
             SPI_DEASSERT_TIME};

    rfm69_internals_t rfm69_internals;
    // no list init since = {not, safe, allows only first part, and most types are uint8_t}
    rfm69_internals.error_bits               = ERROR_BITS_NONE;
    rfm69_internals.radio_mode               = RF69_MODE_VOID; // To trig setMode first time
    rfm69_internals.doListenToAll            = false;
    rfm69_internals.powerLevel_dBm           = APPPPOWERLEVEL_MAX_DBM;
    rfm69_internals.isRadio_RFM69HW          = ISRFM69HW;
    rfm69_internals.nodeID                   = 0;
    rfm69_internals.device_type              = RF_VERSION_VER_VOID;
    rfm69_internals.APPPROTLEN               = 0;
    rfm69_internals.SENDERID                 = 0;
    rfm69_internals.TARGETID                 = 0;
    rfm69_internals.PACKETLEN                = 0;
    rfm69_internals.waitForIRQInterruptCause = no_IRQExpected;

    unsigned spi_aux_bits = 0;

    while (1) {
        select {
            case i_radio.do_spi_aux_pin (const unsigned maskof_pin, const pin_e value) : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,1);
                func_do_spi_aux_pin (maskof_pin, value, spi_aux_bits, p_spi_aux);
                VALUE_XSCOPE_LIB (RFM69_VALUE,2);
            } break;

            case i_radio.getAndClearErrorBits (void) -> {error_bits_e return_error_bits, is_error_e is_error_return} : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,3);
                if ((rfm69_internals.error_bits != ERROR_BITS_NONE) and (rfm69_internals.error_bits != error_bits_previous)) {
                    is_error_return = IS_ERR;
                    PRINT_ERROR_BITS_HIGHPRI (rfm69_internals.error_bits);
                } else {
                    is_error_return = NO_ERR;
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.getAndClearErrorBits", rfm69_internals.error_bits);
                }
                return_error_bits          = rfm69_internals.error_bits;
                error_bits_previous        = rfm69_internals.error_bits;
                rfm69_internals.error_bits = ERROR_BITS_NONE;
                VALUE_XSCOPE_LIB (RFM69_VALUE,4);
            } break;

            case i_radio.getAndClearErrorBits_ (void) -> {error_t return_error} : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,5);
                if ((rfm69_internals.error_bits != ERROR_BITS_NONE) and (rfm69_internals.error_bits != error_bits_previous)) {
                    return_error.is_error = IS_ERR;
                    PRINT_ERROR_BITS_HIGHPRI (rfm69_internals.error_bits);
                } else {
                    return_error.is_error= NO_ERR;
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.getAndClearErrorBits", rfm69_internals.error_bits);
                }

                return_error.error_bits    = rfm69_internals.error_bits;
                error_bits_previous        = rfm69_internals.error_bits;
                rfm69_internals.error_bits = ERROR_BITS_NONE;
                VALUE_XSCOPE_LIB (RFM69_VALUE,6);
            } break;

            // true  = disable filtering to capture all frames on network
            // false = enable node/broadcast filtering (in sw here) to capture only frames sent to this/broadcast address
            case i_radio.setListenToAll (const bool doListenToAll) : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,7);
                rfm69_internals.doListenToAll = doListenToAll;
                // Assumes RF_PACKET1_ADRSFILTERING_OFF, as it's a mechanism on top of it with all packets read
                PRINT_ADD_ERROR_BITS_IFF ("i_radio.doListenToAll", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,8);
            } break; // i_radio.doListenToAll

            case i_radio.uspi_rcCalibration (void) : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,9);
                // I have kept this function even if I don't use it, since I did code it and it's rather complex

                func_rcCalibration (i_spi, spi_params, rfm69_internals);

                PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_rcCalibration", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,10);
            } break; // i_radio.uspi_rcCalibration

            case i_radio.uspi_readTemperature_degC (const int8_t calOffset) -> int8_t temperature_degC : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,11);
                temperature_degC = readTemperature_degC (calOffset, i_spi, spi_params, rfm69_internals);
                PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_readTemperature_degC", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,12);
            } break; // i_radio.readTemperature

            #if (ALLOW_SESSION_TYPE_TRANS_LIB==0) // Linker seems not to exclude these
                //           uspi_do_aux_adafruit_rfm69hcw_RST_pulse MUST BE CALLED FIRST AND ONLY ONCE, EVEN BEFORE func_initialize. Reason: it blocks on await_spi_port_init_by_all_clients
                case i_radio.uspi_do_aux_adafruit_rfm69hcw_RST_pulse (const unsigned maskof_pin) : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,13);
                    func_do_aux_adafruit_rfm69hcw_RST_pulse (maskof_pin, spi_aux_bits, p_spi_aux, i_spi, spi_client);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,14);
                } break;

                case i_radio.uspi_initialize (const rfm69_params_t init) : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,15);
                    func_initialize (init, i_spi, spi_params, rfm69_internals);
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_initialize", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,16);
                } break;

                case i_radio.uspi_handleSPIInterrupt (void) -> {
                        some_rfm69_internals_t      return_some_rfm69_internals,
                        packet_t                    return_PACKET,
                        interruptAndParsingResult_e return_interruptAndParsingResult } : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,17);

                    some_rfm69_internals_t      some_rfm69_internals;
                    packet_t                    PACKET;
                    interruptAndParsingResult_e interruptAndParsingResult;

                    // AQU=065 inlining the code here does not help:
                    func_handleSPIInterrupt (some_rfm69_internals, PACKET, interruptAndParsingResult, i_spi, spi_params, rfm69_internals);

                    VALUE_XSCOPE_LIB (RFM69_VALUE,1818); // AQU=065d prints this!

                    return_some_rfm69_internals = some_rfm69_internals;
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1819);
                    return_PACKET = PACKET;
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1820);
                    return_interruptAndParsingResult = interruptAndParsingResult;
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1821); // AQU=065d prints this!

                    // PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_handleSPIInterrupt", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,18); // AQU=065d prints this!
                    // delay_milliseconds (1); // THIS NEVER RETURNS WHEN func_handleSPIInterrupt IS RUN!
                    VALUE_XSCOPE_LIB (RFM69_VALUE,1823); // AQU=065d does NOT print this!
                } break; // i_radio.uspi_handleSPIInterrupt

                // Get the Received Signal Strength Indication (RSSI)
                case i_radio.uspi_readRSSI_dBm (const forceTrigger_t forceTrigger) -> dBm_t return_rssi_dBm : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,19);
                    return_rssi_dBm = readRSSI_dBm (forceTrigger, i_spi, spi_params, rfm69_internals);
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_readRSSI_dBm", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,20);
                } break; // i_radio.uspi_readRSSI_dBm

                case i_radio.uspi_send (const uint8_t  TARGETID_toAddress, const packet_t PACKET) -> waitForIRQInterruptCause_e return_waitForIRQInterruptCause : {

                    VALUE_XSCOPE_LIB (RFM69_VALUE,21);

                    func_send (TARGETID_toAddress, PACKET, i_spi, spi_params, rfm69_internals);

                    return_waitForIRQInterruptCause = rfm69_internals.waitForIRQInterruptCause;

                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_send", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,22);
                } break; // i_radio.uspi_send

                case i_radio.uspi_setPowerLevel_dBm (const uint8_t powerLevel_dBm) : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,23);
                    func_setPowerLevel_dBm (powerLevel_dBm, i_spi, spi_params, rfm69_internals);

                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.setPowerLevel", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,24);
                } break; // i_radio.setPowerLevel

                case i_radio.uspi_encrypt16 (const char key[], unsigned const static len) : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,25);
                    char key_[len];
                    for (unsigned iof_key=0; iof_key < len; iof_key++) {
                        key_[iof_key] = key[iof_key];
                    }
                    func_encrypt16 (key_, len, i_spi, spi_params, rfm69_internals);
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_encrypt16", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,26);
                } break;

                case i_radio.uspi_getDeviceType (void) -> uint8_t returnDeviceType : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,27);
                    returnDeviceType = func_getDeviceType (i_spi, spi_client, spi_params, rfm69_internals);

                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_getDeviceType", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,28);
                } break;

                case i_radio.uspi_setHighPower (const bool isHighPowerOn) : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,29);
                    setHighPower (isHighPowerOn, i_spi, spi_params, rfm69_internals);
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_setHighPower", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,30);
                } break; // i_radio.uspi_setHighPower

                case i_radio.uspi_receiveDone (void) -> bool return_done: {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,31);
                    return_done = func_receiveDone (i_spi, spi_params, rfm69_internals);
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_receiveDone", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,32);
                } break; // i_radio.uspi_receiveDone

            #endif

            #if (ALLOW_SESSION_TYPE_TRANS_LIB>=1)

                case i_radio.do_aux_adafruit_rfm69hcw_RST_pulse_trans1 (const unsigned maskof_pin) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,33);
                    internal_transx.param_for_trans1.u_in.maskof_pin = maskof_pin;
                    internal_transx.return_trans3.id_trans1          = session_do_aux_adafruit_rfm69hcw_RST_pulse_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ---
                    VALUE_XSCOPE_LIB (RFM69_VALUE,34);
                } break;

                case i_radio.initialize_trans1 (const rfm69_params_t init) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,35);
                    internal_transx.param_for_trans1.u_in.init = init;
                    internal_transx.return_trans3.id_trans1    = session_initialize_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ---
                    VALUE_XSCOPE_LIB (RFM69_VALUE,36);
                } break;

                case i_radio.getDeviceType_trans1 (void) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,37);
                    internal_transx.return_trans3.id_trans1 = session_getDeviceType_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ---
                    VALUE_XSCOPE_LIB (RFM69_VALUE,38);
                } break;

                case i_radio.encrypt16_trans1 (const char key[KEY_LEN]) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,39);
                    for (unsigned iof_key=0; iof_key < KEY_LEN; iof_key++) {
                        internal_transx.param_for_trans1.u_in.key[iof_key] = key[iof_key];
                    }

                    internal_transx.return_trans3.id_trans1 = session_encrypt16_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ---
                    VALUE_XSCOPE_LIB (RFM69_VALUE,40);
                } break;

                case i_radio.send_trans1 (const uint8_t TARGETID_toAddress, const packet_t PACKET) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,41);
                    internal_transx.param_for_trans1.u_in.send.TARGETID_toAddress = TARGETID_toAddress;
                    internal_transx.param_for_trans1.u_in.send.PACKET             = PACKET;
                    internal_transx.return_trans3.id_trans1                       = session_send_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,42);
                } break;

                case i_radio.readRSSI_dBm_trans1 (const forceTrigger_t forceTrigger) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,43);
                    internal_transx.param_for_trans1.u_in.forceTrigger = forceTrigger;
                    internal_transx.return_trans3.id_trans1            = session_readRSSI_dBm_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,44);
                } break;

                case i_radio.handleSPIInterrupt_trans1 (void) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,45);
                    internal_transx.return_trans3.id_trans1 = session_handleSPIInterrupt_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,46);
                } break;

                case i_radio.ultimateIRQclear_trans1 (void) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,47);
                    internal_transx.return_trans3.id_trans1 = session_ultimateIRQclear_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,48);
                } break;

                case i_radio.receiveDone_trans1 (void) -> time32_t return_start_time_trans1 : {
                    internal_transx.return_trans3.id_trans1 = session_receiveDone_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                } break;

                case i_radio.rcCalibration_trans1 (void) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,49);
                    internal_transx.return_trans3.id_trans1 = session_rcCalibration_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,50);
                } break;

                case i_radio.readTemperature_degC_trans1 (const int8_t calOffset_degC) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,51);
                    internal_transx.param_for_trans1.u_in.calOffset_degC = calOffset_degC;
                    internal_transx.return_trans3.id_trans1              = session_readTemperature_degC_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,52);
                } break;

                case i_radio.setHighPower_trans1 (const bool isHighPowerOn) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,53);
                    internal_transx.param_for_trans1.u_in.isHighPowerOn = isHighPowerOn;
                    internal_transx.return_trans3.id_trans1             = session_setHighPower_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,54);
                } break;

                case i_radio.setMode_trans1 (const uint8_t newMode) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,55);
                    internal_transx.param_for_trans1.u_in.newMode = newMode;
                    internal_transx.return_trans3.id_trans1       = session_setMode_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,56);
                } break;

                case i_radio.setPowerLevel_dBm_trans1 (const uint8_t powerLevel_dBm) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,57);
                    internal_transx.param_for_trans1.u_in.powerLevel_dBm = powerLevel_dBm;
                    internal_transx.return_trans3.id_trans1              = session_setPowerLevel_dBm_trans1;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,58);
                } break;

                case i_radio.setFrequencyRegister_trans1 (const uint32_t register_value) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,59);
                    internal_transx.param_for_trans1.u_in.register_value = register_value;
                    internal_transx.return_trans3.id_trans1              = session_setFrequencyRegister;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,60);
                } break;

                case i_radio.getFrequencyRegister_trans1 (void) -> time32_t return_start_time_trans1 : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,61);
                    internal_transx.return_trans3.id_trans1 = session_getFrequencyRegister;

                    internal_transx.do_async_trans2 = true;
                    internal_transx.tmr :> internal_transx.time; // Immediately
                    return_start_time_trans1 = internal_transx.time;
                    // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                    VALUE_XSCOPE_LIB (RFM69_VALUE,62);
                } break;

                #if (GET_RADIO_DEBUG_REGS_LIB==1)
                    case i_radio.getDebug_trans1 (const debug_state_e debug_state) -> time32_t return_start_time_trans1 : {
                        VALUE_XSCOPE_LIB (RFM69_VALUE,63);
                        internal_transx.param_for_trans1.u_in.debug_state = debug_state;
                        internal_transx.return_trans3.id_trans1           = session_getDebug;

                        internal_transx.do_async_trans2 = true;
                        internal_transx.tmr :> internal_transx.time; // Immediately
                        return_start_time_trans1 = internal_transx.time;
                        // ---- Next MUST BE that client waits for i_radio.timing_transx2 ----
                        VALUE_XSCOPE_LIB (RFM69_VALUE,64);
                    } break;
                #endif

                case (internal_transx.do_async_trans2) => internal_transx.tmr when timerafter(internal_transx.time) :> void: {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,65+internal_transx.return_trans3.id_trans1); // 65-82
                    // Simulates common handling (would have disallowed [[combinable]])
                    internal_transx.do_async_trans2 = false;

                    debug_print ("begin i_radio.timing_transx2 %u\n", internal_transx.return_trans3.id_trans1);

                    // DO BACKGROUND WORK. THIS MAY BLOCK (AQU=065?). IT TALKS OVER SPI. IF IT BLOCKS THEN NO i_radio.session_trans2 WILL BE SENT
                    // THIS WILL BE PICKED OUT WITH do_sessions_trans2to3 IF USED

                    // Observe that this GET_SHARED_GLOBAL, do something then SET_SHARED_GLOBAL is done asynchronously with the
                    // interface call. But it does not help to remove it to resolve RFM69=013
                    SET_FIELD_SHARED_GLOBAL(g_radio_log_value,internal_transx.return_trans3.id_trans1);

                    switch (internal_transx.return_trans3.id_trans1) {

                        case session_do_aux_adafruit_rfm69hcw_RST_pulse_trans1: {
                            func_do_aux_adafruit_rfm69hcw_RST_pulse (internal_transx.param_for_trans1.u_in.maskof_pin, spi_aux_bits, p_spi_aux, i_spi, spi_client);
                        } break;

                        case session_initialize_trans1: {
                            func_initialize (internal_transx.param_for_trans1.u_in.init, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_getDeviceType_trans1: {
                            internal_transx.return_trans3.u_out.deviceType = func_getDeviceType (i_spi, spi_client, spi_params, rfm69_internals);
                        } break;

                        case session_encrypt16_trans1: {
                            // argument to static parameter 2 of `func_encrypt16' must be a constant expression, not enough to assign one from a variable here!
                            func_encrypt16 (internal_transx.param_for_trans1.u_in.key, KEY_LEN, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_send_trans1: {
                            func_send (
                                    internal_transx.param_for_trans1.u_in.send.TARGETID_toAddress,
                                    internal_transx.param_for_trans1.u_in.send.PACKET,
                                    i_spi, spi_params, rfm69_internals);
                            internal_transx.return_trans3.u_out.waitForIRQInterruptCause = rfm69_internals.waitForIRQInterruptCause;
                        } break;

                        case session_readRSSI_dBm_trans1: {
                            internal_transx.return_trans3.u_out.rssi_dBm = readRSSI_dBm (internal_transx.param_for_trans1.u_in.forceTrigger, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_handleSPIInterrupt_trans1: {

                            // Code increased by 192 bytes. TODO remove
                            some_rfm69_internals_t      return_some_rfm69_internals;
                            packet_t                    return_PACKET;
                            interruptAndParsingResult_e return_interruptAndParsingResult;

                            func_handleSPIInterrupt (
                                    return_some_rfm69_internals,
                                    return_PACKET,
                                    return_interruptAndParsingResult,
                                    i_spi, spi_params, rfm69_internals);

                            debug_print_highpri ("session_handleSPIInterrupt_trans1 %u %08X %u\n",
                                return_interruptAndParsingResult,
                                return_some_rfm69_internals.RegIrqFlags2,
                                return_some_rfm69_internals.PACKETLEN);

                            internal_transx.return_trans3.u_out.handleSPIInterrupt.return_some_rfm69_internals      = return_some_rfm69_internals;
                            internal_transx.return_trans3.u_out.handleSPIInterrupt.return_PACKET                    = return_PACKET;
                            internal_transx.return_trans3.u_out.handleSPIInterrupt.return_interruptAndParsingResult = return_interruptAndParsingResult;


                            #if (DEBUG_FOREVER_LOOP_LIB==1)
                                #warning DEBUG_FOREVER_LOOP_LIB
                                while(1) {}
                            #endif

                        } break;

                        case session_ultimateIRQclear_trans1: {
                            func_ultimateIRQclear (i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_receiveDone_trans1: {
                            internal_transx.return_trans3.u_out.receiveDone = func_receiveDone (i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_rcCalibration_trans1: {
                            func_rcCalibration (i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_readTemperature_degC_trans1: {
                            internal_transx.return_trans3.u_out.temperature_degC = readTemperature_degC (internal_transx.param_for_trans1.u_in.calOffset_degC, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_setHighPower_trans1: {
                            setHighPower (internal_transx.param_for_trans1.u_in.isHighPowerOn, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_setMode_trans1: {
                            setMode (4, internal_transx.param_for_trans1.u_in.newMode, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_setPowerLevel_dBm_trans1: {
                            func_setPowerLevel_dBm (internal_transx.param_for_trans1.u_in.powerLevel_dBm, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_setFrequencyRegister: {
                            func_setFrequencyRegister (internal_transx.param_for_trans1.u_in.register_value, i_spi, spi_params, rfm69_internals);
                        } break;

                        case session_getFrequencyRegister: {
                            internal_transx.return_trans3.u_out.register_value = func_getFrequencyRegister (i_spi, spi_params, rfm69_internals);
                        } break;

                        #if (GET_RADIO_DEBUG_REGS_LIB==1)
                            case session_getDebug : {
                                internal_transx.return_trans3.u_out.debug_data = func_getDebug (internal_transx.param_for_trans1.u_in.debug_state, i_spi, spi_params, rfm69_internals);
                            } break;
                        #endif
                    }

                    i_radio.session_trans2 (); // Is the [[notification]]
                    // ---- Next MUST BE that client calls i_radio.timing_transx3 ---
                    VALUE_XSCOPE_LIB (RFM69_VALUE,90+internal_transx.return_trans3.id_trans1); // 90-107
                } break;

                case i_radio.session_trans3 (void) -> return_trans3_t return_trans3 : { // Is the [[clears_notification]]

                    VALUE_XSCOPE_LIB (RFM69_VALUE,121);
                    internal_transx.return_trans3.test_inc_1000 += 1000;
                    return_trans3 = internal_transx.return_trans3;

                    debug_print ("%u", "i_radio.timing_transx3\n");
                    VALUE_XSCOPE_LIB (RFM69_VALUE,122);
                } break;

            #endif

            case i_radio.setNODEID (const uint8_t newNODEID) -> uint8_t return_previous_NODEID : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,123);
                return_previous_NODEID = rfm69_internals.nodeID;
                rfm69_internals.nodeID = newNODEID;

                // Teig: would this side effect into RF69_MAX_FIFO_LEN_65 or RF69_MAX_FIFO_LEN_66: TODO
                // SPI_WRITE_REG (REG_NODEADRS, address); // RegNodeAdrs is Optional Address byte (Node ID)
                // Code above also commented away in RFM69

                PRINT_ADD_ERROR_BITS_IFF ("i_radio.setNODEID", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,124);
            } break; // i_radio.setNODEID

            case i_radio.uspi_setFrequencyRegister (const uint32_t register_value): { // Sets Mode but restores it as well
                VALUE_XSCOPE_LIB (RFM69_VALUE,125);
                func_setFrequencyRegister (register_value, i_spi, spi_params, rfm69_internals);
                PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_setFrequencyRegister", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,126);
            } break; // i_radio.uspi_setFrequencyRegister

            case i_radio.uspi_getFrequencyRegister (void) -> uint32_t return_register_value : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,127);
                return_register_value = func_getFrequencyRegister (i_spi, spi_params, rfm69_internals);
                PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_getFrequencyRegister", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,128);
            } break; // i_radio.uspi_getFrequencyRegister

            case i_radio.uspi_setMode (const uint8_t newMode) : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,129);
                setMode (4, newMode, i_spi, spi_params, rfm69_internals);
                PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_setMode", rfm69_internals.error_bits);
                VALUE_XSCOPE_LIB (RFM69_VALUE,130);
            } break; // i_radio.uspi_setMode

            case i_radio.uspi_ultimateIRQclear() : {
                VALUE_XSCOPE_LIB (RFM69_VALUE,131);
                func_ultimateIRQclear (i_spi, spi_params, rfm69_internals);
                VALUE_XSCOPE_LIB (RFM69_VALUE,132);
            } break;

            #if (GET_RADIO_DEBUG_REGS_LIB==1)
                case i_radio.uspi_getDebug (const debug_state_e debug_state) -> debug_data_t debug_data : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,133);
                    debug_data = func_getDebug (debug_state, i_spi, spi_params, rfm69_internals);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,134);
                } break;
            #endif

            #if (RADIO_IF_READALLREGS == 1)
                #if (WARNINGS == 1)
                    #warning readAllRegs compiled
                #endif
                case i_radio.uspi_readAllRegs (void) : {
                    VALUE_XSCOPE_LIB (RFM69_VALUE,201);
                    uint8_t regVal;

                    #if (DEBUG_PRINT_RFM69_REGISTER_DETAIL==1)
                        int capVal;

                        //... State Variables for intelligent decoding
                        uint8_t  modeFSK    = 0;
                        int      bitRate    = 0;
                        uint32_t freqDev    = 0;
                        uint32_t freqCenter = 0;
                    #endif

                    for (uint8_t regAddr = 1; regAddr <= REG_TEMP2; regAddr++) // REG_TEMP2 is 0x4F
                    {
                        regVal = SPI_READ_REG (regAddr);
                        debug_print_highpri ("-- regAddr:%02X, regVal:%02X --\n", regAddr, regVal);

                        #if (DEBUG_PRINT_RFM69_REGISTER_DETAIL==1)
                        // Detail, but only up to REG_LISTEN1 0x0D
                        switch (regAddr)
                        {
                            case REG_OPMODE : { // 0x01
                                debug_print_highpri ("%s",  "Controls the automatic Sequencer (see section 4.2)\nSequencerOff : ");
                                if (0x80 & regVal) {
                                    debug_print_highpri ("%s",  "1 -> Mode is forced by the user\n");
                                } else {
                                    debug_print_highpri ("%s",
                                            "0 -> Operating mode as selected with Mode bits in RegOpMode is automatically reached with the Sequencer\n");
                                }

                                debug_print_highpri ("%s",
                                        "\nEnables Listen mode, should be enabled whilst in Standby mode:\nListenOn : ");
                                if (0x40 & regVal) {
                                    debug_print_highpri ("%s",  "1 -> On\n");
                                } else {
                                    debug_print_highpri ("%s",  "0 -> Off (see section 4.3)\n");
                                }

                                debug_print_highpri ("%s",
                                        "\nAborts Listen mode when set together with ListenOn=0 See section 4.3.4 for details (Always reads 0)\n");
                                if (0x20 & regVal) {
                                    debug_print_highpri ("%s",  "ERROR - ListenAbort should NEVER return 1 this is a write only register\n");
                                }

                                debug_print_highpri ("%s", "\nTransceiver's operating modes:\nMode : ");
                                capVal = (regVal >> 2) & 0x7;
                                if (capVal == 0b000) {
                                    debug_print_highpri ("%s",  "000 -> Sleep mode (SLEEP)\n");
                                } else if (capVal == 0b001) { // Teig was "="
                                    debug_print_highpri ("%s",  "001 -> Standby mode (STDBY)\n");
                                } else if (capVal == 0b010) { // Teig was "="
                                    debug_print_highpri ("%s",  "010 -> Frequency Synthesizer mode (FS)\n");
                                } else if (capVal == 0b011) { // Teig was "="
                                    debug_print_highpri ("%s",  "011 -> Transmitter mode (TX)\n");
                                } else if (capVal == 0b100) { // Teig was "="
                                    debug_print_highpri ("%s",  "100 -> Receiver Mode (RX)\n");
                                } else {
                                    debug_print_highpri ("0x%#x", capVal);
                                    debug_print_highpri ("%s",  " -> RESERVED\n");
                                }
                                debug_print_highpri ("%s",  "\n");
                                break;
                            }

                            case REG_DATAMODUL : { // 0x02

                                debug_print_highpri ("%s", "Data Processing mode:\nDataMode : ");
                                capVal = (regVal >> 5) & 0x3;
                                if (capVal == 0b00) {
                                    debug_print_highpri ("%s",  "00 -> Packet mode\n");
                                } else if (capVal == 0b01) {
                                    debug_print_highpri ("%s",  "01 -> reserved\n");
                                } else if (capVal == 0b10) {
                                    debug_print_highpri ("%s",  "10 -> Continuous mode with bit synchronizer\n");
                                } else if (capVal == 0b11) {
                                    debug_print_highpri ("%s",  "11 -> Continuous mode without bit synchronizer\n");
                                }

                                debug_print_highpri ("%s", "\nModulation scheme:\nModulation Type : ");
                                capVal = (regVal >> 3) & 0x3;
                                if (capVal == 0b00) {
                                    debug_print_highpri ("%s",  "00 -> FSK\n");
                                    modeFSK = 1;
                                } else if (capVal == 0b01) {
                                    debug_print_highpri ("%s",  "01 -> OOK\n");
                                } else if (capVal == 0b10) {
                                    debug_print_highpri ("%s",  "10 -> reserved\n");
                                } else if (capVal == 0b11) {
                                    debug_print_highpri ("%s",  "11 -> reserved\n");
                                }

                                debug_print_highpri ("%s", "\nData shaping: ");
                                if (modeFSK) {
                                    debug_print_highpri ("%s",  "in FSK:\n");
                                } else {
                                    debug_print_highpri ("%s",  "in OOK:\n");
                                }
                                debug_print_highpri ("%s", "ModulationShaping : ");
                                capVal = regVal & 0x3;
                                if (modeFSK) {
                                    if (capVal == 0b00) {
                                        debug_print_highpri ("%s",  "00 -> no shaping\n");
                                    } else if (capVal == 0b01) {
                                        debug_print_highpri ("%s",  "01 -> Gaussian filter, BT = 1.0\n");
                                    } else if (capVal == 0b10) {
                                        debug_print_highpri ("%s",  "10 -> Gaussian filter, BT = 0.5\n");
                                    } else if (capVal == 0b11) {
                                        debug_print_highpri ("%s",  "11 -> Gaussian filter, BT = 0.3\n");
                                    }
                                } else {
                                    if (capVal == 0b00) {
                                        debug_print_highpri ("%s",  "00 -> no shaping\n");
                                    } else if (capVal == 0b01) {
                                        debug_print_highpri ("%s",  "01 -> filtering with f(cutoff) = BR\n");
                                    } else if (capVal == 0b10) {
                                        debug_print_highpri ("%s",  "10 -> filtering with f(cutoff) = 2*BR\n");
                                    } else if (capVal == 0b11) {
                                        debug_print_highpri ("%s",  "ERROR - 11 is reserved\n");
                                    }
                                }

                                debug_print_highpri ("%s",  "\n");
                                break;
                            }

                            case REG_BITRATEMSB : { // 0x03
                                bitRate = (regVal << 8);
                                break;
                            }

                            case REG_BITRATELSB : { // 0x04
                                bitRate bitor= regVal;
                                debug_print_highpri ("%s",  "Bit Rate (Chip Rate when Manchester encoding is enabled)\nBitRate : ");
                                unsigned long val = 32UL * 1000UL * 1000UL / bitRate;
                                debug_print_highpri ("%d\n", val);
                                break;
                            }

                            case REG_FDEVMSB : { // 0x05
                                freqDev = ((regVal & 0x3f) << 8);
                                break;
                            }

                            case REG_FDEVLSB : { // 0x06
                                freqDev bitor= regVal;
                                debug_print_highpri ("%s",  "Frequency deviation\nFdev : ");
                                uint32_t val = MULT_WITH_U32_X (freqDev, RF69_FSTEP_FLOAT32);
                                debug_print_highpri ("%u\n", val);
                                break;
                            }

                            case REG_FRFMSB : { // 0x07
                                unsigned long tempVal = regVal;
                                freqCenter = (tempVal << 16);
                                break;
                            }

                            case REG_FRFMID : { // 0x08
                                unsigned long tempVal = regVal;
                                freqCenter = freqCenter bitor (tempVal << 8);
                                break;
                            }

                            case REG_FRFLSB : { // 0x09
                                freqCenter = freqCenter bitor regVal;
                                debug_print_highpri ("%s",  "RF Carrier frequency\nFRF : ");
                                uint32_t val = MULT_WITH_U32_X (freqCenter, RF69_FSTEP_FLOAT32);
                                debug_print_highpri ("%u\n", val);
                                break;
                            }

                            case REG_OSC1 : { // 0x0A
                                debug_print_highpri ("%s",  "RC calibration control & status\nRcCalDone : ");
                                if (0x40 bitand regVal) {
                                    debug_print_highpri ("%s",  "1 -> RC calibration is over\n");
                                } else {
                                    debug_print_highpri ("%s",  "0 -> RC calibration is in progress\n");
                                }

                                debug_print_highpri ("%s",  "\n");
                                break;
                            }

                            case REG_AFCCTRL : { // 0x0B
                                debug_print_highpri ("%s",
                                        "Improved AFC routine for signals with modulation index lower than 2. See 3.4.16 for details\nAfcLowBetaOn : ");
                                if (0x20 & regVal) {
                                    debug_print_highpri ("%s",  "1 -> Improved AFC routine\n");
                                } else {
                                    debug_print_highpri ("%s",  "0 -> Standard AFC routine\n");
                                }
                                debug_print_highpri ("%s",  "\n");
                                break;
                            }

                            case REG_LOWBAT : { // 0x0C
                                debug_print_highpri ("%s",  "Reserved\n\n");
                                break;
                            }

                            case REG_LISTEN1 : { // 0x0D
                                uint8_t val;
                                debug_print_highpri ("%s",  "Resolution of Listen mode Idle time (calibrated RC osc):\nListenResolIdle : ");
                                val = regVal >> 6;
                                if (val == 0b00) {
                                    debug_print_highpri ("%s",  "00 -> reserved\n");
                                } else if (val == 0b01) {
                                    debug_print_highpri ("%s",  "01 -> 64 us\n");
                                } else if (val == 0b10) {
                                    debug_print_highpri ("%s",  "10 -> 4.1 ms\n");
                                } else if (val == 0b11) {
                                    debug_print_highpri ("%s",  "11 -> 262 ms\n");
                                }

                                debug_print_highpri ("%s",  "\nResolution of Listen mode Rx time (calibrated RC osc):\nListenResolRx : ");
                                val = (regVal >> 4) & 0x3;
                                if (val == 0b00) {
                                    debug_print_highpri ("%s",  "00 -> reserved\n");
                                } else if (val == 0b01) {
                                    debug_print_highpri ("%s",  "01 -> 64 us\n");
                                } else if (val == 0b10) {
                                    debug_print_highpri ("%s",  "10 -> 4.1 ms\n");
                                } else if (val == 0b11) {
                                    debug_print_highpri ("%s",  "11 -> 262 ms\n");
                                }

                                debug_print_highpri ("%s",  "\nCriteria for packet acceptance in Listen mode:\nListenCriteria : ");
                                if (0x8 & regVal) {
                                    debug_print_highpri ("%s",  "1 -> signal strength is above RssiThreshold and SyncAddress matched\n");
                                } else {
                                    debug_print_highpri ("%s",  "0 -> signal strength is above RssiThreshold\n");
                                }

                                debug_print_highpri ("%s",  "\nAction taken after acceptance of a packet in Listen mode:\nListenEnd : ");
                                val = (regVal >> 1) & 0x3;
                                if (val == 0b00) {
                                    debug_print_highpri ("%s",  "00 -> chip stays in Rx mode. Listen mode stops and must be disabled (see section 4.3)\n");
                                } else if (val == 0b01) {
                                    debug_print_highpri ("%s%s",
                                            "01 -> chip stays in Rx mode until PayloadReady or Timeout interrupt occurs. ",
                                            "It then goes to the mode defined by Mode. Listen mode stops and must be disabled (see section 4.3)\n");
                                } else if (val == 0b10) {
                                    debug_print_highpri ("%s%s",
                                            "10 -> chip stays in Rx mode until PayloadReady or Timeout occurs. ",
                                            "Listen mode then resumes in Idle state.  FIFO content is lost at next Rx wakeup.\n");
                                } else if (val == 0b11) {
                                    debug_print_highpri ("%s",  "11 -> Reserved\n");
                                }

                                debug_print_highpri ("%s",  "\n");
                                break;
                            }
                            default : {
                                // No code
                            } break;
                        }
                        #endif
                    } // for-loop
                    PRINT_ADD_ERROR_BITS_IFF ("i_radio.uspi_readAllRegs", rfm69_internals.error_bits);
                    VALUE_XSCOPE_LIB (RFM69_VALUE,202);
                } break;
        #endif
        }
    }
    // No code here when [[combinable]] (and of course not when [[distributable]]. RFM69=011: we simulate this code with the timerafter
}

// This "interrupt function" is triggered for every change of the defined interrupt pin, p_irq (IRQ).
// The status is sent to the listener that sits in a selective choice (select/case) and would
// receive one of three events: gone high, gone low and high-timeout, sent from here. When this pin goes high
// the RFM69 radio signals something (like RX-data or TX-done or some error code). For the pin to go low again
// the client code must talk with the radio over the i_radio interface handled in code above here. So we
// will not lose any IRQ events. However, since the c_irq_update channel is synchronous it's the client
// that decides when this event will be picked up. Typically, if the handling of the event is long,
// like running the debugger with lots of debug_print and display update, then the SW may empty the FIFO
// and the IRQ may go low but may go high again immediately of there is some radio signal coming in
// from another source. At the moment I have two such sources sending every 4 second, but those
// sendings are not correlated. This fast-coming RX will be attempted sent from here, but not picked up
// in the client before the previous RX's debug_print and display write is done. With synchronous
// debug_print this may take up to 1600 ms.

// The timeout is handled here and pin_*still_high_timeout is repeated until pin_gone_low. This repeat is necessary
// so that, in case the RFM69 radio malfunctions, the client would know that the pin is indeed still high.
// The response to the first of these probably is to issue an i_radio.uspi_ultimateIRQclear(); interface call.
// This should take the IRQ line low. We tested a version where the client asked this function via a separate
// interface call about the state of the pin, but having a task that combines spontaneous sending of IRQ
// update with queries is quite complex for the coder on the client side. It's so easy to deadlock. Besides,
// it takes more resources with one chanend more (it looks like it would have been two chanends more, but the
// compiler knows how to use its chanends economically). This was function IRQ_detect_and_follow_task, now
// obsoleted

[[combinable]] // Cannot be [[distributable]] since pinsneq
void IRQ_interrupt_task (
            chanend        ?c_irq_update,        // Sends irq_update_e
            in port         p_irq,               // Interrupt pin
            probe_pins_t  &?p_probe,             // "null" in call to omit
     const unsigned         irq_high_max_time_ms // Stuck? pin timeout
    )
{
    timer    tmr; // Doesn't seem to cost any HW timer
    time32_t time;
    pin_e    pin_value;

    p_irq :> pin_value;

    while(1) {
        select {
            case p_irq when pinsneq (pin_value) :> pin_value: { // edge triggering

                if (not isnull(p_probe)) {
                    p_probe.probe_when_irq <: pin_value; // A TEST PIN
                } else {}

                if (pin_value == high) {
                    if (not isnull(c_irq_update)) {
                                                       // Pin high by radio like when like when RX-data or TX-done.
                        c_irq_update <: pin_gone_high; // Send "pin_gone_high since last pin_gone_low sent".
                                                       // Handle so that radio takes pin low.
                    } else {}
                    tmr :> time; // Reset "now" time. Read time when pin_gone_high has been seen by client
                    time += (irq_high_max_time_ms * XS1_TIMER_KHZ); // Delay specified time

                } else { // pin_value == low
                    if (not isnull(c_irq_update)) {
                                                      // Radio handled correcly so that pin went low.
                        c_irq_update <: pin_gone_low; // Send "pin_gone_low since last pin_gone_high sent".
                                                      // Wait in selective choice for pin_gone_high again
                    } else {}
                }
            } break;

            case (pin_value == high) => tmr when timerafter(time) :> void: {
                // Observe that even if time ticks while this task is blocking on any sending above,
                // this _timeout_ will not be taken before the task is back in the select
                if (not isnull(c_irq_update)) {
                    c_irq_update <: pin_still_high_timeout; // Send or repeat. Too short timeout or radio stuck
                } else {}
                time += (irq_high_max_time_ms * XS1_TIMER_KHZ); // Delay specified time from "now"
            } break;
        }
    }
}


